/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */

lexer grammar VerilogPreprocessor;

@members {
  /** Reference to the parent Verilog front-end. */
  private ru.ispras.verilog.parser.VerilogFrontend frontend;

  /**
   * Sets the Verilog front-end.
   *
   * @param frontend the reference to the parent Verilog front-end.
   */
  public void setFrontend(final ru.ispras.verilog.parser.VerilogFrontend frontend) {
    this.frontend = frontend;
  }

  /**
   * Skips a token if the scope is hidden.
   */
  private void pp() {
    if(frontend.isHidden()) {
      skip();
    }
  }

  /**
   * Injects a special token indicating start of a new sub-lexer.
   */
  private void returnNewSourceToken() {
    state.token = new CommonToken(ru.ispras.verilog.parser.util.TokenSourceStack.NEW_SOURCE_TOKEN);
  }

  @Override
  public void emitErrorMessage(final String msg) {
    if (!frontend.isHidden()) {
      ru.ispras.verilog.parser.VerilogLogger.error(msg);
    }
  }
}

//--------------------------------------------------------------------------------------------------
// Preprocessor: If-Then-Else Directives
//--------------------------------------------------------------------------------------------------

PP_IFDEF : '`ifdef' WS key=IDENT
{
  // `ifdef is processed even in the hidden scope.
  frontend.onIfdef($key.getText());
  skip();
};

PP_IFNDEF : '`ifndef' WS key=IDENT
{
  // `ifndef is processed even in the hidden scope.
  frontend.onIfndef($key.getText());
  skip();
};

PP_ELSIF : '`elsif' WS key=IDENT
{
  // `elsif is processed even in the hidden scope.
  frontend.onElsif($key.getText());
  skip();
};

PP_ELSE : '`else'
{
  // `else is processed even in the hidden scope.
  frontend.onElse();
  skip();
};

PP_ENDIF : '`endif'
{
  // `endif is processed even in the hidden scope.
  frontend.onEndif();
  skip();
};

//--------------------------------------------------------------------------------------------------
// Preprocessor: Begin-End-Keywords Directives
//--------------------------------------------------------------------------------------------------

PP_BEGIN_KEYWORDS : '`begin_keywords' WS '"' version=VERSION '"' REST
{
  if (!frontend.isHidden()) {
    frontend.onBeginKeywords($version.getText());
  }
  skip();
};

PP_END_KEYWORDS : '`end_keywords' REST
{
  if (!frontend.isHidden()) {
    frontend.onEndKeywords();
  }
  skip();
};

//--------------------------------------------------------------------------------------------------
// Preprocessor: Define-Undefine Directives
//--------------------------------------------------------------------------------------------------

PP_DEFINE : '`define' WS key=IDENT
              ( ('(')=> '(' WS? defParams=PARAMS WS? ')' val=LINE
              | val=LINE ) (NL | EOF)
{
  if (!frontend.isHidden()) {
    frontend.onDefine($key.getText(), $val.getText(), $defParams == null? null : $defParams.getText());
  }
  skip();
};

PP_UNDEF : '`undef' WS key=IDENT REST
{
  if (!frontend.isHidden()) {
    frontend.onUndef($key.getText());
  }
  skip();
};

//--------------------------------------------------------------------------------------------------
// Preprocessor: Line Directive
//--------------------------------------------------------------------------------------------------

PP_LINE : '`line' WS NUMBER WS '"' filename=FILENAME '"' WS NUMBER REST
{
  skip();
};

//--------------------------------------------------------------------------------------------------
// Preprocessor: Include & Expand Directives
//--------------------------------------------------------------------------------------------------

PP_INCLUDE : '`include' WS '"' filename=FILENAME '"' REST
{
  if (!frontend.isHidden()) {
    frontend.lexInclude($filename.getText());
    returnNewSourceToken();
  } else {
    skip();
  }
};

PP_EXPAND : '`' key=IDENT ( params =  INLINE_PARAMS )?
{
  if (!frontend.isHidden()) {
    final String substitution = frontend.expand($key.getText(),
      $params == null ? null : $params.getText().substring(1, $params.getText().length() - 1));

    frontend.lexExpand(substitution);
    returnNewSourceToken();
  } else {
    skip();
  }
};

//--------------------------------------------------------------------------------------------------
// Preprocessor: Other Directives
//--------------------------------------------------------------------------------------------------

PP_TIMESCALE               : '`timescale'               REST { skip(); };
PP_CELLDEFINE              : '`celldefine'              REST { skip(); };
PP_ENDCELLDEFINE           : '`endcelldefine'           REST { skip(); };
PP_UNCONNECTED_DRIVE       : '`unconnected_drive'       REST { skip(); };
PP_NOUNCONNECTED_DRIVE     : '`nounconnected_drive'     REST { skip(); };
PP_PRAGMA                  : '`pragma'                  REST { skip(); };
PP_RESETALL                : '`resetall'                REST { skip(); };
PP_DEFAULT_DECAY_TIME      : '`default_decay_time'      REST { skip(); };
PP_DEFAULT_TRIREG_STRENGTH : '`default_trireg_strength' REST { skip(); };
PP_DELAY_MODE_DISTRIBUTED  : '`delay_mode_distributed'  REST { skip(); };
PP_DELAY_MODE_PATH         : '`delay_mode_path'         REST { skip(); };
PP_DELAY_MODE_UNIT         : '`delay_mode_unit'         REST { skip(); };
PP_DELAY_MODE_ZERO         : '`delay_mode_zero'         REST { skip(); };

//--------------------------------------------------------------------------------------------------
// Common: Whitespace and Comments
//--------------------------------------------------------------------------------------------------


fragment
SPACE : (' ' | '\t') ;

fragment
WS : (SPACE)+ ;

fragment
LINE : (~('\n' | '\r'))* ;

fragment
NL : (('\r')? '\n')+ ;

fragment
REST : LINE (NL | EOF) ;

fragment
NUMBER : DIGIT (DIGIT | '_')* ;

fragment
DIGIT : ('0'..'9') ;

fragment
IDENT : ('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'_'|'$'|'0'..'9')* ;

fragment
VERSION : (~('"' | '\n'))* ;

fragment
FILENAME : (~('"' | '\n'))* ;

fragment
PARAMS : IDENT (WS? ',' WS? IDENT)* ;

fragment
INLINE_PARAMS
@init {
  int o = 1;
} :
('(' ({o>0}?=>(('(' {o++;}) | (~('\n' | '\r' | '(' | ')')) | (')' {o--;}) ))*) ;

//--------------------------------------------------------------------------------------------------
