/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.RuleReturnScope;
import org.antlr.runtime.TokenRewriteStream;
import org.antlr.runtime.TokenSource;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeAdaptor;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.antlr.runtime.tree.Tree;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.grammar.VerilogLexer;
import ru.ispras.verilog.parser.grammar.VerilogParser;
import ru.ispras.verilog.parser.grammar.VerilogTreeBuilder;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.util.DefineStructure;
import ru.ispras.verilog.parser.util.IncludeFileFinder;
import ru.ispras.verilog.parser.util.IncludeStructure;
import ru.ispras.verilog.parser.util.TokenSourceCallback;
import ru.ispras.verilog.parser.util.TokenSourceStack;

import java.io.BufferedReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;
import java.util.ListIterator;
import java.util.Stack;

/**
 * {@link VerilogFrontend} implements the Verilog front-end.
 *
 * <p>The front-end includes the preprocessor, the lexer, the parser and the static analyzer.</p>
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogFrontend {
  public static boolean DEBUG = false;

  //================================================================================================
  // Preprocessor
  //================================================================================================

  private class OnTokenSourceEnd implements TokenSourceCallback {
    @Override
    public void action(final TokenSource source) {}
  }

  enum IfDefScope {
    IF_TRUE,
    IF_FALSE,
    ELSE_TRUE,
    ELSE_FALSE
  }

  private final IncludeFileFinder finder = new IncludeFileFinder();
  private final Hashtable<String, DefineStructure> defines = new Hashtable<>();
  private final Stack<IfDefScope> ifdefs = new Stack<>();
  private final List<IncludeStructure> includeFileNames = new ArrayList<>();
  private final Stack<VerilogStandard> keywords = new Stack<>();

  private void addFile(final String source, final String fileName) {
    this.includeFileNames.add(new IncludeStructure(fileName, source));
  }

  /**
   * Returns the name of the file being processed by the front-end.
   *
   * @return the name of the file.
   */
  public String getCurrentFileName() {
    return this.source.getSourceName().isEmpty() ? "<none>" : this.source.getSourceName();
  }

  public void addPath(final String path) {
    InvariantChecks.checkNotNull(path);
    finder.addPaths(path);
  }

  public boolean isDefined(final String key) {
    InvariantChecks.checkNotNull(key);
    return defines.containsKey(key.trim());
  }

  public boolean isHidden() {
    if (!ifdefs.empty()) {
      final IfDefScope scope = ifdefs.peek();
      return scope == IfDefScope.IF_FALSE || scope == IfDefScope.ELSE_FALSE;
    }

    return false;
  }

  public void onDefine(final String key, final String val, final String defParams) {
    // A multi-line comment inside a value becomes its part, while a single-line comment does not.
    final int index = val.indexOf("//");

    String value = index == -1 ? val : val.substring(0, index);

    DefineStructure defineStructure = new DefineStructure(defParams != null);

    if (defParams != null) {

      String[] params = defParams.split(",");

      for (int p = 0; p < params.length; p++) {

        List<Integer> listIdx = getListIdx(value, params[p]);

        for (Integer i : listIdx) {
          if (!isParam(value, params[p], i)) {
            listIdx.remove(i);
          }
        }

        String param = "%" + (p + 1) + "$s";

        StringBuilder stringBuffer = new StringBuilder(value);

        int iadding = 0;

        for (Integer i : listIdx) {
          stringBuffer.replace(i + iadding, i + iadding + params[p].length(), param);
          iadding += param.length() - params[p].length();
        }

        value = stringBuffer.toString();
      }
    }

    defineStructure.setValue(value.trim());

    defines.put(key.trim(), defineStructure);
  }

  private List<Integer> getListIdx(String str, String substr) {

    List<Integer> listIdx = new ArrayList<>();

    int lastIndex = -1;

    while ((lastIndex = str.indexOf(substr, lastIndex + 1)) != -1) {
      listIdx.add(lastIndex);
    }

    return listIdx;
  }

  private Boolean isParam(String value, String param, int index) {
    // IDENT : ('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'_'|'$'|'0'..'9')*
    if (index != 0) {

      char preChar = value.charAt(index - 1);

      if (preChar == '_'
          || (preChar >= 'a' && preChar <= 'z')
          || (preChar >= 'A' && preChar <= 'Z')
          || preChar == '`') {
        return false;
      }
    }

    if (index + param.length() < value.length()) {

      char postChar = value.charAt(index + param.length());

      if (postChar == '_'
          || (postChar >= 'a' && postChar <= 'z')
          || (postChar >= 'A' && postChar <= 'Z')
          || (postChar >= '0' && postChar <= '9')
          || postChar == '$') {
        return false;
      }
    }

    return true;
  }

  public void onUndef(final String key) {
    defines.remove(key.trim());
  }

  public void onIfdef(final String key) {
    if (isHidden() || !isDefined(key)) {
      ifdefs.push(IfDefScope.IF_FALSE);
    } else {
      ifdefs.push(IfDefScope.IF_TRUE);
    }
  }

  public void onIfndef(final String key) {
    if (isHidden() || isDefined(key)) {
      ifdefs.push(IfDefScope.IF_FALSE);
    } else {
      ifdefs.push(IfDefScope.IF_TRUE);
    }
  }

  public void onElse() {
    InvariantChecks.checkTrue(!ifdefs.empty());

    final IfDefScope scope = ifdefs.pop();

    if (isHidden() || scope == IfDefScope.IF_TRUE || scope == IfDefScope.ELSE_FALSE) {
      ifdefs.push(IfDefScope.ELSE_FALSE);
    } else {
      ifdefs.push(IfDefScope.ELSE_TRUE);
    }
  }

  public void onElsif(final String key) {
    InvariantChecks.checkTrue(!ifdefs.empty());

    final IfDefScope scope = ifdefs.pop();

    if (isHidden() || scope == IfDefScope.IF_TRUE || scope == IfDefScope.ELSE_FALSE) {
      ifdefs.push(IfDefScope.ELSE_FALSE);
    } else {
      if (!isDefined(key)) {
        ifdefs.push(IfDefScope.IF_FALSE);
      } else {
        ifdefs.push(IfDefScope.IF_TRUE);
      }
    }
  }

  public void onEndif() {
    InvariantChecks.checkTrue(!ifdefs.empty());
    ifdefs.pop();
  }

  public String expand(final String key, final String params) {

    final String defineKey = key.trim();

    List<String> paramsList = splitParams(params);

    DefineStructure defineStructure = defines.get(defineKey);

    if (defineStructure == null) {
      VerilogLogger.error("Definition of '%s' has not been found", defineKey);
      VerilogTranslator.exit();
    }

    return defineStructure.getString(paramsList);
  }

  private List<String> splitParams(String params) {

    List<String> paramList = new ArrayList<>();

    if (params == null) {
      return paramList;
    }

    StringBuilder param = new StringBuilder();

    int bracketCounter = 0;

    for (Character c : params.toCharArray()) {
      switch (c) {
        case '(':
          bracketCounter++;
          param.append(c);
          break;

        case ')':
          bracketCounter--;
          param.append(c);
          break;

        case ',':
          if (bracketCounter == 0) {
            paramList.add(param.toString().trim());
            param = new StringBuilder();
          } else {
            param.append(c);
          }
          break;

        default:
          param.append(c);
          break;
      }
    }

    paramList.add(param.toString());

    return paramList;
  }

  public void onBeginKeywords(final String versionSpecifier) {
    final VerilogStandard standard;

    if (VerilogStandard.IEEE_1364_1995.getName().equals(versionSpecifier)) {
      standard = VerilogStandard.IEEE_1364_1995;
    } else if (VerilogStandard.IEEE_1364_2001_NC.getName().equals(versionSpecifier)) {
      standard = VerilogStandard.IEEE_1364_2001_NC;
    } else if (VerilogStandard.IEEE_1364_2001.getName().equals(versionSpecifier)) {
      standard = VerilogStandard.IEEE_1364_2001;
    } else if (VerilogStandard.IEEE_1364_2005.getName().equals(versionSpecifier)) {
      standard = VerilogStandard.IEEE_1364_2005;
    } else {
      standard = null;
    }

    InvariantChecks.checkNotNull(standard);
    onBeginKeywords(standard);
  }

  public void onBeginKeywords(final VerilogStandard standard) {
    VerilogLogger.debug("begin_keywords %s", standard);
    keywords.push(standard);
  }

  public void onEndKeywords() {
    InvariantChecks.checkNotEmpty(keywords);
    keywords.pop();
  }

  public boolean isAllowed(final VerilogStandard keywordStandard) {
    final VerilogStandard supportedStandard =
        keywords.isEmpty() ? VerilogStandard.IEEE_1364_2005 : keywords.peek();

    return keywordStandard.isCompliantWith(supportedStandard);
  }

  //================================================================================================
  // Lexer
  //================================================================================================

  private final TokenSourceStack source = new TokenSourceStack(new OnTokenSourceEnd());

  public TokenSource startLexer(final List<String> filenames) {
    ListIterator<String> iterator = filenames.listIterator(filenames.size());

    // Process the files in reverse order (emulate inclusion).
    while (iterator.hasPrevious()) {
      final String filename = iterator.previous();
      includeFile(filename);
    }

    return source;
  }

  public void includeFile(String filename){
    final ANTLRFileStream stream = finder.openFileAsAntlrStream(filename);

    if (stream == null) {
      VerilogLogger.error("Include file '%s' has not been found", filename);
      return;
    }
    includeFileNames.add(new IncludeStructure(filename, ""));

    startSublexer(stream);
  }

  public void startSublexer(final CharStream chars) {
    source.push(new VerilogLexer(chars, this));
  }

  public void lexExpand(final String substitution) {
    if (substitution != null && !substitution.isEmpty()) {
      VerilogLogger.debug("Expanding macro '%s' ...", substitution);

      final ANTLRStringStream stream = new ANTLRStringStream(substitution);
      startSublexer(stream);
    }
  }

  public void lexInclude(final String filename) {
    final ANTLRFileStream stream = finder.openFileAsAntlrStream(filename);

    if (stream == null) {
      VerilogLogger.error("Include file '%s' has not been found", filename);
      return;
    }

    if (!isFilenameContains(stream.getSourceName())) {
      VerilogLogger.debug("Including file '%s' ...", filename);

      addFile(getCurrentFileName(), stream.getSourceName());
      startSublexer(stream);
    } else {
      VerilogLogger.error("Cycle inclusion at: '%s'", stream.getSourceName());
    }
  }

  private boolean isFilenameContains(String filename) {
    IncludeStructure includeStructure = new IncludeStructure(source.getSourceName(), "");

    while (includeFileNames.lastIndexOf(includeStructure) != -1
        && includeFileNames.get(includeFileNames.lastIndexOf(includeStructure)) != null) {

      includeStructure = includeFileNames.get(includeFileNames.lastIndexOf(includeStructure));

      if (includeStructure.getSource().equals(filename)) {
        return true;
      } else {
        includeStructure = new IncludeStructure(includeStructure.getSource(), "");
      }
    }
    return false;
  }

  //================================================================================================
  // Parser
  //================================================================================================

  public CommonTree startParser(final CommonTokenStream tokens) throws RecognitionException {
    InvariantChecks.checkNotNull(tokens);

    final VerilogParser parser = new VerilogParser(tokens);
    parser.setTreeAdaptor(new CommonTreeAdaptor());

    final RuleReturnScope result = parser.startRule();
    return (CommonTree) result.getTree();
  }

  public CommonTree startParser(final TokenSource source) throws RecognitionException {
    InvariantChecks.checkNotNull(source);

    final CommonTokenStream tokens = new TokenRewriteStream();
    tokens.setTokenSource(source);

    return startParser(tokens);
  }

  //================================================================================================
  // Tree Builder
  //================================================================================================

  public VerilogNode startBuilder(final CommonTreeNodeStream nodes) throws RecognitionException {
    InvariantChecks.checkNotNull(nodes);

    final VerilogTreeBuilder builder = new VerilogTreeBuilder(nodes, this);
    return builder.startRule();
  }

  public VerilogNode startBuilder(final CommonTree tree) throws RecognitionException {
    InvariantChecks.checkNotNull(tree);
    return startBuilder(new CommonTreeNodeStream(tree));
  }

  //================================================================================================
  // Analyzer
  //================================================================================================

  public void loadLibrary(final String filename) {
    final BufferedReader reader = finder.openFileAsReader(filename);

    if (reader == null) {
      VerilogLogger.error("Library file '%s' has not been found", filename);
      VerilogTranslator.exit();
    }

    VerilogLibrary.get().loadLibrary(reader);
  }

  public VerilogNode start(final List<String> filenames) throws RecognitionException {
    final TokenSource tokens = startLexer(filenames);
    final CommonTree tree = startParser(tokens);

    if (DEBUG) {
      printTree(tree, 0);
    }

    return startBuilder(tree);
  }

  public VerilogNode start(final String[] filenames) throws RecognitionException {
    return start(Arrays.asList(filenames));
  }

  //================================================================================================
  // Debug
  //================================================================================================

  private void printTree(final CommonTree tree, final int indent) {
    if (tree != null) {
      final StringBuffer buffer = new StringBuffer(indent);

      for (int i = 0; i < indent; i++) {
        buffer.append("  ");
      }

      final String indentation = buffer.toString();

      if (tree.getParent() == null) {
        VerilogLogger.debug("%s%s", indentation, tree.getText());
      }

      for (int i = 0; i < tree.getChildCount(); i++) {
        final Tree child = tree.getChild(i);
        InvariantChecks.checkNotNull(child);

        VerilogLogger.debug("%sBEGIN: %s", indentation, child);
        printTree((CommonTree) tree.getChild(i), indent + 1);
        VerilogLogger.debug("%sEND: %s", indentation, child);
      }
    }
  }
}
