/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.VerilogModule;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogNullStatement;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogElementType.Type;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.StringTokenizer;

/**
 * {@link VerilogLibrary} stores the standard declarations.
 *
 * <p>
 * The standard declarations includes system tasks and functions, primitive gates, etc.
 * </p>
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogLibrary {
  private static VerilogLibrary instance;
  
  /**
   * Initializes (resets) the library.
   */
  public static void init() {
    instance = null;
  }

  /**
   * Returns an library instance.
   * 
   * @return a new library instance.
   */
  public static VerilogLibrary get() {
    if (instance == null) {
      instance = new VerilogLibrary();
    }

    return instance;
  }

  /** Contains the standard declarations. */
  private final Collection<AbstractNode<VerilogNode.Tag>> declarations = new ArrayList<>();

  private VerilogLibrary() {
    // [IEEE Std 1364-2005]
    //
    // 7. Gate- and switch-level modeling
    //
    // 7.2 and, nand, nor, or, xor, and xnor gates
    gatePrimitive("and");
    gatePrimitive("nand");
    gatePrimitive("nor");
    gatePrimitive("or");
    gatePrimitive("xor");
    gatePrimitive("xnor");

    // 7.3 buf and not gates
    bufPrimitive("buf");
    bufPrimitive("not");

    // 7.4 bufif1, bufif0, notif1, and notif0 gates
    bufifPrimitive("bufif1");
    bufifPrimitive("bufif0");
    bufifPrimitive("notif1");
    bufifPrimitive("notif0");

    // 7.5 MOS switches
    mosPrimitive("cmos");
    mosPrimitive("nmos");
    mosPrimitive("pmos");
    mosPrimitive("rcmos");
    mosPrimitive("rnmos");
    mosPrimitive("rpmos");

    // 7.6 Bidirectional pass switches
    tranPrimitive("tran");
    tranPrimitive("rtran");
    tranPrimitive("tranif1");
    tranPrimitive("rtranif1");
    tranPrimitive("tranif0");
    tranPrimitive("rtranif0");

    // 7.8 pullup and pulldown sources
    pullPrimitive("pullup");
    pullPrimitive("pulldown");

    // 15. Timing checks
    //
    // 15.2 Timing checks using stability window
    systemTask("$setup");
    systemTask("$hold");
    systemTask("$setuphold");
    systemTask("$removal");
    systemTask("$recovery");
    systemTask("$recrem");

    // 15.3 Timing checks for clock and control signals
    systemTask("$skew");
    systemTask("$timeskew");
    systemTask("$fullskew");
    systemTask("$width");
    systemTask("$period");
    systemTask("$nochange");

    // 17. System tasks and functions
    //
    // Display tasks
    systemTask("$display");
    systemTask("$displayb");
    systemTask("$displayh");
    systemTask("$displayo");
    systemTask("$strobe");
    systemTask("$strobeb");
    systemTask("$strobeh");
    systemTask("$strobeo");
    systemTask("$write");
    systemTask("$writeb");
    systemTask("$writeh");
    systemTask("$writeo");
    systemTask("$monitor");
    systemTask("$monitorb");
    systemTask("$monitorh");
    systemTask("$monitoro");
    systemTask("$monitoroff");
    systemTask("$monitoron");

    // File I/O tasks
    systemTask("$fclose");
    systemTask("$fdisplay");
    systemTask("$fdisplayb");
    systemTask("$fdisplayh");
    systemTask("$fdisplayo");
    systemTask("$fstrobe");
    systemTask("$fstrobeb");
    systemTask("$fstrobeh");
    systemTask("$fstrobeo");
    systemTask("$swrite");
    systemTask("$swriteb");
    systemTask("$swriteh");
    systemTask("$swriteo");
    systemTask("$fflush");
    systemTask("$sdf_annotate");
    systemTask("$fwrite");
    systemTask("$fwriteb");
    systemTask("$fwriteh");
    systemTask("$fwriteo");
    systemTask("$fmonitor");
    systemTask("$fmonitorb");
    systemTask("$fmonitorh");
    systemTask("$fmonitoro");
    systemTask("$sformat");
    systemTask("$readmemb");
    systemTask("$readmemh");

    systemFunction("$feof",   Type.INTEGER);
    systemFunction("$ferror", Type.INTEGER);
    systemFunction("$fgetc",  Type.INTEGER);
    systemFunction("$fgets",  Type.INTEGER);
    systemFunction("$fopen",  Type.INTEGER);
    systemFunction("$fread",  Type.INTEGER);
    systemFunction("$fscanf", Type.INTEGER);
    systemFunction("$fseek",  Type.INTEGER);
    systemFunction("$ftell",  Type.INTEGER);
    systemFunction("$rewind", Type.INTEGER);
    systemFunction("$sscanf", Type.INTEGER);
    systemFunction("$ungetc", Type.INTEGER);

    // Timescale tasks
    systemTask("$printtimescale");
    systemTask("$timeformat");

    // Simulation control tasks
    systemTask("$finish");
    systemTask("$stop");

    // PLA modeling tasks
    systemTask("$async$and$array");
    systemTask("$async$nand$array");
    systemTask("$async$or$array");
    systemTask("$async$nor$array");
    systemTask("$async$and$plane");
    systemTask("$async$nand$plane");
    systemTask("$async$or$plane");
    systemTask("$async$nor$plane");

    systemTask("$sync$and$array");
    systemTask("$sync$nand$array");
    systemTask("$sync$or$array");
    systemTask("$sync$nor$array");
    systemTask("$sync$and$plane");
    systemTask("$sync$nand$plane");
    systemTask("$sync$or$plane");
    systemTask("$sync$nor$plane");

    // Stochastic analysis tasks
    systemTask("$q_initialize");
    systemTask("$q_remove");
    systemTask("$q_add");
    systemTask("$q_full");
    systemTask("$q_exam");

    // Simulation time functions
    systemFunction("$time",             Type.TIME);
    systemFunction("$stime",            Type.TIME);
    systemFunction("$realtime",         Type.REALTIME);

    // Conversion functions
    systemFunction("$itor",             Type.REAL);
    systemFunction("$rtoi",             Type.INTEGER);
    systemFunction("$bitstoreal",       Type.REAL);
    systemFunction("$realtobits",       Type.INTEGER);
    systemFunction("$signed",           Type.INTEGER);
    systemFunction("$unsigned",         Type.INTEGER);

    // Probabilistic distribution functions
    systemFunction("$random",           Type.INTEGER); 
    systemFunction("$dist_chi_square",  Type.INTEGER);
    systemFunction("$dist_erlang",      Type.INTEGER);
    systemFunction("$dist_exponential", Type.INTEGER);
    systemFunction("$dist_normal",      Type.INTEGER);
    systemFunction("$dist_poisson",     Type.INTEGER);
    systemFunction("$dist_t",           Type.INTEGER);
    systemFunction("$dist_uniform",     Type.INTEGER);

    // Command line input
    systemFunction("$test$plusargs",    Type.INTEGER); // Boolean
    systemFunction("$value$plusargs",   Type.INTEGER); // String

    // Math functions
    systemFunction("$clog2",            Type.INTEGER);
    systemFunction("$ln",               Type.REAL);
    systemFunction("$log10",            Type.REAL); 
    systemFunction("$exp",              Type.REAL);
    systemFunction("$sqrt",             Type.REAL);
    systemFunction("$pow",              Type.REAL);
    systemFunction("$floor",            Type.REAL);
    systemFunction("$ceil",             Type.REAL);
    systemFunction("$sin",              Type.REAL);
    systemFunction("$cos",              Type.REAL);
    systemFunction("$tan",              Type.REAL);
    systemFunction("$asin",             Type.REAL);
    systemFunction("$acos",             Type.REAL);
    systemFunction("$atan",             Type.REAL);
    systemFunction("$atan2",            Type.REAL);
    systemFunction("$hypot",            Type.REAL);
    systemFunction("$sinh",             Type.REAL);
    systemFunction("$cosh",             Type.REAL);
    systemFunction("$tanh",             Type.REAL);
    systemFunction("$asinh",            Type.REAL);
    systemFunction("$acosh",            Type.REAL);
    systemFunction("$atanh",            Type.REAL);

    // 18. Value change dump (VCD) files
    //
    // 18.1 Creating four-state VCD file
    //
    // 18.1.1 Specifying name of dump file ($dumpfile)
    systemTask("$dumpfile");

    // 18.1.2 Specifying variables to be dumped ($dumpvars)
    systemTask("$dumpvars");

    // 18.1.3 Stopping and resuming the dump ($dumpoff/$dumpon)
    systemTask("$dumpoff");
    systemTask("$dumpon");

    // 18.1.4 Generating a checkpoint ($dumpall)
    systemTask("$dumpall");

    // 18.1.5 Limiting size of dump file ($dumplimit)
    systemTask("$dumplimit");

    // 18.1.6 Reading dump file during simulation ($dumpflush)
    systemTask("$dumpflush");

    // 18.3 Creating extended VCD file
    //
    // 18.3.1 Specifying dump file name and ports to be dumped ($dumpports)
    systemTask("$dumpports");

    // 18.3.2 Stopping and resuming the dump ($dumpportsoff/$dumpportson)
    systemTask("$dumpportsoff");
    systemTask("$dumpportson");

    // 18.3.3 Generating a checkpoint ($dumpportsall)
    systemTask("$dumpportsall");

    // 18.3.4 Limiting size of dump file ($dumpportslimit)
    systemTask("$dumpportslimit");

    // 18.3.5 Reading dump file during simulation ($dumpportsflush)
    systemTask("$dumpportsflush");

    // [IEEE Std 1800-2012]
    //
    // SVA system tasks & functions
    //
    // 16.9.3 Sampled value functions

    // Sampled value
    systemFunction("$sampled",        Type.REG);

    // Sampled value in a prior time step
    systemFunction("$past",           Type.REG);

    // Value change functions
    systemFunction("$rose",           Type.INTEGER); // Boolean
    systemFunction("$fell",           Type.INTEGER); // Boolean
    systemFunction("$stable",         Type.INTEGER); // Boolean
    systemFunction("$changed",        Type.INTEGER); // Boolean

    // 16.9.4 Global clocking past and future sampled value functions

    // Global clocking past sampled value functions
    systemFunction("$past_gclk",      Type.REG);
    systemFunction("$rose_gclk",      Type.INTEGER); // Boolean
    systemFunction("$fell_gclk",      Type.INTEGER); // Boolean
    systemFunction("$stable_gclk",    Type.INTEGER); // Boolean
    systemFunction("$changed_gclk",   Type.INTEGER); // Boolean

    // Global clocking future sampled value functions
    systemFunction("$future_gclk",    Type.REG);
    systemFunction("$rising_gclk",    Type.INTEGER); // Boolean
    systemFunction("$falling_gclk",   Type.INTEGER); // Boolean
    systemFunction("$steady_gclk",    Type.INTEGER); // Boolean
    systemFunction("$changing_gclk",  Type.INTEGER); // Boolean

  }

  public Collection<AbstractNode<VerilogNode.Tag>> getDeclarations() {
    return declarations;
  }

  public void loadLibrary(final BufferedReader reader) {
    try {
      for (String line = reader.readLine(); line != null; line = reader.readLine()) {
        final StringTokenizer tokenizer = new StringTokenizer(line);

        if (tokenizer.countTokens() == 0) {
          continue;
        }

        InvariantChecks.checkNotNull(tokenizer.countTokens() == 2 || tokenizer.countTokens() == 3);
        final String kind = tokenizer.nextToken();
        final String type = kind.equals("function") ? tokenizer.nextToken() : null;
        final String name = tokenizer.nextToken();

        if (type != null) {
          systemFunction(name, type.equals("real") ? Type.REAL : Type.INTEGER);
        } else {
          systemTask(name);
        }
      }
    } catch (final IOException e) {
      e.printStackTrace();
    }
  }
  
  private void gatePrimitive(final String name) {
    final VerilogModule gatePrimitive = new VerilogModule(null);

    gatePrimitive.setType(VerilogModule.Type.GATE_PRIMITIVE);
    gatePrimitive.setName(name);

    declarations.add(gatePrimitive);
  }
  
  private void bufPrimitive(final String name) {
    final VerilogModule bufPrimitive = new VerilogModule(null);

    bufPrimitive.setType(VerilogModule.Type.GATE_PRIMITIVE);
    bufPrimitive.setName(name);

    declarations.add(bufPrimitive);
  }
  
  public void bufifPrimitive(final String name) {
    final VerilogModule bufifPrimitive = new VerilogModule(null);

    bufifPrimitive.setType(VerilogModule.Type.GATE_PRIMITIVE);
    bufifPrimitive.setName(name);

    declarations.add(bufifPrimitive);
  }
  
  private void mosPrimitive(final String name) {
    final VerilogModule mosPrimitive = new VerilogModule(null);

    mosPrimitive.setType(VerilogModule.Type.GATE_PRIMITIVE);
    mosPrimitive.setName(name);

    declarations.add(mosPrimitive);
  }
  
  private void tranPrimitive(final String name) {
    final VerilogModule tranPrimitive = new VerilogModule(null);

    tranPrimitive.setType(VerilogModule.Type.GATE_PRIMITIVE);
    tranPrimitive.setName(name);

    declarations.add(tranPrimitive);
  }
  
  private void pullPrimitive(final String name) {
    final VerilogModule pullPrimitive = new VerilogModule(null);

    pullPrimitive.setType(VerilogModule.Type.GATE_PRIMITIVE);
    pullPrimitive.setName(name);

    declarations.add(pullPrimitive);
  }
  
  private void systemTask(final String name) {
    final VerilogProcedure systemTask = new VerilogProcedure(null);

    systemTask.setType(VerilogProcedure.Type.TASK);
    systemTask.setName(name);
    systemTask.setStatement(new VerilogNullStatement(null));

    declarations.add(systemTask);
  }
  
  private void systemFunction(final String name, final VerilogElementType.Type type) {
    final VerilogElementType elementType = new VerilogElementType();
    elementType.setType(type);

    final VerilogProcedure systemFunction = new VerilogProcedure(null);
    systemFunction.setType(VerilogProcedure.Type.FUNCTION);
    systemFunction.setName(name);
    systemFunction.setOutput(name, elementType);
    systemFunction.setStatement(new VerilogNullStatement(null));

    declarations.add(systemFunction);
  }
}
