/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.verilog.parser;

import org.apache.log4j.Appender;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Layout;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;

/**
 * {@link VerilogLogger} implements the logging facilities.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogLogger {
  /** The message logger. */
  private final Logger logger = Logger.getLogger(VerilogLogger.class);
  /** The number of errors. */
  private int errors = 0;
  /** The number of warnings. */
  private int warnings = 0;

  private VerilogLogger() {
    final Layout layout = new PatternLayout(PatternLayout.DEFAULT_CONVERSION_PATTERN);
    final Appender appender = new ConsoleAppender(layout);

    logger.addAppender(appender);
  }

  private static final VerilogLogger LOG = new VerilogLogger();

  public static void verbose(final boolean verbose) {
    LOG.setVerbose(verbose);
  }

  public static void reset() {
    LOG.errors = 0;
    LOG.warnings = 0;
  }

  public static void info(final String format, final Object ... args) {
    LOG.logInfo("INFO: " + String.format(format, args));
  }

  public static void debug(final String format, final Object ... args) {
    LOG.logDebug("DEBUG: " + String.format(format, args));
  }

  public static void warn(final String format, final Object ... args) {
    LOG.logWarn("WARN: " + String.format(format, args));
  }

  public static void error(final String format, final Object ... args) {
    LOG.logError("ERROR: " + String.format(format, args));
  }

  public static void stat() {
    LOG.logInfo("%d error(s), %d warning(s)", LOG.errors, LOG.warnings);
  }

  public static boolean failed() {
    return LOG.errors > 0;
  }

  private boolean verbose = false;

  public void setVerbose(final boolean verbose) {
    this.verbose = verbose;
  }

  public void logInfo(final String format, final Object ... args) {
    if (verbose) {
      logger.info(String.format(format, args));
    }
  }

  public void logDebug(final String format, final Object ... args) {
    if (verbose) {
      logger.debug(String.format(format, args));
    }
  }

  public void logWarn(final String format, final Object ... args) {
    logger.warn(String.format(format, args));
    warnings++;
  }

  public void logError(final String format, final Object ... args) {
    logger.error(String.format(format, args));
    errors++;
  }
}
