/*
 * Copyright 2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.backends.design.checker;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.walker.VerilogEmptyVisitor;

/**
 * {@link AssignCheckerModuleVisitor} is a key component for Verilog assign checker back-end.
 *
 * @author <a href="mailto:aadanilov_3@edu.hse.ru">Alexey Danilov</a>
 */
public class AssignCheckerModuleVisitor extends VerilogEmptyVisitor {

  /**
   * Performs checking assignment statement required by [IEEE Std 1364-2005]. 4.9.3.1.2 Assignment
   * to array elements
   *
   * @param node assign statement
   */
  @Override
  public Result onAssignStatementBegin(VerilogAssignStatement node) {
    if (node.isAssign()) {
      return assignmentTypeAssign(node);
    } else if (node.isBlocking()) {
      return assignmentTypeBlocking(node);
    }

    return super.onAssignStatementBegin(node);
  }

  private Result assignmentTypeAssign(final VerilogAssignStatement node) {
    final VerilogAssignment assignment = node.getAssignment();
    final VerilogExpression rhs = assignment.getRhsExpression().getMinExpression();
    final Node lhs = assignment.getLhsNode();

    if (lhs.isType(DataTypeId.BIT_VECTOR)) {
      return assignmentTypeAssignLhsBVector(rhs.getNode(), lhs);
    }

    return super.onAssignStatementBegin(node);
  }

  private Result assignmentTypeAssignLhsBVector(final Node rhs,
      final Node lhs) {
    if (rhs.getKind().equals(Kind.VALUE)) {
      final DataType dataType = lhs.getDataType();
      final int parameter = Integer.parseInt(dataType.getParameters()[0].toString());
      final String toString = rhs.getUserData().toString();
      final int indexOfOne = toString.indexOf('1');

      final boolean notAllBitsAreZeros = indexOfOne != -1;
      final boolean valueBitIsOutOfRange = parameter < toString.length() - indexOfOne;
      if (notAllBitsAreZeros && valueBitIsOutOfRange) {
        VerilogLogger.warn("Some bits from assignment statement with node '%s' will be ignored",
            lhs.toString());
      }
    }

    return Result.OK;
  }

  private Result assignmentTypeBlocking(final VerilogAssignStatement node) {
    final VerilogAssignment assignment = node.getAssignment();
    final Node lhs = assignment.getLhsNode();
    final VerilogExpression rhs = assignment.getRhsExpression().getMinExpression();

    if (lhs.isType(DataTypeId.MAP)) {
      return assignmentTypeBlockingLhsMap(rhs.getNode());
    } else if (lhs.isType(DataTypeId.BIT_VECTOR)) {
      return assignmentTypeBlockingLhsBitVector(rhs.getNode());
    }

    return Result.OK;
  }

  private Result assignmentTypeBlockingLhsBitVector(final Node exprNode) {
    if (exprNode.isType(DataTypeId.MAP)) {
      VerilogLogger.error("Attempt to write map to a bit vector");
    }
    return Result.OK;
  }

  private Result assignmentTypeBlockingLhsMap(final Node exprNode) {
    if (exprNode.isType(DataTypeId.BIT_VECTOR)) {
      VerilogLogger.error("Attempt to write bit vector to a map");
    }
    return Result.OK;
  }
}
