/*
 * Copyright 2019-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.backends.syntax.instance.transformer;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.sva.SvaAssertionStatement;
import ru.ispras.verilog.parser.sva.SvaDeclaration;
import ru.ispras.verilog.parser.sva.SvaInstantiation;
import ru.ispras.verilog.parser.sva.SvaPropertyDeclaration;
import ru.ispras.verilog.parser.sva.SvaSequenceDeclaration;
import ru.ispras.verilog.parser.sva.basis.SvaPortConnection;
import ru.ispras.verilog.parser.sva.expression.SvaCaseExpression;
import ru.ispras.verilog.parser.sva.expression.SvaCaseExpressionItem;
import ru.ispras.verilog.parser.sva.util.SvaUtils;
import ru.ispras.verilog.parser.walker.VerilogEmptyVisitor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * {@link SvaInstanceTransformerVisitor} represents the visitor that transforms mistakenly
 * parsed function and variable expressions into proper SVA instance expressions.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class SvaInstanceTransformerVisitor extends VerilogEmptyVisitor {

  @Override
  public Result onSequenceDeclarationBegin(final SvaSequenceDeclaration node) {
    final VerilogExpression verilogExpr = node.getExpression();
    transform(node, verilogExpr);
    return Result.OK;
  }

  @Override
  public Result onPropertyDeclarationBegin(final SvaPropertyDeclaration node) {
    final VerilogExpression verilogExpr = node.getProperty().getExpression();
    transform(node, verilogExpr);
    return Result.OK;
  }

  @Override
  public Result onAssertionStatementBegin(final SvaAssertionStatement node) {
    if (node.isConcurrent()) {
      final VerilogExpression propExpr = node.getProperty().getExpression();
      transform(node, propExpr);
    }

    if (node.isImmediate()) {
      final VerilogExpression expr = node.getExpression();
      transform(node, expr);
    }

    return Result.OK;
  }

  private void transform(final VerilogNode scope, final VerilogExpression verilogExpr) {
    if (verilogExpr instanceof SvaCaseExpression) {
      transform(scope, (SvaCaseExpression) verilogExpr);
    } else {
      final Node expr = verilogExpr.getNode();
      verilogExpr.setNode(transformNode(scope, expr));
    }
  }

  private void transform(final VerilogNode scope, final SvaCaseExpression caseExpression) {
    for (SvaCaseExpressionItem item : caseExpression.getItems()) {
      final VerilogExpression property = item.getProperty();
      transform(scope, property);
    }
  }

  private Node transformNode(final VerilogNode scope, final Node expr) {
    InvariantChecks.checkNotNull(expr);
    Node transformedExpr =
        Transformer.transform(expr, StandardOperation.FUNCTION, new TransformerRule() {
          @Override
          public boolean isApplicable(final Node node) {
            return ExprUtils.isOperation(node, StandardOperation.FUNCTION);
          }

          @Override
          public Node apply(final Node node) {
            final VerilogExpression expression = (VerilogExpression) node.getUserData();
            final String name = expression.getPath().toString();
            return getNode(scope, name, node);
          }
        });

    transformedExpr =
        Transformer.transform(transformedExpr, Node.Kind.VARIABLE, new TransformerRule() {
          @Override
          public boolean isApplicable(final Node node) {
            return ExprUtils.isKind(Node.Kind.VARIABLE, node);
          }

          @Override
          public Node apply(final Node node) {
            final String name = ((NodeVariable) node).getName();
            return getNode(scope, name, node);
          }
        });

    return transformedExpr;
  }

  private Node getNode(final VerilogNode scope, final String name, final Node node) {
    Node newNode = node;
    final SvaDeclaration declaration = SvaUtils.findSvaDeclaration(name, scope);
    if (declaration != null) {
      final List<Node> args =
          ExprUtils.isOperation(node) ? ((NodeOperation) node).getOperands() : null;
      final Collection<SvaPortConnection> connections = getConnections(args);
      final SvaInstantiation instance =
          SvaUtils.createSvaInstance(scope, name, declaration, connections, args);
      newNode = instance.getNode();
    }
    return newNode;
  }

  private Collection<SvaPortConnection> getConnections(final List<Node> operands) {
    final Collection<SvaPortConnection> connections = new ArrayList<>();
    if (operands != null) {
      for (int i = 1; i < operands.size(); i++) {
        final Node operand = operands.get(i);
        final SvaPortConnection connection = new SvaPortConnection();
        connection.setArgument(operand.getUserData());
        connections.add(connection);
      }
    }
    return connections;
  }

}
