/*
 * Copyright 2012-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.core;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * {@link AbstractNode} implements the basic abstract syntax tree (AST) node.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class AbstractNode<Tag extends Enum<Tag>> extends AbstractSymbolTable<Tag> {

  /** {@link Name} shows whether a node has a name. */
  private enum Name { YES, NO }

  /** {@link Scope} shows whether a node has a scope. */
  private enum Scope { YES, NO, TRANSPARENT }

  /** {@link NodeKind} classifies nodes according to their abilities to have names and scopes. */
  protected enum NodeKind {
    /** Module, function, etc. */
    CLASS(Name.YES, Scope.YES),
    /** Instance, variable, etc. */
    OBJECT(Name.YES, Scope.NO),
    /** Statement, expression, etc. */
    ELEMENT(Name.NO, Scope.NO),
    /** Block, loop body, etc. */
    BLOCK(Name.NO, Scope.YES),
    /** Named block. */
    NAMED_BLOCK(Name.YES, Scope.YES),
    /** Statement, expression, etc. */
    PSEUDO_ELEMENT(Name.NO, Scope.TRANSPARENT),
    /** Unnamed pseudo-block. */
    PSEUDO_BLOCK(Name.NO, Scope.TRANSPARENT),
    /** Named pseudo-block. */
    NAMED_PSEUDO_BLOCK(Name.YES, Scope.TRANSPARENT);

    /** The name existence indicator. */
    private Name hasName;
    /** The scope existence indicator. */
    private Scope hasScope;

    /**
     * Creates a node kind.
     *
     * @param hasName the name existence indicator.
     * @param hasScope the scope existence indicator.
     */
    NodeKind(final Name hasName, final Scope hasScope) {
      InvariantChecks.checkNotNull(hasName);
      InvariantChecks.checkNotNull(hasScope);

      this.hasName = hasName;
      this.hasScope = hasScope;
    }

    /**
     * Returns the name existence indicator.
     *
     * @return the name existence indicator.
     */
    public final Name hasName() {
      return hasName;
    }

    /**
     * Returns the scope existence indicator.
     *
     * @return the scope existence indicator.
     */
    public final Scope hasScope() {
      return hasScope;
    }
  }

  /** The name of the node. */
  private String name;

  /** The tag of the node. */
  private Tag tag;

  /** The name existence indicator. */
  private Name hasName;

  /** The scope existence indicator. */
  private Scope hasScope;

  /** The admissible tags of the node's children. */
  private EnumSet<Tag> childrenTags;

  /** Maps a name to a node (i.e. an entity descriptor). */
  private Map<String, AbstractNode<Tag>> table = new LinkedHashMap<>();

  /** Maps a tag to a linked list of all nodes with that tag. */
  private Map<Tag, List<AbstractNode<Tag>>> slices = new HashMap<>();

  /** Contains all children of the node. */
  private List<AbstractNode<Tag>> children = new LinkedList<>();

  /** The default redefinition handler. */
  private ErrorHandler<Tag> redefinitionHandler = new ErrorHandler<Tag>() {
    @Override
    public void handle(final AbstractNode<Tag> newNode, final AbstractNode<Tag> oldNode) {
      throw new IllegalStateException(
          String.format("The symbol table already contains the name '%s'", newNode.getName()));
    }
  };

  /**
   * Creates a node.
   *
   * @param tag the tag.
   * @param childrenTags the admissible tags of the node's children.
   * @param kind the kind.
   * @param parent the parent node.
   */
  protected AbstractNode(
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final NodeKind kind,
      final AbstractNode<Tag> parent) {
    InvariantChecks.checkNotNull(tag);
    InvariantChecks.checkNotNull(childrenTags);
    // The parent is allowed to be null.

    this.tag = tag;
    this.childrenTags = childrenTags;
    this.hasName = kind.hasName();
    this.hasScope = kind.hasScope();

    if (parent != null) {
      // NOTE: Should be after the symbol table initialization.
      setParent(parent);
    }
  }

  /**
   * Creates a node.
   *
   * @param tag the tag.
   * @param childrenTags the admissible tags of the node's children.
   * @param kind the kind.
   */
  protected AbstractNode(
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final NodeKind kind) {
    this(tag, childrenTags, kind, null);
  }

  /**
   * Creates a copy of the node.
   *
   * @param other the node to be copied.
   */
  protected AbstractNode(final AbstractNode<Tag> other) {
    super(other);

    this.name = other.name;
    this.tag = other.tag;
    this.childrenTags = other.childrenTags;
    this.hasName = other.hasName;
    this.hasScope = other.hasScope;

    // Deep copy of the symbol table.
    for (final AbstractNode<Tag> child : other.children) {
      final AbstractNode<Tag> childClone = child.clone();

      childClone.setParent(this);
      this.add(childClone);
    }
  }

  /**
   * Creates a copy of the node.
   *
   * @param other the node to be copied.
   * @param parent the parent node to be set.
   */
  protected AbstractNode(final AbstractNode<Tag> other, final AbstractNode<Tag> parent) {
    this(other);

    setParent(parent);
  }

  /**
   * Returns the name of the node.
   *
   * @return the name of the node.
   */
  public final String getName() {
    return name;
  }

  /**
   * Returns the fully qualified name of the node.
   *
   * @return the full name of the node.
   */
  public final String getFullName() {
    if (getParent() == null) {
      return name;
    }

    final String path = getParent().getFullName();

    if (path == null) {
      return name;
    }

    return name != null ? String.format("%s.%s", path, name) : path;
  }

  /**
   * Sets the name of the node.
   *
   * @param name the name to be set.
   */
  public final void setName(final String name) {
    // The name is allowed to be null.
    this.name = name;
  }

  /**
   * Returns the tag of the node.
   *
   * @return the tag of the node.
   */
  public final Tag getTag() {
    return tag;
  }

  /**
   * Sets the tag of the node.
   *
   * @param tag the tag to be set.
   */
  public final void setTag(final Tag tag) {
    InvariantChecks.checkNotNull(tag);
    this.tag = tag;
  }

  /**
   * Checks whether the node has a name.
   *
   * @return {@code true} iff the node has a name.
   */
  public final boolean hasName() {
    return hasName == Name.YES;
  }

  /**
   * Checks whether the node has a scope.
   *
   * @return {@code true} iff the node has a scope.
   */
  public final boolean hasScope() {
    return hasScope == Scope.YES;
  }

  /** Checks whether the node is transparent (i.e. has no a real scope).
   *
   * @return {@code true} iff the node is transparent.
   */
  public final boolean isTransparent() {
    return hasScope == Scope.TRANSPARENT;
  }

  /**
   * Returns the parent of the node.
   *
   * @return the parent node or {@code null} if the node is a top-level one.
   */
  public final AbstractNode<Tag> getParent() {
    return (AbstractNode<Tag>) getUpperTable();
  }

  /**
   * Sets the parent of the node.
   *
   * @param parent the parent node to be set.
   */
  public final void setParent(final AbstractNode<Tag> parent) {
    InvariantChecks.checkNotNull(parent);
    setUpperTable(parent);
  }

  @Override
  public String toString() {
    return String.format("%s(%s)", tag, (name != null ? name : ""));
  }

  @Override
  public abstract AbstractNode<Tag> clone();

  //================================================================================================
  // Symbol Table Methods
  //================================================================================================

  @Override
  public final Collection<AbstractNode<Tag>> items() {
    return children;
  }

  @Override
  public final Collection<AbstractNode<Tag>> items(final Tag tag) {
    InvariantChecks.checkNotNull(tag);

    final Collection<AbstractNode<Tag>> slice = slices.get(tag);

    if (slice == null) {
      return Collections.<AbstractNode<Tag>>emptyList();
    }

    return slice;
  }

  @Override
  public final Collection<AbstractNode<Tag>> items(final EnumSet<Tag> tags) {
    InvariantChecks.checkNotNull(tags);

    final Collection<AbstractNode<Tag>> selection = new ArrayList<>();

    for (final Tag tag : tags) {
      selection.addAll(items(tag));
    }

    return selection;
  }

  @Override
  public final AbstractNode<Tag> find(final String name) {
    InvariantChecks.checkNotNull(name);

    final AbstractNode<Tag> node = table.get(name);

    if (node != null) {
      return node;
    }

    // NOTE: Sometimes, there are references to declarations inside the declarations.
    if (this.name != null && this.name.equals(name)) {
      return this;
    }

    return null;
  }

  @Override
  public final void add(final AbstractNode<Tag> node) {
    InvariantChecks.checkNotNull(node);

    // Check admissibility of the child's tag.
    InvariantChecks.checkTrue(childrenTags.contains(node.getTag()),
        String.format("The child of '%s' has the illegal tag '%s'", getTag(), node.getTag()));

    // Set the child's parent.
    final AbstractNode<Tag> parent = node.getParent();
    InvariantChecks.checkTrue(parent == null || parent == this || parent == getParent());

    if (parent == null) {
      node.setParent(this);
    }

    // NOTE: Even unnamed children are put into the symbol table.
    final String name = node.getName();
    final Tag tag = node.getTag();

    // If the node has a name, add the node into the symbol table.
    if (node.hasName() && name != null) {
      if (table.containsKey(name)) {
        // Invoke the redefinition callback.
        redefinitionHandler.handle(node, table.get(name));
      }

      table.put(name, node);
    }

    // Add the node into the slice.
    List<AbstractNode<Tag>> slice = slices.get(tag);

    if (slice == null) {
      slice = new LinkedList<>();
    }

    slice.add(node);
    slices.put(tag, slice);

    // Add the node into the list of children.
    children.add(node);
  }

  @Override
  public final void remove(final AbstractNode<Tag> node) {
    InvariantChecks.checkNotNull(node);

    final String name = node.getName();
    final Tag tag = node.getTag();

    // If the node has a name, remove the node from the symbol table.
    if (node.hasName() && name != null) {
      table.remove(name);
    }

    // Remove the node from the slice.
    final List<AbstractNode<Tag>> slice = slices.get(tag);
    slice.remove(node);

    // Remove the node from the list of children.
    children.remove(node);
  }

  @Override
  public void replace(final AbstractNode<Tag> oldNode,
      final Collection<AbstractNode<Tag>> newNodes) {

    final String oldName = oldNode.getName();
    final Tag oldTag = oldNode.getTag();

    // Replace the node in the name table.
    table.remove(oldName);

    for (final AbstractNode<Tag> newNode : newNodes) {
      // Set the child's parent.
      final AbstractNode<Tag> parent = newNode.getParent();
      InvariantChecks.checkTrue(parent == null || parent == oldNode.getParent());

      if (parent == null) {
        newNode.setParent(oldNode.getParent());
      }

      // NOTE: Even unnamed children are put into the symbol table.
      final String newName = newNode.getName();
      final Tag newTag = newNode.getTag();
      InvariantChecks.checkTrue(newTag == oldTag);

      // If the node has a name, add the node into the symbol table.
      if (newNode.hasName() && newName != null) {
        if (table.containsKey(newName)) {
          // Invoke the redefinition callback.
          redefinitionHandler.handle(newNode, table.get(newName));
        }

        table.put(newName, newNode);
      }
    }

    // Replace the node in the slice.
    final List<AbstractNode<Tag>> slice = slices.get(tag);
    final int i = slice.indexOf(oldNode);
    InvariantChecks.checkTrue(i != -1);

    slice.remove(i);
    slice.addAll(i, newNodes);

    // Replace the node in the list of children.
    final int j = children.indexOf(oldNode);
    InvariantChecks.checkTrue(j != -1);

    children.remove(j);
    children.addAll(j, newNodes);
  }

  //================================================================================================
  // Error Handler
  //================================================================================================

  /**
   * Sets the redefinition handler.
   *
   * @param redefinitionHandler the handler to be set.
   */
  protected final void setRedefinitionHandler(final ErrorHandler<Tag> redefinitionHandler) {
    InvariantChecks.checkNotNull(redefinitionHandler);
    this.redefinitionHandler = redefinitionHandler;
  }
}
