/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.core;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.EnumSet;
import java.util.LinkedList;
import java.util.Queue;

/**
 * {@link AbstractSymbolTable} represents an abstract symbol table.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class AbstractSymbolTable<Tag extends Enum<Tag>> {
  /** The symbol table of the upper scope. */
  private AbstractSymbolTable<Tag> upper;

  /**
   * Creates a symbol table.
   */
  public AbstractSymbolTable() {}

  /**
   * Creates a symbol table with an upper scope.
   * 
   * @param upper the symbol table of the upper scope.
   */
  public AbstractSymbolTable(final AbstractSymbolTable<Tag> upper) {
    InvariantChecks.checkNotNull(upper);
    this.upper = upper;
  }

  /**
   * Returns the symbol table of the upper scope.
   * 
   * @return the upper symbol table or {@code null}.
   */
  protected final AbstractSymbolTable<Tag> getUpperTable() {
    return upper;
  }

  /**
   * Sets the symbol table of the upper scope.
   * 
   * @param upper the symbol table to be set.
   */
  protected final void setUpperTable(final AbstractSymbolTable<Tag> upper) {
    InvariantChecks.checkNotNull(upper);
    this.upper = upper;
  }

  /**
   * Returns the symbol table's items.
   * 
   * @return the collection of items.
   */
  public abstract Collection<AbstractNode<Tag>> items();

  /**
   * Returns the items whose tags belong to the given set.
   * 
   * @param tags the set of tags.
   * @return the collection of items.
   */
  public abstract Collection<AbstractNode<Tag>> items(final EnumSet<Tag> tags);

  /**
   * Returns the items with the given tag.
   * 
   * @param tag the tag.
   * @return the collection of items.
   */
  public abstract Collection<AbstractNode<Tag>> items(final Tag tag);

  /**
   * Returns the items with the given tags.
   * 
   * @param tag1 the tag 1.
   * @param tag2 the tag 2.
   * @return the collection of items.
   */
  public final Collection<AbstractNode<Tag>> items(final Tag tag1, final Tag tag2) {
    return items(EnumSet.of(tag1, tag2));
  }

  /**
   * Returns the items with the given tags.
   * 
   * @param tag1 the tag 1.
   * @param tag2 the tag 2.
   * @param tag3 the tag 3.
   * @return the collection of items.
   */
  public final Collection<AbstractNode<Tag>> items(final Tag tag1, final Tag tag2, final Tag tag3) {
    return items(EnumSet.of(tag1, tag2, tag3));
  }

  /**
   * Returns an item of the symbol table with the given name.
   * 
   * @param name the name.
   * @return an item or {@code null}.
   */
  public abstract AbstractNode<Tag> find(final String name);

  /**
   * Returns an item of the symbol table with the given name and the given tag.
   * 
   * <p>
   * If the tag is null, then it is ignored.
   * </p>
   * 
   * @param name the name.
   * @param tag the tag.
   * @return an item or {@code null}.
   */
  public final AbstractNode<Tag> findRecursively(final String name, final Tag tag) {
    for (AbstractSymbolTable<Tag> table = this; table != null; table = table.getUpperTable()) {
      final AbstractNode<Tag> child = table.find(name);

      if (child != null && (tag == null || child.getTag() == tag)) {
        return child;
      }
    }

    return null;
  }

  /**
   * Returns an item of the symbol table with the given name.
   * 
   * @param name the name.
   * @return an item or {@code null}.
   */
  public final AbstractNode<Tag> findRecursively(final String name) {
    return findRecursively(name, null);
  }

  /**
   * Returns an item of the symbol table with the given name and the given tag.
   * 
   * <p>
   * If the tag is null, then it is ignored.
   * </p>
   * 
   * <p>
   * The purpose of this method is to handle the following situation, where {@code scope} declares
   * some element, which is visible from outside.
   * </p>
   * 
   * <pre>
   * scope {
   *   declare(element);
   * }
   * use(element).
   * </pre>
   * 
   * <p>
   * Such situation is possible, when {@code scope} is transparent (i.e. it is not a real scope and
   * will be removed by a preprocessor).
   * </p>
   * 
   * @param name the name.
   * @param tag the tag.
   * @return an item or {@code null}.
   */
  public final AbstractNode<Tag> findAroundRecursively(final String name, final Tag tag) {
    for (AbstractSymbolTable<Tag> table = this; table != null; table = table.getUpperTable()) {
      final Queue<AbstractSymbolTable<Tag>> queue = new LinkedList<>();
      queue.add(table);

      while (!queue.isEmpty()) {
        final AbstractSymbolTable<Tag> scope = queue.remove();
        final AbstractNode<Tag> node = scope.find(name);

        if (node != null && (tag == null || node.getTag() == tag)) {
          return node;
        }

        final Collection<AbstractNode<Tag>> children = scope.items();
        for (final AbstractNode<Tag> child : children) {
          if (child.isTransparent()) {
            queue.add(child);
          }
        }
      }
    }

    return null;
  }

  /**
   * Returns an item of the symbol table with the given name.
   * 
   * @param name the name.
   * @return an item or {@code null}.
   */
  public final AbstractNode<Tag> findAroundRecursively(final String name) {
    return findAroundRecursively(name, null);
  }

  /**
   * Adds the item into the symbol table.
   * 
   * @param item the item to be added.
   */
  public abstract void add(final AbstractNode<Tag> item);

  /**
   * Adds the items into the symbol table.
   * 
   * @param items the collection of items to be added.
   */
  public final void addAll(final Collection<? extends AbstractNode<Tag>> items) {
    InvariantChecks.checkNotNull(items);

    for (final AbstractNode<Tag> item : items) {
      add(item);
    }
  }

  /**
   * Removes the item from the symbol table.
   * 
   * @param item the item to be removed.
   */
  public abstract void remove(final AbstractNode<Tag> item);

  /**
   * Replaces the item of the symbol table with the given items.
   * 
   * <p>All items should have the same tag.</p>
   * 
   * @param oldItem the existing item to be replaced.
   * @param newItems the items to to replace the existing one.
   */
  public abstract void replace(
      final AbstractNode<Tag> oldItem, final Collection<AbstractNode<Tag>> newItems);

  /**
   * Clones the symbol table.
   * 
   * @return a copy of the symbol table.
   */
  public abstract AbstractSymbolTable<Tag> clone();
}
