/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogNode;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * {@link VerilogContext} keeps a current path.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogContext {

  private static final class Scope {
    /** Instantiation node. */
    final VerilogNode node;
    /** Instance bindings. */
    final Map<String, VerilogBinding> bindings = new LinkedHashMap<>();
    /** Values of the generate variables. */
    final Map<String, Data> genvars = new LinkedHashMap<>();

    Scope(final VerilogNode node) {
      // The argument can be null.
      this.node = node;
    }

    Scope(final Scope other) {
      InvariantChecks.checkNotNull(other);

      this.node = other.node;
      this.bindings.putAll(other.bindings);
      this.genvars.putAll(other.genvars);
    }

    @Override
    public Scope clone() {
      return new Scope(this);
    }
  }

  /** Path to the current instance. */
  private final List<Scope> path;

  public VerilogContext() {
    this(new ArrayList<>());
  }

  public VerilogContext(final VerilogContext other) {
    this(new ArrayList<>());

    for (final Scope scope : other.path) {
      this.path.add(scope.clone());
    }
  }

  private VerilogContext(final List<Scope> path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  public void begin(final VerilogNode node, boolean reuseParentBindings) {
    // The node parameter can be null.
    final Scope scope = new Scope(node);

    if (reuseParentBindings && !path.isEmpty()) {
      scope.bindings.putAll(getBindings());
    }

    path.add(scope);
  }

  public void begin(final VerilogNode node) {
    begin(node, false);
  }

  public void end() {
    InvariantChecks.checkTrue(!path.isEmpty());
    path.remove(path.size() - 1);
  }

  private String getPathName() {
    final StringBuilder builder = new StringBuilder();

    for (final Scope scope : path) {
      if (builder.length() != 0) {
        builder.append('.');
      }

      final VerilogNode node = scope.node;
      if (node != null && node.getName().length() > 0) {
        builder.append(node.getName());
      }
    }

    return builder.toString();
  }

  public String getFullName(final String localName) {
    InvariantChecks.checkNotNull(localName);

    final String pathName = getPathName();

    // Temporary fix: check name prefix to avoid erroneous repeated actions.
    return pathName.length() != 0 && !localName.startsWith(pathName)
        ? String.format("%s.%s", getPathName(), localName) : localName;
  }

  public NodeVariable getVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    final String name = getFullName(variable.getName());
    final DataType type = variable.getDataType();

    return new NodeVariable(name, type);
  }

  public VerilogContext getParentContext() {
    return new VerilogContext(this.path.subList(0, path.size() - 1));
  }

  public VerilogNode getNode() {
    InvariantChecks.checkNotEmpty(path);
    return path.get(path.size() - 1).node;
  }

  public Map<String, VerilogBinding> getBindings() {
    InvariantChecks.checkNotEmpty(path);
    return path.get(path.size() - 1).bindings;
  }

  public Map<String, Data> getGenvars() {
    InvariantChecks.checkNotEmpty(path);
    return path.get(path.size() - 1).genvars;
  }

  public Map<String, VerilogBinding> getParentBindings() {
    InvariantChecks.checkNotEmpty(path);
    return path.size() > 1 ? path.get(path.size() - 2).bindings : null;
  }

  public void addGenvar(final String name, final Data data) {
    InvariantChecks.checkNotEmpty(path);
    path.get(path.size() - 1).genvars.put(name, data);
  }

  public Data getGenvarValue(final String name) {
    for (int i = path.size() - 1; i >= 0; i--) {
      final Scope scope = path.get(i);
      final Data value = scope.genvars.get(name);

      if (value != null) {
        return value;
      }
    }

    return null;
  }

  @Override
  public String toString() {
    return getPathName();
  }

  @Override
  public VerilogContext clone() {
    return new VerilogContext(this);
  }
}
