/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.EnumSet;

/**
 * {@link VerilogActivity} represents processes.
 * 
 * <p>There are the following process types:</p>
 * <ul>
 * <li>{@code always} (ordinary process);</li>
 * <li>{@code initial} (initial process).</li>
 * </ul>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogActivity extends VerilogNode {
  public static final Tag TAG = Tag.ACTIVITY;
  public static final EnumSet<Tag> TAGS_CHILDREN = VerilogStatement.TAGS;

  /** {@link Type} contains the process types. */
  public enum Type {
    /** Ordinary process: {@code always ...}. */
    ALWAYS,
    /** Initial process: {@code initial ...}. */
    INITIAL
  }

  /** Type of the process (by default, {@code ALWAYS}). */
  private Type type = Type.ALWAYS;

  /**
   * Creates a process.
   * 
   * @param parent the node parent.
   */
  public VerilogActivity(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the process.
   * 
   * @param other the process to be copied.
   * @param parent the parent node.
   */
  public VerilogActivity(final VerilogActivity other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
  }
  
  /**
   * Returns the type of the process.
   * 
   * @return the process type: {@code ALWAYS} or {@code INITIAL}.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the process.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the process type is {@code ALWAYS}.
   * 
   * @return {@code true} iff the process type is {@code ALWAYS}.
   */
  public boolean isAlways() {
    return type == Type.ALWAYS;
  }

  /**
   * Checks whether the process type is {@code INITIAL}.
   * 
   * @return {@code true} iff the process type is {@code INITIAL}.
   */
  public boolean isInitial() {
    return type == Type.INITIAL;
  }

  /**
   * Sets the process type to {@code ALWAYS}.
   */
  public void setAlways() {
    type = Type.ALWAYS;
  }

  /**
   * Sets the process type to {@code INITIAL}.
   */
  public void setInitial() {
    type = Type.INITIAL;
  }

  /**
   * Returns the statement of the process.
   * 
   * @return the process statement.
   */
  public VerilogStatement getStatement() {
    return (VerilogStatement) items(VerilogStatement.TAGS).iterator().next();
  }

  /**
   * Sets the statement to the process.
   * 
   * @param statement the statement to be set.
   */
  public void setStatement(final VerilogStatement statement) {
    InvariantChecks.checkNotNull(statement);
    add(statement);
  }

  @Override
  public VerilogActivity clone() {
    return new VerilogActivity(this, getParentNode());
  }
}
