/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogAssignment} represents assignments.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogAssignment extends VerilogNode {
  public static final Tag TAG = Tag.ASSIGNMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** L-value (reference). */
  private List<VerilogReference> references = new ArrayList<>();
  /** R-value (expression). */
  private VerilogMinTypMax expression;

  /**
   * Creates an assignment.
   * 
   * @param parent the parent node.
   */
  public VerilogAssignment(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the assignment.
   * 
   * @param other the assignment to be copied.
   * @param parent the parent node.
   */
  public VerilogAssignment(final VerilogAssignment other, final VerilogNode parent) {
    super(other, parent);

    for (final VerilogReference reference : other.references) {
      this.references.add(reference.clone());
    }

    // The right-hand-side expression is allowed to be null (e.g., in deassign).
    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the l-value (reference) of the assignment.
   * 
   * @return the reference.
   */
  public List<VerilogReference> getReferences() {
    return references;
  }

  /**
   * Adds the reference to the l-value.
   * 
   * @param reference the reference to be added.
   */
  public void addReference(final VerilogReference reference) {
    InvariantChecks.checkNotNull(reference);
    this.references.add(reference);
  }

  /**
   * Returns the expression representation of the assignment's left hand side.
   * 
   * @return the left hand side.
   */
  public Node getLhsNode() {
    final Collection<Node> operands = new ArrayList<>(references.size());

    for (final VerilogReference reference : references) {
      operands.add(reference.getNode());
    }

    return operands.size() == 1 ? ((ArrayList<Node>) operands).get(0) : Nodes.bvconcat(operands);
  }

  /**
   * Returns the r-value (expression) of the assignment.
   * 
   * @return the expression.
   */
  public VerilogMinTypMax getRhsExpression() {
    return expression;
  }

  /**
   * Sets the r-value (expression) of the assignment.
   * 
   * @param expression the expression to be set.
   */
  public void setRhsExpression(final VerilogMinTypMax expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  @Override
  public VerilogAssignment clone() {
    return new VerilogAssignment(this, getParentNode());
  }
}
