/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogLiteral;
import ru.ispras.verilog.parser.model.util.ModelUtils;

import java.util.EnumSet;
import java.util.LinkedList;
import java.util.List;

/**
 * {@link VerilogCaseStatementItem} represents case statement items.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogCaseStatementItem extends VerilogNode {
  public static final Tag TAG = Tag.CASE_STATEMENT_ITEM;
  public static final EnumSet<Tag> TAGS_CHILDREN = VerilogStatement.TAGS;

  /** Case expressions (to be matched with the case selection expression). */
  private List<VerilogExpression> expressions = new LinkedList<>();

  /**
   * Creates a case item.
   *
   * @param parent the parent node.
   */
  public VerilogCaseStatementItem(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the case item.
   *
   * @param other the case item to be copied.
   * @param parent the parent node.
   */
  public VerilogCaseStatementItem(final VerilogCaseStatementItem other, final VerilogNode parent) {
    super(other, parent);

    for (final VerilogExpression expression : other.expressions) {
      this.expressions.add(expression.clone());
    }
  }

  /**
   * Returns the expressions of the case.
   *
   * @return the list of expressions.
   */
  public List<VerilogExpression> getExpressions() {
    return expressions;
  }

  /**
   * Adds the expression to the case.
   *
   * @param expression the expression to be added.
   */
  public void addExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    expressions.add(expression);
  }

  /**
   * Adds all the expressions to the case.
   * @param list The list of expressions.
   */
  public void addExpressions(final List<VerilogExpression> list) {
    for (final VerilogExpression expr : list) {
      addExpression(expr);
    }
  }

  /**
   * Removes all the expressions of the case.
   */
  public void removeExpressions() {
    expressions.clear();
  }

  /**
   * Returns the statement of the case.
   *
   * @return the statement.
   */
  public VerilogStatement getStatement() {
    return (VerilogStatement) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the statement of the case.
   *
   * @param statement the statement to be set.
   */
  public void setStatement(final VerilogStatement statement) {
    add(statement);
  }

  /**
   * Returns the condition associated with the case.
   *
   * @return the case condition.
   */
  public VerilogExpression getExpression() {
    // Create a result.
    final VerilogExpression expression = new VerilogExpression();
    // Get the parent node.
    final VerilogCaseStatement parent = (VerilogCaseStatement)getParent();

    final VerilogExpression a = parent.getExpression();

    // A non-default case.
    for (final VerilogExpression b : expressions) {
      Node lhs = a.getNode();
      Node rhs = b.getNode();

      if (rhs.getKind() == Kind.VALUE) {
        VerilogLiteral mask = null;

        final NodeValue value = (NodeValue) rhs;
        final VerilogLiteral literal = ModelUtils.extractLiteral(value.getData());

        // Choose mask according to the case statement type.
        switch (parent.getType()) {
          case CASE:
            // No mask.
            mask = new VerilogLiteral(0, literal.getBitSize());
            break;
          case CASEX:
            // X mask.
            mask = (literal.getXMask() != null)
                ? new VerilogLiteral(literal.getXMask())
                : new VerilogLiteral(0, literal.getBitSize());
            break;
          case CASEZ:
            // Z mask.
            mask = (literal.getZMask() != null)
                ? new VerilogLiteral(literal.getZMask())
                : new VerilogLiteral(0, literal.getBitSize());
            break;
          default:
            break;
        }

        // For casex and casez, ((a|mask) === (b|mask)) is used instead of (a === b).
        if (!mask.getBitVector().isAllReset()) {
          lhs = new NodeOperation(StandardOperation.BVOR, lhs, ModelUtils.getValue(mask));
          rhs = new NodeOperation(StandardOperation.BVOR, rhs, ModelUtils.getValue(mask));
        }
      }

      // EQCASE (===) compares undefined values: X === X is true.
      final NodeOperation equality = new NodeOperation(StandardOperation.EQCASE, lhs, rhs);

      if (expression.isNull()) {
        // case(a) b ... => (a === b).
        expression.setNode(equality);
      } else {
        // case(a) b, c ... => (a === b) || (a === c).
        expression.setNode(
            new NodeOperation(StandardOperation.OR, expression.getNode(), equality));
      }
    }

    // A default case.
    if (expression.isNull()) {
      // Iterate over all cases.
      for (final VerilogCaseStatementItem item : parent.getCases()) {
        // Except the default one.
        if (item == this) {
          continue;
        }

        // Construct the disjunction of all non-default conditions.
        final VerilogExpression b = item.getExpression();

        if (expression.isNull()) {
          // case(a) b ... => non-default = (a === b).
          expression.setNode(b.getNode());
        } else {
          // case(a) b, c ... => non-default = ((a === b) || (a === c)).
          expression.setNode(
              new NodeOperation(StandardOperation.OR, expression.getNode(), b.getNode()));
        }
      }

      // Negate the non-default condition.
      if (expression.isNull()) {
        expression.setNode(NodeValue.newBoolean(true));
      } else {
        expression.setNode(
            new NodeOperation(StandardOperation.NOT, expression.getNode()));
      }
    }

    return expression;
  }

  @Override
  public VerilogCaseStatementItem clone() {
    return new VerilogCaseStatementItem(this, getParentNode());
  }
}
