/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogRange;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogDeclaration} represents declarations.
 * 
 * <p>There are the following types of declarations:</p>
 * <ul>
 * <li>{@code input};</li>
 * <li>{@code output};</li>
 * <li>{@code inout};</li>
 * <li>{@code event};</li>
 * <li>{@code parameter};</li>
 * <li>{@code specparam};</li>
 * <li>{@code localparam};</li>
 * <li>{@code genvar};</li>
 * <li>{@code variable}.</li>
 * </ul>
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogDeclaration extends VerilogNode {
  public static final Tag TAG = Tag.DECLARATION;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** {@link Type} contains the declaration types. */
  public enum Type {
    /** Input declaration: {@code input x}. */
    INPUT,
    /** Output declaration: {@code output x}. */
    OUTPUT,
    /** Input/output declaration: {@code inout x}. */
    INOUT,
    /** Event declaration: {@code event x}. */
    EVENT,
    /** Parameter declaration: {@code parameter x}. */
    PARAMETER,
    /** Specify parameter declaration: {@code specparam x}. */
    SPECPARAM,
    /** Local parameter declaration: {@code localparam x}. */
    LOCALPARAM,
    /** Generate variable declaration: {@code genvar x}. */
    GENVAR,
    /** Local variable declaration: {@code x}. */
    VARIABLE
  }

  /** Type of the declaration (by default, {@code VARIABLE}). */
  private Type type = Type.VARIABLE;
  /** Type of the declared element. */
  private VerilogElementType elementType;
  /** Dimension of the declared element. */
  private List<VerilogRange> dimensions = new ArrayList<>();
  /** Expression representing the initial value. */
  private VerilogMinTypMax expression;

  /**
   * Creates a declaration.
   *
   * @param parent the parent node.
   */
  public VerilogDeclaration(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
  }

  /**
   * Creates a copy of the declaration.
   * 
   * @param other the declaration to be copied.
   * @param parent the parent node.
   */
  public VerilogDeclaration(final VerilogDeclaration other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
    this.elementType = other.elementType != null ? other.elementType.clone() : null;

    for (final VerilogRange range : other.dimensions) {
      this.dimensions.add(range.clone());
    }

    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the declaration type.
   *
   * @return the declaration type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the declaration type.
   *
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the declaration is {@code INPUT}.
   *
   * @return {@code true} iff the declaration is {@code INPUT}.
   */
  public boolean isInput() {
    return type == Type.INPUT;
  }

  /**
   * Checks whether the declaration is {@code OUTPUT}.
   *
   * @return {@code true} iff the declaration is {@code OUTPUT}.
   */
  public boolean isOutput() {
    return type == Type.OUTPUT;
  }

  /**
   * Checks whether the declaration is {@code INOUT}.
   *
   * @return {@code true} iff the declaration is {@code INOUT}.
   */
  public boolean isInout() {
    return type == Type.INOUT;
  }

  /**
   * Checks whether the declaration is a port ({@code INPUT}, {@code OUTPUT}, or {@code INOUT}).
   *
   * @return {@code true} iff the declaration is a port.
   */
  public boolean isPort() {
    return isInput() || isOutput() || isInout();
  }

  /**
   * Checks whether the declaration is {@code EVENT}.
   *
   * @return {@code true} iff the declaration is {@code EVENT}.
   */
  public boolean isEvent() {
    return type == Type.EVENT;
  }

  /**
   * Checks whether the declaration is {@code PARAMETER}.
   *
   * @return {@code true} iff the declaration is {@code PARAMETER}.
   */
  public boolean isParameter() {
    return type == Type.PARAMETER;
  }

  /**
   * Checks whether the declaration is {@code SPECPARAMETER}.
   *
   * @return {@code true} iff the declaration is {@code SPECPARAMETER}.
   */
  public boolean isSpecparam() {
    return type == Type.SPECPARAM;
  }

  /**
   * Checks whether the declaration is {@code LOCALPARAM}.
   *
   * @return {@code true} iff the declaration is {@code LOCALPARAM}.
   */
  public boolean isLocalparam() {
    return type == Type.LOCALPARAM;
  }

  /**
   * Checks whether the declaration is a parameter
   * ({@code PARAMETER}, {@code SPECPARAM}, or {@code LOCALPARAM}).
   *
   * @return {@code true} iff the declaration is a port.
   */
  public boolean isParam() {
    return isParameter() || isSpecparam() || isLocalparam();
  }

  /**
   * Checks whether the declaration is {@code GENVAR}.
   *
   * @return {@code true} iff the declaration is {@code GENVAR}.
   */
  public boolean isGenvar() {
    return type == Type.GENVAR;
  }

  /**
   * Checks whether the declaration is {@code VARIABLE}.
   *
   * @return {@code true} iff the declaration is {@code VARIABLE}.
   */
  public boolean isVariable() {
    return type == Type.VARIABLE;
  }

  /**
   * Sets the type of the statement to {@code INPUT}.
   */
  public void setInput() {
    type = Type.INPUT;
  }

  /**
   * Sets the type of the statement to {@code OUTPUT}.
   */
  public void setOutput() {
    type = Type.OUTPUT;
  }

  /**
   * Sets the type of the statement to {@code INOUT}.
   */
  public void setInout() {
    type = Type.INOUT;
  }

  /**
   * Sets the type of the statement to {@code EVENT}.
   */
  public void setEvent() {
    type = Type.EVENT;
  }

  /**
   * Sets the type of the statement to {@code PARAMETER}.
   */
  public void setParameter() {
    type = Type.PARAMETER;
  }

  /**
   * Sets the type of the statement to {@code SPECPARAM}.
   */
  public void setSpecparam() {
    type = Type.SPECPARAM;
  }

  /**
   *  Sets the type of the statement to {@code LOCALPARAM}.
   */
  public void setLocalparam() {
    type = Type.LOCALPARAM;
  }

  /**
   * Sets the type of the statement to {@code GENVAR}.
   */
  public void setGenvar() {
    type = Type.GENVAR;
  }

  /**
   * Sets the type of the statement to {@code VARIABLE}.
   */
  public void setVariable() {
    type = Type.VARIABLE;
  }

  /**
   * Returns the type of the declared element.
   *
   * @return the element type.
   */
  public VerilogElementType getElementType() {
    return elementType;
  }

  /**
   * Sets the type of the declared element.
   *
   * @param elementType the type to be set.
   */
  public void setElementType(final VerilogElementType elementType) {
    InvariantChecks.checkNotNull(elementType);
    this.elementType = elementType;
  }

  /**
   * Returns the dimensions of the declared element.
   *
   * @return the element dimensions.
   */
  public List<VerilogRange> getDimensions() {
    return dimensions;
  }

  /**
   * Extends the dimension of the declared element.
   *
   * @param range the new dimension.
   */
  public void addDimension(final VerilogRange range) {
    InvariantChecks.checkNotNull(range);
    this.dimensions.add(range);
  }

  /**
   * Returns the expression representing the initial value.
   *
   * @return the initial value expression.
   */
  public VerilogMinTypMax getExpression() {
    return expression;
  }

  /**
   * Sets the expression representing the initial value.
   *
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogMinTypMax expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Merges this declaration with the given one.
   * 
   * <p>
   * [IEEE Std 1364-2005]
   * 
   * 12.3.3 Port declarations
   *
   * If a port declaration does not include a net or variable type, then the port can be again
   * declared in a net or variable declaration. If the net or variable is declared as a vector,
   * the range specification between the two declarations of a port shall be identical. Once a name
   * is used in a port declaration, it shall not be declared again in another port declaration or
   * in a data type declaration.
   * </p>
   * 
   * @param declaration the declaration to be merged with this one.
   */
  public void addDeclaration(final VerilogDeclaration declaration) {
    InvariantChecks.checkNotNull(declaration);

    if (this.type == Type.VARIABLE && declaration.type != Type.VARIABLE) {
      this.type = declaration.type;
    }
    if (!this.elementType.isSpecified() && declaration.elementType.isSpecified()) {
      this.elementType = declaration.elementType;
    }
    if (this.dimensions.isEmpty() && !declaration.dimensions.isEmpty()) {
      this.dimensions = declaration.dimensions;
    }
    if (this.expression == null && declaration.expression != null) {
      this.expression = declaration.expression;
    }
  }

  @Override
  public VerilogDeclaration clone() {
    return new VerilogDeclaration(this, getParentNode());
  }
}
