/*
 * Copyright 2012-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.EnumSet;

/**
 * {@link VerilogLoopStatement} represents loop statements.
 *
 * <p>There are the following loop types:</p>
 * <ul>
 * <li>{@code forever};</li>
 * <li>{@code repeat};</li>
 * <li>{@code while};</li>
 * <li>{@code for}.</li>
 * </ul>
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogLoopStatement extends VerilogStatement {
  public static final Tag TAG = Tag.LOOP_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = VerilogStatement.TAGS;

  /** {@link Type} contains the loop statement types. */
  public enum Type {
    /** Forever loop: {@code forever ...}. */
    FOREVER,
    /** Forever loop: {@code repeat(n) ...}. */
    REPEAT,
    /** While loop: {@code while(c) ...}. */
    WHILE,
    /** For loop: {@code for(a; b; c) ...}. */
    FOR
  }

  /** Loop statement type (by default, {@code FOREVER}). */
  private Type type = Type.FOREVER;
  /** Loop initialization assignment (only for {@code FOR} loops). */
  private VerilogAssignment initialization;
  /** Loop expression (condition or iteration number). */
  private VerilogExpression expression;
  /** Loop iteration assignment (only for {@code FOR} loops). */
  private VerilogAssignment iteration;

  /**
   * Creates a loop statement.
   *
   * @param parent the node parent.
   */
  public VerilogLoopStatement(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the loop statement.
   *
   * @param other the loop statement to be copied.
   * @param parent the parent node.
   */
  public VerilogLoopStatement(final VerilogLoopStatement other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
    this.initialization = other.initialization != null ? other.initialization.clone() : null;
    this.expression = other.expression != null ? other.expression.clone() : null;
    this.iteration = other.iteration != null ? other.iteration.clone() : null;
  }

  /**
   * Returns the type of the loop.
   *
   * @return the loop type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the loop.
   *
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the loop is {@code FOREVER}.
   *
   * @return {@code true} iff the loop is {@code FOREVER}.
   */
  public boolean isForever() {
    return type == Type.FOREVER;
  }

  /**
   * Checks whether the loop is {@code REPEAT}.
   *
   * @return {@code true} iff the loop is {@code REPEAT}.
   */
  public boolean isRepeat() {
    return type == Type.REPEAT;
  }

  /**
   * Checks whether the loop is {@code WHILE}.
   *
   * @return {@code true} iff the loop is {@code WHILE}.
   */
  public boolean isWhile() {
    return type == Type.WHILE;
  }

  /**
   * Checks whether the loop is {@code FOR}.
   *
   * @return {@code true} iff the loop is {@code FOR}.
   */
  public boolean isFor() {
    return type == Type.FOR;
  }

  /**
   * Sets the type of loop to {@code FOREVER}.
   */
  public void setForever() {
    type = Type.FOREVER;
  }

  /**
   * Sets the type of loop to {@code REPEAT}.
   */
  public void setRepeat() {
    type = Type.REPEAT;
  }

  /**
   * Sets the type of loop to {@code WHILE}.
   */
  public void setWhile() {
    type = Type.WHILE;
  }

  /**
   * Sets the type of loop to {@code FOR}.
   */
  public void setFor() {
    type = Type.FOR;
  }

  /**
   * Returns the initialization assignment of the {@code FOR} loop.
   *
   * @return the initialization assignment.
   */
  public VerilogAssignment getInitialization() {
    return initialization;
  }

  /**
   * Sets the initialization assignment of the {@code FOR} loop.
   *
   * @param initialization the assignment to be set.
   */
  public void setInitialization(final VerilogAssignment initialization) {
    InvariantChecks.checkNotNull(initialization);
    this.initialization = initialization;
  }

  /**
   * Returns the expression of the loop.
   *
   * @return the expression.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the expression of the loop.
   *
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the iteration assignment of the {@code FOR} loop.
   *
   * @return the iteration assignment.
   */
  public VerilogAssignment getIteration() {
    return iteration;
  }

  /**
   * Sets the iteration assignment of the {@code FOR} loop.
   *
   * @param iteration the assignment to be set.
   */
  public void setIteration(final VerilogAssignment iteration) {
    InvariantChecks.checkNotNull(iteration);
    this.iteration = iteration;
  }

  /**
   * Returns the loop body.
   *
   * @return the loop body.
   */
  public VerilogStatement getStatement() {
    return (VerilogStatement) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the loop body.
   *
   * @param statement the body to be set.
   */
  public void setStatement(final VerilogStatement statement) {
    add(statement);
  }

  @Override
  public VerilogLoopStatement clone() {
    return new VerilogLoopStatement(this, getParentNode());
  }
}
