/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.core.ErrorHandler;

import java.util.Collection;
import java.util.EnumSet;

/**
 * {@link VerilogNode} represents a basic AST node for the Verilog HDL.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class VerilogNode extends AbstractNode<VerilogNode.Tag> {

  /** {@link Tag} contains the node types. */
  public enum Tag {
    // Root node.
    CODE,
    // Top-level entities.
    MODULE,
    // Ports.
    PORT,
    // Declarations.
    DECLARATION,
    // Module items.
    ACTIVITY,
    ASSIGN,
    ASSIGNMENT,
    INSTANTIATION,
    DEFINE_PARAMETER,
    PORT_CONNECTION,
    PROCEDURE,
    TABLE,
    TABLE_ENTRY,
    // Statements.
    NULL_STATEMENT,
    ASSIGN_STATEMENT,
    TASK_STATEMENT,
    WAIT_STATEMENT,
    DISABLE_STATEMENT,
    TRIGGER_STATEMENT,
    DELAYED_STATEMENT,
    IF_STATEMENT,
    IF_STATEMENT_BRANCH,
    CASE_STATEMENT,
    CASE_STATEMENT_ITEM,
    LOOP_STATEMENT,
    BLOCK_STATEMENT,
    // Generate constructs.
    GENERATE,
    IF_GENERATE,
    IF_GENERATE_BRANCH,
    CASE_GENERATE,
    CASE_GENERATE_ITEM,
    LOOP_GENERATE,
    BLOCK_GENERATE,
    // Specify constructs.
    SPECIFY,
    PULSE_STYLE,
    SHOW_CANCELLED,
    PATH_DECLARATION,
    // Attributes.
    ATTRIBUTE,
    // SVA
    ASSERTION_STATEMENT,
    SEQUENCE_DECLARATION,
    PROPERTY_DECLARATION,
    SVA_PORT,
    SVA_INSTANCE
  }

  /**
   * Unites two sets of tags.
   * 
   * @param lhs the set of tags.
   * @param rhs the set of tags.
   * @return the union the given sets.
   */
  public static EnumSet<Tag> union(final EnumSet<Tag> lhs, final EnumSet<Tag> rhs) {
    InvariantChecks.checkNotNull(lhs);
    InvariantChecks.checkNotNull(rhs);

    final EnumSet<Tag> result = EnumSet.copyOf(lhs);
    result.addAll(rhs);

    return result;
  }

  /** List of attributes. */
  private Collection<VerilogAttribute> attributes;

  /**
   * Creates a node.
   * 
   * @param tag the tag.
   * @param childrenTags the tags of admissible children.
   * @param kind the kind.
   * @param parent the parent node.
   */
  protected VerilogNode(
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final NodeKind kind,
      final VerilogNode parent) {
    super(tag, childrenTags, kind, parent);

    setRedefinitionHandler(new ErrorHandler<VerilogNode.Tag>() {
      @Override
      public void handle(
          final AbstractNode<VerilogNode.Tag> newNode,
          final AbstractNode<VerilogNode.Tag> oldNode) {
        // Declarations of the same variable can be merged.
        if (newNode.getTag() == Tag.DECLARATION && oldNode.getTag() == Tag.DECLARATION) {
          final VerilogDeclaration newDeclaration = (VerilogDeclaration) newNode;
          final VerilogDeclaration oldDeclaration = (VerilogDeclaration) oldNode;

          // [IEEE Std 1364-2005]
          //
          // 12.3.3 Port declarations
          //
          // If a port declaration includes a net or variable type, then the port is considered
          // completely declared, and it is an error for the port to be declared again in a
          // variable or net data type declaration. Because of this, all other aspects of the port
          // shall be declared in such a port declaration, including the signed and range
          // definitions if needed.

          // FIXME:

          // If a port declaration does not include a net or variable type, then the port can be
          // again declared in a net or variable declaration. If the net or variable is declared as
          // a vector, the range specification between the two declarations of a port shall be
          // identical. Once a name is used in a port declaration, it shall not be declared again
          // in another port declaration or in a data type declaration.
          oldDeclaration.addDeclaration(newDeclaration);
        } else {
          VerilogLogger.debug(String.format("Redefinition of the name '%s'", newNode.getName()));
        }
      }
    });
  }

  /**
   * Creates a top-level node (i.e. a node with {@code null} parent).
   * 
   * @param tag the tag.
   * @param childrenTags the tags of admissible children.
   * @param kind the kind.
   */
  protected VerilogNode(
      final Tag tag,
      final EnumSet<Tag> childrenTags,
      final NodeKind kind) {
    this(tag, childrenTags, kind, null);
  }

  /**
   * Creates a copy of the node.
   * 
   * @param other the node to be copied.
   * @param parent the parent node to be set.
   */
  protected VerilogNode(final VerilogNode other, final VerilogNode parent) {
    super(other, parent);

    // There is no need to copy attributes.
    this.attributes = other.attributes;
  }

  /**
   * Checks whether the node is {@code CODE}.
   * 
   * @return {@code true} iff the node is {@code CODE}.
   */
  public boolean isCode() {
    return getTag() == Tag.CODE;
  }

  /**
   * Checks whether the node is {@code MODULE}.
   * 
   * @return {@code true} iff the node is {@code MODULE}.
   */
  public boolean isModule() {
    return getTag() == Tag.MODULE;
  }

  /**
   * Checks whether the node is {@code PORT}.
   * 
   * @return {@code true} iff the node is {@code PORT}.
   */
  public boolean isPort() {
    return getTag() == Tag.PORT;
  }

  /**
   * Checks whether the node is {@code SVA_PORT}.
   *
   * @return {@code true} iff the node is {@code SVA_PORT}.
   */
  public boolean isSvaPort() {
    return getTag() == Tag.SVA_PORT;
  }

  /**
   * Checks whether the node is {@code DECLARATION}.
   * 
   * @return {@code true} iff the node is {@code DECLARATION}.
   */
  public boolean isDeclaration() {
    return getTag() == Tag.DECLARATION;
  }

  /**
   * Checks whether the node is {@code SEQUENCE_DECLARATION}.
   *
   * @return {@code true} iff the node is {@code SEQUENCE_DECLARATION}.
   */
  public boolean isSequenceDeclaration() {
    return getTag() == Tag.SEQUENCE_DECLARATION;
  }

  /**
   * Checks whether the node is {@code PROPERTY_DECLARATION}.
   *
   * @return {@code true} iff the node is {@code PROPERTY_DECLARATION}.
   */
  public boolean isPropertyDeclaration() {
    return getTag() == Tag.PROPERTY_DECLARATION;
  }

  /**
   * Checks whether the node is {@code ACTIVITY}.
   * 
   * @return {@code true} iff the node is {@code ACTIVITY}.
   */
  public boolean isActivity() {
    return getTag() == Tag.ACTIVITY;
  }

  /**
   * Checks whether the node is {@code ASSIGN}.
   * 
   * @return {@code true} iff the node is {@code ASSIGN}.
   */
  public boolean isAssign() {
    return getTag() == Tag.ASSIGN;
  }

  /**
   * Checks whether the node is {@code ASSIGNMENT}.
   * 
   * @return {@code true} iff the node is {@code ASSIGNMENT}.
   */
  public boolean isAssignment() {
    return getTag() == Tag.ASSIGNMENT;
  }

  /**
   * Checks whether the node is {@code INSTANTIATION}.
   * 
   * @return {@code true} iff the node is {@code INSTANTIATION}.
   */
  public boolean isInstantiation() {
    return getTag() == Tag.INSTANTIATION;
  }

  /**
   * Checks whether the node is {@code PORT_CONNECTION}.
   * 
   * @return {@code true} iff the node is {@code PORT_CONNECTION}.
   */
  public boolean isConnection() {
    return getTag() == Tag.PORT_CONNECTION;
  }

  /**
   * Checks whether the node is {@code PROCEDURE}.
   * 
   * @return {@code true} iff the node is {@code PROCEDURE}.
   */
  public boolean isProcedure() {
    return getTag() == Tag.PROCEDURE;
  }

  /**
   * Checks whether the node is {@code TABLE}.
   * 
   * @return {@code true} iff the node is {@code TABLE}.
   */
  public boolean isTable() {
    return getTag() == Tag.TABLE;
  }

  /**
   * Checks whether the node is {@code TABLE_ENTRY}.
   * 
   * @return {@code true} iff the node is {@code TABLE_ENTRY}.
   */
  public boolean isTableEntry() {
    return getTag() == Tag.TABLE_ENTRY;
  }

  /**
   * Checks whether the node is {@code NULL_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code NULL_STATEMENT}.
   */
  public boolean isNullStatement() {
    return getTag() == Tag.NULL_STATEMENT;
  }

  /**
   * Checks whether the node is {@code ASSIGN_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code ASSIGN_STATEMENT}.
   */
  public boolean isAssignStatement() {
    return getTag() == Tag.ASSIGN_STATEMENT;
  }

  /**
   * Checks whether the node is {@code ASSERTION_STATEMENT}.
   *
   * @return {@code true} iff the node is {@code ASSERTION_STATEMENT}.
   */
  public boolean isAssertionStatement() {
    return getTag() == Tag.ASSERTION_STATEMENT;
  }

  /**
   * Checks whether the node is {@code TASK_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code TASK_STATEMENT}.
   */
  public boolean isTaskStatement() {
    return getTag() == Tag.TASK_STATEMENT;
  }

  /**
   * Checks whether the node is {@code WAIT_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code WAIT_STATEMENT}.
   */
  public boolean isWaitStatement() {
    return getTag() == Tag.WAIT_STATEMENT;
  }

  /**
   * Checks whether the node is {@code DISABLE_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code DISABLE_STATEMENT}.
   */
  public boolean isDisableStatement() {
    return getTag() == Tag.DISABLE_STATEMENT;
  }

  /**
   * Checks whether the node is {@code TRIGGER_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code TRIGGER_STATEMENT}.
   */
  public boolean isTriggerStatement() {
    return getTag() == Tag.TRIGGER_STATEMENT;
  }

  /**
   * Checks whether the node is {@code DELAYED_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code DELAYED_STATEMENT}.
   */
  public boolean isDelayedStatement() {
    return getTag() == Tag.DELAYED_STATEMENT;
  }

  /**
   * Checks whether the node is {@code IF_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code IF_STATEMENT}.
   */
  public boolean isIfStatement() {
    return getTag() == Tag.IF_STATEMENT;
  }

  /**
   * Checks whether the node is {@code IF_STATEMENT_BRANCH}.
   * 
   * @return {@code true} iff the node is {@code IF_STATEMENT_BRANCH}.
   */
  public boolean isIfStatementBranch() {
    return getTag() == Tag.IF_STATEMENT_BRANCH;
  }

  /**
   * Checks whether the node is {@code CASE_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code CASE_STATEMENT}.
   */
  public boolean isCaseStatement() {
    return getTag() == Tag.CASE_STATEMENT;
  }

  /**
   * Checks whether the node is {@code CASE_STATEMENT_ITEM}.
   * 
   * @return {@code true} iff the node is {@code CASE_STATEMENT_ITEM}.
   */
  public boolean isCaseStatementItem() {
    return getTag() == Tag.CASE_STATEMENT_ITEM;
  }

  /**
   * Checks whether the node is {@code LOOP_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code LOOP_STATEMENT}.
   */
  public boolean isLoopStatement() {
    return getTag() == Tag.LOOP_STATEMENT;
  }

  /**
   * Checks whether the node is {@code BLOCK_STATEMENT}.
   * 
   * @return {@code true} iff the node is {@code BLOCK_STATEMENT}.
   */
  public boolean isBlockStatement() {
    return getTag() == Tag.BLOCK_STATEMENT;
  }

  /**
   * Checks whether the node is {@code GENERATE}.
   * 
   * @return {@code true} iff the node is {@code GENERATE}.
   */
  public boolean isGenerate() {
    return getTag() == Tag.GENERATE;
  }

  /**
   * Checks whether the node is {@code IF_GENERATE}.
   * 
   * @return {@code true} iff the node is {@code IF_GENERATE}.
   */
  public boolean isIfGenerate() {
    return getTag() == Tag.IF_GENERATE;
  }

  /**
   * Checks whether the node is {@code IF_GENERATE_BRANCH}.
   * 
   * @return {@code true} iff the node is {@code IF_GENERATE_BRANCH}.
   */
  public boolean isIfGenerateBranch() {
    return getTag() == Tag.IF_GENERATE_BRANCH;
  }

  /**
   * Checks whether the node is {@code CASE_GENERATE}.
   * 
   * @return {@code true} iff the node is {@code CASE_GENERATE}.
   */
  public boolean isCaseGenerate() {
    return getTag() == Tag.CASE_GENERATE;
  }

  /**
   * Checks whether the node is {@code CASE_GENERATE_ITEM}.
   * 
   * @return {@code true} iff the node is {@code CASE_GENERATE_ITEM}.
   */
  public boolean isCaseGenerateItem() {
    return getTag() == Tag.CASE_GENERATE_ITEM;
  }

  /**
   * Checks whether the node is {@code LOOP_GENERATE}.
   * 
   * @return {@code true} iff the node is {@code LOOP_GENERATE}.
   */
  public boolean isLoopGenerate() {
    return getTag() == Tag.LOOP_GENERATE;
  }

  /**
   * Checks whether the node is {@code BLOCK_GENERATE}.
   * 
   * @return {@code true} iff the node is {@code BLOCK_GENERATE}.
   */
  public boolean isBlockGenerate() {
    return getTag() == Tag.BLOCK_GENERATE;
  }

  /**
   * Checks whether the node is {@code SPECIFY}.
   * 
   * @return {@code true} iff the node is {@code SPECIFY}.
   */
  public boolean isSpecify() {
    return getTag() == Tag.SPECIFY;
  }

  /**
   * Checks whether the node is {@code PULSE_STYLE}.
   * 
   * @return {@code true} iff the node is {@code PULSE_STYLE}.
   */
  public boolean isPulseStyle() {
    return getTag() == Tag.PULSE_STYLE;
  }

  /**
   * Checks whether the node is {@code SHOW_CANCELLED}.
   * 
   * @return {@code true} iff the node is {@code SHOW_CANCELLED}.
   */
  public boolean isShowCancelled() {
    return getTag() == Tag.SHOW_CANCELLED;
  }

  /**
   * Checks whether the node is {@code PATH_DECLARATION}.
   * 
   * @return {@code true} iff the node is {@code PATH_DECLARATION}.
   */
  public boolean isPathDeclaration() {
    return getTag() == Tag.PATH_DECLARATION;
  }

  /**
   * Checks whether the node is {@code ATTRIBUTE}.
   * 
   * @return {@code true} iff the node is {@code ATTRIBUTE}.
   */
  public boolean isAttribute() {
    return getTag() == Tag.ATTRIBUTE;
  }

  /**
   * Checks whether the node is {@code PORT_CONNECTION}.
   * 
   * @return {@code true} iff the node is {@code PORT_CONNECTION}.
   */
  public boolean isPortConnection() {
    return getTag() == Tag.PORT_CONNECTION;
  }

  /**
   * Returns the attributes of the node.
   * 
   * @return the attributes.
   */
  public Collection<VerilogAttribute> getAttributes() {
    return attributes;
  }

  /**
   * Sets attributes of the node.
   * 
   * @param attributes the attributes to be set.
   */
  public void setAttributes(final Collection<VerilogAttribute> attributes) {
    this.attributes = attributes;
  }

  /**
   * Returns the parent node.
   * 
   * @return the parent node.
   */
  public VerilogNode getParentNode() {
    return (VerilogNode) getParent();
  }

  @Override
  public abstract VerilogNode clone();
}
