/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogProcedure} represents functions and tasks.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogProcedure extends VerilogNode {
  public static final Tag TAG = Tag.PROCEDURE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(
      // Variables and parameters.
      Tag.DECLARATION,
      // Statements.
      Tag.NULL_STATEMENT,
      Tag.ASSIGN_STATEMENT,
      Tag.TASK_STATEMENT,
      Tag.WAIT_STATEMENT,
      Tag.DISABLE_STATEMENT,
      Tag.TRIGGER_STATEMENT,
      Tag.DELAYED_STATEMENT,
      Tag.IF_STATEMENT,
      Tag.CASE_STATEMENT,
      Tag.LOOP_STATEMENT,
      Tag.BLOCK_STATEMENT
  );

  /** {@link Type} contains the procedure types. */
  public enum Type {
    /** Task declaration. */
    TASK,
    /** Function declaration. */
    FUNCTION
  }

  /** Type of procedure (by default, {@code TASK}). */
  private Type type = Type.TASK;
  /** Return value type (only for functions). */
  private VerilogElementType elementType;
  /** Automatic procedure flag. */
  private boolean automatic = false;
  /** Arguments of the procedure. */
  private List<VerilogPort> ports = new ArrayList<>();
  /** Output of the procedure (only for functions). */
  private VerilogPort output;

  /**
   * Creates a procedure.
   *
   * @param parent the parent node.
   */
  public VerilogProcedure(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.CLASS, parent);
  }

  /**
   * Creates a copy of the procedure.
   *
   * @param other the procedure to be copied.
   * @param parent the parent node.
   */
  public VerilogProcedure(final VerilogProcedure other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
    this.elementType = other.elementType != null ? other.elementType.clone() : null;
    this.automatic = other.automatic;

    for (final VerilogPort port : other.ports) {
      this.ports.add(port.clone());
    }

    this.output = other.output;
  }

  /**
   * Returns the type of the procedure.
   *
   * @return the procedure type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the procedure.
   *
   * @param type the procedure type.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the procedure is {@code TASK}.
   *
   * @return {@code true} iff the procedure is {@code TASK}.
   */
  public boolean isTask() {
    return type == Type.TASK;
  }

  /**
   * Checks whether the procedure is {@code FUNCTION}.
   *
   * @return {@code true} iff the procedure is {@code FUNCTION}.
   */
  public boolean isFunction() {
    return type == Type.FUNCTION;
  }

  /**
   * Sets the type of the procedure to {@code TASK}.
   */
  public void setTask() {
    type = Type.TASK;
  }

  /**
   * Sets the type of the procedure to {@code FUNCTION}.
   */
  public void setFunction() {
    type = Type.FUNCTION;
  }

  /**
   * Checks whether the procedure is a system one.
   *
   * @return {@code true} iff the procedure is a system one.
   */
  public boolean isSystem() {
    final String name = getName();
    return name != null && name.startsWith("$");
  }

  /**
   * Returns the return value type of the procedure.
   *
   * @return the return value type of the procedure.
   */
  public VerilogElementType getElementType() {
    return elementType;
  }

  /**
   * Sets the return value type of the procedure and declares the output.
   *
   * @param name the procedure name.
   * @param elementType the return value type.
   */
  public void setOutput(final String name, final VerilogElementType elementType) {
    final VerilogDeclaration declaration = new VerilogDeclaration(this);

    declaration.setName(name);
    declaration.setType(VerilogDeclaration.Type.OUTPUT);
    declaration.setElementType(elementType);

    add(declaration);
    setOutput(declaration);

    this.elementType = elementType;
  }

  /**
   * Set the output from the declaration.
   *
   * @param declaration the declaration, whose output to be set.
   */
  public void setOutput(final VerilogDeclaration declaration) {
    InvariantChecks.checkNotNull(declaration);
    InvariantChecks.checkTrue(declaration.isOutput());

    final VerilogPort output = new VerilogPort(this);

    final VerilogReference reference = new VerilogReference();
    final VerilogPath path = new VerilogPath();
    final VerilogPathItem item = new VerilogPathItem(declaration.getName());

    path.addItem(item);
    reference.setPath(path);
    reference.setDeclaration(declaration);

    output.setName(declaration.getName());
    output.addReference(reference);

    this.output = output;
  }

  /**
   * Checks whether the procedure is automatic.
   *
   * @return {@code true} iff the procedure is automatic.
   */
  public boolean isAutomatic() {
    return automatic;
  }

  /**
   * Makes the procedure automatic.
   */
  public void setAutomatic() {
    automatic = true;
  }

  /**
   * Returns the signature of the procedure (i.e., the list of its ports).
   *
   * @return the signature of the procedure.
   */
  public List<VerilogPort> getPorts() {
    return ports;
  }

  /**
   * Adds the port to the signature of the procedure.
   *
   * @param port the port to be added.
   */
  public void addInput(final VerilogPort port) {
    InvariantChecks.checkNotNull(port);
    ports.add(port);
  }

  /**
   * Adds the port from the declaration to the signature of the procedure.
   *
   * @param declaration the declaration, whose port to be added.
   */
  public void addPort(final VerilogDeclaration declaration) {
    InvariantChecks.checkNotNull(declaration);

    if (declaration.isPort()) {
      final VerilogPort port = new VerilogPort(this);

      final VerilogReference reference = new VerilogReference();
      final VerilogPath path = new VerilogPath();
      final VerilogPathItem item = new VerilogPathItem(declaration.getName());

      path.addItem(item);
      reference.setPath(path);
      reference.setDeclaration(declaration);

      port.setName(declaration.getName());
      port.addReference(reference);

      addInput(port);
    }
  }

  /**
   * Returns the output of the procedure.
   *
   * @return the output of the procedure.
   */
  public VerilogPort getOutput() {
    return output;
  }

  /**
   * Returns the statement of the procedure.
   *
   * @return the procedure statement.
   */
  public VerilogStatement getStatement() {
    return (VerilogStatement) items(VerilogStatement.TAGS).iterator().next();
  }

  /**
   * Sets the statement of the procedure.
   *
   * @param statement the statement to be added.
   */
  public void setStatement(final VerilogStatement statement) {
    add(statement);
  }

  @Override
  public VerilogProcedure clone() {
    return new VerilogProcedure(this, getParentNode());
  }
}
