/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogPulseStyle} represents pulse style specifications.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogPulseStyle extends VerilogNode {
  public static final Tag TAG = Tag.PULSE_STYLE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** {@link VerilogPulseStyle} contains the pulse style types. */
  public enum Type {
    /** On event: {@code onevent ...}. */
    ON_EVENT,
    /** On detect: {@code ondetect ...}. */
    ON_DETECT
  }

  /** Type of the pulse style (by default, {@code ON_EVENT}). */
  private Type type = Type.ON_EVENT;
  /** References. */
  private List<VerilogReference> references = new ArrayList<>();

  /**
   * Creates a pulse style specification.
   * 
   * @param parent the parent node.
   */
  public VerilogPulseStyle(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
  }

  /**
   * Creates a copy of the pulse style specification.
   * 
   * @param other the pulse style specification to be copied.
   * @param parent the parent node.
   */
  public VerilogPulseStyle(final VerilogPulseStyle other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;

    for (final VerilogReference reference : other.references) {
      this.references.add(reference.clone());
    }
  }

  /**
   * Returns the type of the pulse style specification.
   * 
   * @return the pulse style type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the pulse style specification.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the pulse style is {@code ON_EVENT}.
   * 
   * @return {@code true} iff the pulse style is {@code ON_EVENT}.
   */
  public boolean isOnEvent() {
    return type == Type.ON_EVENT;
  }

  /**
   * Checks whether the pulse style is {@code ON_DETECT}.
   * 
   * @return {@code true} iff the pulse style is {@code ON_DETECT}.
   */
  public boolean isOnDetect() {
    return type == Type.ON_DETECT;
  }

  /**
   * Sets a pulse style type to {@code ON_EVENT}.
   */
  public void setOnEvent() {
    type = Type.ON_EVENT;
  }

  /**
   * Sets a pulse style type to {@code ON_DETECT}.
   */
  public void setOnDetect() {
    type = Type.ON_DETECT;
  }

  /**
   * Returns the references of the pulse style specification.
   * 
   * @return the list of references.
   */
  public List<VerilogReference> getReferences() {
    return references;
  }

  /**
   * Adds the reference to the pulse style specification.
   * 
   * @param reference the reference to be added.
   */
  public void addReference(final VerilogReference reference) {
    InvariantChecks.checkNotNull(reference);
    this.references.add(reference);
  }

  @Override
  public VerilogPulseStyle clone() {
    return new VerilogPulseStyle(this, getParentNode());
  }
}
