/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogEventControl} represents event control constructs.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogEventControl {
  /** Delay: {@code #d}. */
  private VerilogDelay delay;
  /** Events: {@code posedge x, negedge y}. */
  private List<VerilogEvent> events;
  /** Repeat: {@code repeat n ...}. */
  private VerilogExpression repeat;

  /**
   * Creates an event control.
   */
  public VerilogEventControl() {}

  /**
   * Creates a copy of the event control.
   * 
   * @param other the event control to be copied.
   */
  public VerilogEventControl(final VerilogEventControl other) {
    InvariantChecks.checkNotNull(other);

    this.delay = other.delay != null ? other.delay.clone() : null;
    this.repeat = other.repeat != null ? other.repeat.clone() : null;

    if (other.events != null) {
      this.events = new ArrayList<>(other.events.size());

      for (final VerilogEvent event : other.events) {
        this.events.add(event.clone());
      }
    }
  }

  /**
   * Checks whether the control is delay-based.
   * 
   * @return {@code true} iff the control is delay-based.
   */
  public boolean isDelays() {
    return delay != null;
  }

  /**
   * Checks whether the control is event-based.
   * 
   * @return {@code true} iff the control is event-based.
   */
  public boolean isEvents() {
    return events != null && !events.isEmpty() && repeat == null;
  }

  /**
   * Checks whether the control is event-based and has the repeat construct.
   * 
   * @return {@code true} iff the control is event-based and has the repeat construct.
   */
  public boolean isRepeat() {
    return events != null && !events.isEmpty() && repeat != null;
  }

  /**
   * Checks whether the control is absent.
   * 
   * @return {@code true} iff the control is absent.
   */
  public boolean isEmpty() {
    return !(isDelays() || isEvents() || isRepeat());
  }

  /**
   * Returns the delay of the delay-based control.
   * 
   * @return the delay.
   */
  public VerilogDelay getDelay() {
    return delay;
  }

  /**
   * Sets the delay of the delay-based control.
   * 
   * @param delay the delay to be set.
   */
  public void setDelay(final VerilogDelay delay) {
    InvariantChecks.checkNotNull(delay);
    this.delay = delay;
  }

  /**
   * Returns the events of the event-based control.
   * 
   * @return the events.
   */
  public List<VerilogEvent> getEvents() {
    return events;
  }

  /**
   * Sets the events of the event-based control.
   * 
   * @param events the events to be set.
   */
  public void setEvents(final List<VerilogEvent> events) {
    InvariantChecks.checkNotNull(events);
    this.events = events;
  }

  /**
   * Returns the repetition number of the event-based control.
   * 
   * @return the repetition number.
   */
  public VerilogExpression getRepeat() {
    return repeat;
  }

  /**
   * Sets the number of repetitions of the event-based control.
   * 
   * @param repeat the repetition number to be set.
   */
  public void setRepeat(final VerilogExpression repeat) {
    InvariantChecks.checkNotNull(repeat);
    this.repeat = repeat;
  }

  /**
   * Sets the number of repetitions and the events of the event-based control.
   * 
   * @param repeatEvents the repeat events construct.
   */
  public void setRepeatEvents(final VerilogRepeatEvents repeatEvents) {
    InvariantChecks.checkNotNull(repeatEvents);

    setRepeat(repeatEvents.getExpression());
    setEvents(repeatEvents.getEvents());
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();

    if (delay != null) {
      if (builder.length() != 0) {
        builder.append(' ');
      }
      builder.append(delay);
    }

    if (events != null) {
      if (builder.length() != 0) {
        builder.append(' ');
      }
      builder.append(events);
    }

    if (repeat != null) {
      if (builder.length() != 0) {
        builder.append(' ');
      }
      builder.append(repeat);
    }

    return builder.toString();
  }

  @Override
  public VerilogEventControl clone() {
    return new VerilogEventControl(this);
  }
}
