/*
 * Copyright 2012-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.util.Parser;

import java.math.BigInteger;

/**
 * {@link VerilogLiteral} represents numeric and string literals.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogLiteral {

  /** {@link Type} contains literal types. */
  public enum Type {
    BIT_VECTOR,
    INTEGER,
    REAL,
    STRING
  }

  /**
   * Parses the string representing a numeric literal.
   * @param text the string to be parsed.
   * @return the constructed literal.
   */
  public static VerilogLiteral parseNumber(final String text) {
    return Parser.parseNumber(text);
  }

  /**
   * Parses the string representing a string literal.
   * @param text the string to be parsed.
   * @param withQuotes the flag indicating whether the string contains quotes to be removed.
   * @return the constructed literal.
   */
  public static VerilogLiteral parseString(final String text, final boolean withQuotes) {
    return new VerilogLiteral(withQuotes ? text.substring(1, text.length() - 1) : text);
  }

  /** The literal type. */
  private final Type type;

  /** The bit vector value. */
  private final BitVector value;
  /** The integer value. */
  private final BigInteger integer;
  /** The real value. */
  private final double real;
  /** The string value. */
  private final String string;

  /** The X mask (bit vector MASK such that {@code MASK[i] == 1} iff {@code VECTOR[i] === X)}. */
  private final BitVector xmask;
  /** The Z mask (bit vector MASK such that {@code MASK[i] == 1} iff {@code VECTOR[i] === Z)}. */
  private final BitVector zmask;

  /**
   * Creates a bit vector literal.
   *
   * @param value the value.
   * @param xmask the X mask.
   * @param zmask the Z mask.
   */
  public VerilogLiteral(
      final BitVector value,
      final BitVector xmask,
      final BitVector zmask) {
    InvariantChecks.checkNotNull(value);
    // The xmask and zmask parameters can be null.

    this.type = Type.BIT_VECTOR;
    this.value = value;
    this.xmask = xmask;
    this.zmask = zmask;

    this.integer = null;
    this.real = 0;
    this.string = null;
  }

  /**
   * Creates a bit vector literal with no X and Z digits.
   *
   * @param value the value.
   */
  public VerilogLiteral(final BitVector value) {
    this(value, null, null);
  }

  /**
   * Creates a bit vector literal.
   *
   * @param value the value.
   * @param width the width.
   */
  public VerilogLiteral(final int value, int width) {
    this(BitVector.valueOf(value, width));
  }

  /**
   * Creates an integer literal.
   *
   * @param value the value.
   */
  public VerilogLiteral(final int value) {
    this.type = Type.INTEGER;
    this.integer = BigInteger.valueOf(value);
    this.value = BitVector.valueOf(value, Integer.SIZE);

    this.real = 0;
    this.string = null;

    this.xmask = null;
    this.zmask = null;
  }

  /**
   * Creates an integer literal.
   *
   * @param value the value.
   */
  public VerilogLiteral(final boolean value) {
    this(value ? 1 : 0);
  }

  /**
   * Creates an integer literal.
   *
   * @param value the value.
   */
  public VerilogLiteral(final BigInteger value) {
    this.type = Type.INTEGER;
    this.integer = value;
    this.value = BitVector.valueOf(value, Integer.SIZE);

    this.real = 0;
    this.string = null;
    this.xmask = null;
    this.zmask = null;
  }

  /**
   * Creates a real literal.
   *
   * @param value the value.
   */
  public VerilogLiteral(final double value) {
    this.type = Type.REAL;
    this.real = value;

    this.integer = null;
    this.string = null;
    this.value = null;
    this.xmask = null;
    this.zmask = null;
  }

  /**
   * Creates a string literal.
   *
   * @param value the value.
   */
  public VerilogLiteral(final String value) {
    InvariantChecks.checkNotNull(value);

    this.type = Type.STRING;
    this.string = value;

    // Convert the string into the bit vector.
    this.value = BitVector.newEmpty(8 * string.length());

    for (int i = 0; i < value.length(); i++) {
      this.value.field(8 * i, 8 * i + 7).assign(BitVector.valueOf(string.charAt(i), 8));
    }

    this.integer = null;
    this.real = 0;
    this.xmask = null;
    this.zmask = null;
  }

  /**
   * Creates a literal.
   *
   * @param value the value.
   */
  public VerilogLiteral(final Data value) {
    switch (value.getType().getTypeId()) {
      case BIT_VECTOR:
        this.type = Type.BIT_VECTOR;
        this.value = value.getBitVector();

        this.integer = null;
        this.real = 0;
        this.string = null;

        break;
      case LOGIC_BOOLEAN:
        this.type = Type.INTEGER;
        this.integer = value.getBoolean() ? BigInteger.ONE : BigInteger.ZERO;
        this.value = BitVector.valueOf(value.getBoolean() ? 1 : 0, Integer.SIZE);

        this.real = 0;
        this.string = null;

        break;
      case LOGIC_INTEGER:
        this.type = Type.INTEGER;
        this.integer = value.getInteger();
        this.value = BitVector.valueOf(value.getInteger(), Integer.SIZE);

        this.real = 0;
        this.string = null;

        break;
      case LOGIC_REAL:
        this.type = Type.REAL;
        this.real = value.getReal();

        this.value = null;
        this.integer = null;
        this.string = null;

        break;
      case LOGIC_STRING:
        this.type = Type.STRING;
        this.string = value.getString();

        this.value = null;
        this.integer = null;
        this.real = 0;

        break;
      default:
        this.type = null;
        this.value = null;
        this.integer = null;
        this.real = 0;
        this.string = null;
    }

    this.xmask = null;
    this.zmask = null;
  }

  /**
   * Creates a copy of the literal.
   *
   * @param other the literal to be copied.
   */
  public VerilogLiteral(final VerilogLiteral other) {
    InvariantChecks.checkNotNull(other);

    this.type = other.type;
    this.integer = other.integer;
    this.real = other.real;
    this.string = other.string;

    this.value = other.value != null ? other.value.copy() : null;
    this.xmask = other.xmask != null ? other.xmask.copy() : null;
    this.zmask = other.zmask != null ? other.zmask.copy() : null;
  }

  /**
   * Returns the literal type ({@code INTEGER}, {@code REAL}, or {@code STRING}).
   *
   * @return the literal type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Checks whether the literal is {@code BIT_VECTOR}.
   *
   * @return {@code true} iff the literal is {@code BIT_VECTOR}.
   */
  public boolean isBitVector() {
    return type == Type.BIT_VECTOR;
  }

  /**
   * Checks whether the literal is {@code INTEGER}.
   *
   * @return {@code true} iff the literal is {@code INTEGER}.
   */
  public boolean isInteger() {
    return type == Type.INTEGER;
  }

  /**
   * Checks whether the literal is {@code REAL}.
   *
   * @return {@code true} iff the literal is {@code REAL}.
   */
  public boolean isReal() {
    return type == Type.REAL;
  }

  /**
   * Checks whether the literal is {@code STRING}.
   *
   * @return {@code true} iff the literal is {@code STRING}.
   */
  public boolean isString() {
    return type == Type.STRING;
  }

  /**
   * Returns the bit vector value of the literal.
   *
   * @return the bit vector value.
   */
  public BitVector getBitVector() {

    if (isReal()) {
      return BitVector.valueOf(getInteger(), Integer.SIZE);
    }

    return value;
  }

  /**
   * Returns the integer value of the literal.
   *
   * @return the integer value.
   */
  public int getInteger() {

    if (isReal()) {

      /*
       * 4.8.2 Conversion
       *
       * Real numbers shall be converted to integers by rounding the real number to the nearest
       * integer, rather than by truncating it. Implicit conversion shall take place when a real
       * number is assigned to an integer. If the fractional part of the real number is exactly 0.5,
       * it shall be rounded away from zero.
       */
      return (int)getReal() + ((getReal() % 1 >= 0.5) ? 1 : 0);
    }

    return isInteger() ? integer.intValue() : value.intValue();
  }

  /**
   * Returns the boolean value of the literal.
   *
   * @return the boolean value.
   */
  public boolean getBoolean() {
    InvariantChecks.checkTrue(isInteger());
    return integer.intValue() != 0;
  }

  /**
   * Returns the big integer value of the literal.
   *
   * @return the big integer value.
   */
  public BigInteger getBigInteger() {
    InvariantChecks.checkTrue(!isReal());
    return isInteger() ? integer : value.bigIntegerValue();
  }

  /**
   * Returns the real value of the literal.
   *
   * @return the real value.
   */
  public double getReal() {
    InvariantChecks.checkTrue(isReal());
    return real;
  }

  /**
   * Returns the string value of the literal.
   *
   * @return the string value.
   */
  public String getString() {
    InvariantChecks.checkTrue(isString());
    return string;
  }

  /**
   * Returns the byte array representation of the literal.
   *
   * @return the byte array representation.
   */
  public byte[] getByteArray() {
    return value.toByteArray();
  }

  /**
   * Checks whether the literal contains X digits.
   *
   * @return {@code true} iff the literal contains X digits.
   */
  public boolean hasX() {
    return xmask != null && !xmask.isAllReset();
  }

  /**
   * Checks whether the literal contains Z digits.
   *
   * @return {@code true} iff the literal contains Z digits.
   */
  public boolean hasZ() {
    return zmask != null && !zmask.isAllReset();
  }

  /**
   * Checks whether the literal contains X or Z digits.
   *
   * @return {@code true} iff the literal contains X or Z digits.
   */
  public boolean hasXorZ() {
    return hasX() || hasZ();
  }

  /**
   * Returns the X mask of the literal.
   *
   * @return the X mask.
   */
  public BitVector getXMask() {
    return xmask;
  }

  /**
   * Returns the Z mask of the literal.
   *
   * @return the Z mask.
   */
  public BitVector getZMask() {
    return zmask;
  }

  /**
   * Returns the bit width of the literal.
   *
   * @return the bit width.
   */
  public int getBitSize() {
    return value.getBitSize();
  }

  @Override
  public String toString() {
    switch (type) {
      case BIT_VECTOR:
        return String.format("%d'h%s", value.getBitSize(), value.toHexString());
      case INTEGER:
        return String.format("%d", integer.intValue());
      case REAL:
        return Double.toString(real);
      default:
        return string;
    }
  }

  @Override
  public VerilogLiteral clone() {
    return new VerilogLiteral(this);
  }
}
