/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogPath} represents paths.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogPath {
  /** Path is a sequence of path items. */
  private List<VerilogPathItem> path = new ArrayList<>();

  /**
   * Creates an empty path.
   */
  public VerilogPath() {}

  /**
   * Creates a path consisting of one path item.
   * 
   * @param item the path item.
   */
  public VerilogPath(final VerilogPathItem item) {
    InvariantChecks.checkNotNull(item);
    path.add(item);
  }

  /**
   * Creates a copy of the path.
   * 
   * @param other the path to be copied.
   */
  public VerilogPath(final VerilogPath other) {
    InvariantChecks.checkNotNull(other);

    this.path = new ArrayList<>(other.path.size());
    for (final VerilogPathItem item : other.path) {
      this.path.add(item.clone());
    }
  }

  /**
   * Checks whether the path is hierarchical.
   * 
   * @return {@code true} iff the path is hierarchical.
   */
  public boolean isHierarchical() {
    return path.size() == 1;
  }

  /**
   * Returns the list of path items.
   * 
   * @return the list of path items.
   */
  public List<VerilogPathItem> getPath() {
    return path;
  }

  /**
   * Return the first path item of the path.
   * 
   * @return the first path item.
   */
  public VerilogPathItem getPathItem() {
    InvariantChecks.checkNotEmpty(path);
    return path.iterator().next();
  }

  /**
   * Adds an item to the path.
   * 
   * @param item the item to be added.
   */
  public void addItem(final VerilogPathItem item) {
    path.add(item);
  }

  @Override
  public String toString() {
    final StringBuffer buffer = new StringBuffer();

    for (final VerilogPathItem item : path) {
      if (buffer.length() != 0) {
        buffer.append('.');
      }

      buffer.append(item);
    }

    return buffer.toString();
  }

  @Override
  public VerilogPath clone() {
    return new VerilogPath(this);
  }
}
