/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.util.ModelUtils;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogReference} represents references.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogReference {
  /** Path to an object. */
  private VerilogPath path;
  /** Indices and bits selection range. */
  private List<VerilogRange> indicesAndBitsSelection = new ArrayList<>();
  /** Flag indicating whether the reference has bits selection (set by the static checker). */
  private boolean bitsSelection = false;

  /** Reference declaration. */
  private VerilogDeclaration declaration;

  /** Expression. */
  private Node node;

  /**
   * Creates a reference.
   */
  public VerilogReference() {}

  /**
   * Creates a copy of the reference.
   *
   * @param other the reference to be copied.
   */
  public VerilogReference(final VerilogReference other) {
    InvariantChecks.checkNotNull(other);

    this.path = other.path != null ? other.path.clone() : null;

    for (final VerilogRange range : other.indicesAndBitsSelection) {
      this.indicesAndBitsSelection.add(range.clone());
    }

    this.bitsSelection = other.bitsSelection;
    this.declaration = other.declaration;

    this.node = other.node != null ? other.node.deepestCopy() : null;
  }

  /**
   * Returns the reference path.
   *
   * @return the path.
   */
  public VerilogPath getPath() {
    return path;
  }

  /**
   * Sets the reference path.
   *
   * @param path the path to be set.
   */
  public void setPath(final VerilogPath path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  /**
   * Returns the list of ranges.
   *
   * @return the list of ranges.
   */
  public List<VerilogRange> getIndicesAndBitsSelection() {
    // For simplicity, the reference grammar is specified as PATH (RANGE)*,
    // but it should be PATH (EXPRESSION)* (RANGE)?.
    return indicesAndBitsSelection;
  }

  /**
   * Returns the range for bits selection.
   *
   * @return the range for bits selection.
   */
  public VerilogRange getBitsSelection() {
    // For simplicity, the reference grammar is specified as PATH (RANGE)*,
    // but it should be PATH (EXPRESSION)* (RANGE)?.
    return !indicesAndBitsSelection.isEmpty()
        ? indicesAndBitsSelection.get(indicesAndBitsSelection.size() - 1)
        : null;
  }

  /**
   * Adds the selection to the reference.
   *
   * @param range the selection to be added.
   */
  public void addIndexOrBitsSelection(final VerilogRange range) {
    this.indicesAndBitsSelection.add(range);
  }

  /**
   * Checks whether the reference has bits selection.
   *
   * @return {@code true} iff the reference has bits selection.
   */
  public boolean hasBitsSelection() {
    return bitsSelection;
  }

  /**
   * Sets the bits selection flag.
   *
   * @param bitsSelection the bits selection flag.
   */
  public void setBitsSelection(boolean bitsSelection) {
    this.bitsSelection = bitsSelection;
  }

  /**
   * Returns the declaration (the reference descriptor).
   *
   * @return the declaration.
   */
  public VerilogDeclaration getDeclaration() {
    return declaration;
  }

  /**
   * Sets the declaration (the reference descriptor).
   *
   * @param declaration the declaration to be set.
   */
  public void setDeclaration(final VerilogDeclaration declaration) {
    InvariantChecks.checkNotNull(declaration);
    this.declaration = declaration;

    bitsSelection = (declaration.getDimensions().size() < indicesAndBitsSelection.size());
  }

  /**
   * Returns the node representing the reference.
   *
   * @return the node.
   */
  public Node getNode() {
    // If there is a node, just return it.
    if (node != null) {
      return node;
    }

    final Node variable = ModelUtils.getVariable(this);
    final List<VerilogRange> indicesAndBitSelection = getIndicesAndBitsSelection();

    if (indicesAndBitSelection.isEmpty() || declaration == null) {
      // Save the result.
      this.node = variable;
      return variable;
    }

    final VerilogDeclaration variableDeclaration = (VerilogDeclaration) declaration;
    final List<VerilogRange> dimensions = variableDeclaration.getDimensions();

    Node node = variable;

    for (int i = 0; i < indicesAndBitSelection.size(); i++) {
      final VerilogRange range = indicesAndBitSelection.get(i);

      final VerilogExpression lhs = range.getLeftExpression();
      final VerilogExpression rhs = range.getRightExpression();

      final Node upper;
      final Node lower;

      switch (range.getType()) {
        case PLUS:
          lower = lhs.getNode();
          upper = Nodes.bvsub(Nodes.bvadd(lhs.getNode(), rhs.getNode()), NodeValue.newInteger(1));
          break;
        case MINUS:
          lower = Nodes.bvadd(Nodes.bvsub(lhs.getNode(), rhs.getNode()), NodeValue.newInteger(1));
          upper = lhs.getNode();
          break;
        default:
          lower = rhs != null ? rhs.getNode() : lhs.getNode();
          upper = lhs.getNode();
          break;
      }

      final DataType lowerType = lower.getDataType();
      final DataType upperType = upper.getDataType();
      final DataType commonType =
          upperType.getSize() >= lowerType.getSize() ? upperType : lowerType;

      final Node high = TypeConversion.coerce(upper, commonType);
      final Node low  = TypeConversion.coerce(lower, commonType);

      if (i < dimensions.size()) {
        node = Nodes.select(node, low);
      } else {
        node = Nodes.bvextract(high, low, node);
      }
    }

    // Save the result.
    this.node = node;
    return node;
  }

  /**
   * Sets the node representation of the reference.
   *
   * @param node the node to be set.
   */
  public void setNode(final Node node) {
    InvariantChecks.checkNotNull(node);
    this.node = node;
  }

  @Override
  public String toString() {
    return indicesAndBitsSelection.isEmpty()
        ? path.toString()
        : String.format("%s%s", path, indicesAndBitsSelection);
  }

  @Override
  public VerilogReference clone() {
    return new VerilogReference(this);
  }
}
