/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.util;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogLiteral;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

/**
 * {@link Parser} implements a parser of integer, real and string literals.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class Parser {

  /**
   * {@link SizeBase} encapsulates literal size/base information.
   */
  public static class SizeBase {
    /** Bit size (-1 if unknown). */
    private int bitSize;
    /** Signed / unsigned. */
    private boolean isSigned;
    /** Number format (2, 8, 10 or 16). */
    private int base;
    /** Bits per digit (not for decimal). */
    private int bitsPerDigit;

    /**
     * Constructs a size/base object.
     * 
     * @param bitSize the bit size of the literal.
     * @param isSigned the signed/unsigned flag.
     * @param base the number system of the literal (2, 8, 10 or 16).
     */
    public SizeBase(final int bitSize, final boolean isSigned, final int base) {
      this.bitSize = bitSize;
      this.isSigned = isSigned;
      this.base = base;
    }

    /**
     * Constructs a default size/base object.
     */
    public SizeBase() {
      this(-1, false, -1);
    }

    /**
     * Returns the bit size of the literal.
     * 
     * @return the bit size.
     */
    public int getBitSize() {
      return bitSize;
    }

    /**
     * Sets the bit size of the literal.
     * 
     * @param bitSize the bit size to be set.
     */
    public void setBitSize(final int bitSize) {
      this.bitSize = bitSize;
    }

    /**
     * Returns the signed/unsigned flag of the literal.
     * 
     * @return {@code true} iff the literal is signed.
     */
    public boolean isSigned() {
      return isSigned;
    }

    /**
     * Sets the signed/unsigned flag of the literal.
     * 
     * @param isSigned the flag to be set.
     */
    public void setSigned(final boolean isSigned) {
      this.isSigned = isSigned;
    }

    /**
     * Returns the number system of the literal.
     * 
     * @return the number system.
     */
    public int getBase() {
      return base;
    }

    /**
     * Sets the number system of the literal.
     * 
     * @param base the number system to be set.
     */
    public void setBase(final int base) {
      this.base = base;

      switch (base) {
        case 2: {
          this.bitsPerDigit = 1;
          break;
        }
        case 8: {
          this.bitsPerDigit = 3;
          break;
        }
        default: {
          this.bitsPerDigit = 4;
          break;
        }
      }
    }

    /**
     * Returns the number of bits per digit.
     * 
     * @return the number of bits per digit.
     */
    public int getBitsPerDigit() {
      return bitsPerDigit;
    }

    /**
     * Checks whether size/base information is undefined.
     * 
     * @return {@code true} iff the information is undefined.
     */
    public boolean isUndefined() {
      return getBase() == -1;
    }
  }

  /**
   * Parses the string and constructs the literal object.
   * 
   * <p>
   * The number format is follows:
   * 
   * NUMBER   : ((SIZE)? BASE)? INTEGER (FRACTION)? (EXPONENT)?
   * SIZE     : (DIGIT)+
   * BASE     : '\'' ('s' | 'S')? ('d' | 'h' | 'o' | 'b')
   * INTEGER  : DIGIT (DIGIT | '_')*
   * FRACTION : '.' (DIGIT | '_')*
   * EXPONENT : 'e' ('+'|'-')? (DIGIT)+
   * DIGIT    : ('0'..'9') | ('a'..'f') | 'x' | 'z' | '?'
   * </p>
   * 
   * @param text the string to be parsed.
   * @return the literal object.
   */
  public static VerilogLiteral parseNumber(final String text) {
    InvariantChecks.checkNotNull(text);

    String number = preProcess(text);

    final SizeBase base = new SizeBase();
    number = parseSizeBase(number, base);

    if (base.isUndefined() && (number.indexOf('e') != -1 || number.indexOf('.') != -1)) {
      return new VerilogLiteral(Double.parseDouble(number));
    }

    return parseInteger(number, base);
  }

  /**
   * Parses the string and constructs the literal object.
   * 
   * <p>
   * The integer format is follows:
   * 
   * INTEGER : DIGIT+
   * DIGIT   : ('0'..'9') | ('a'..'f') | 'x' | 'z'
   * </p>
   * 
   * @param text the string to be parsed.
   * @param base the size and the base of the literal.
   * @return the constructed literal. 
   */
  private static VerilogLiteral parseInteger(final String text, final SizeBase base) {
    BigInteger integer = BigInteger.ZERO;

    if (base.isUndefined()) {
      base.setBase(10);
    }

    final List<Byte> xmaskList = base.getBase() == 10 ? null : new ArrayList<Byte>();
    final List<Byte> zmaskList = base.getBase() == 10 ? null : new ArrayList<Byte>();

    int currentXMask = 0;
    int currentZMask = 0;
    int currentDigit = 0;
    int currentWidth = 0;

    for (int i = 0; i < text.length(); i++) {
      integer = integer.multiply(BigInteger.valueOf(base.getBase()));

      currentXMask <<= base.getBitsPerDigit();
      currentZMask <<= base.getBitsPerDigit();
      currentWidth +=  base.getBitsPerDigit();

      final int c = text.charAt(i);

      if ('0' <= c && c <= '9') {
        currentDigit = (c - '0');
      } else if ('a' <= c && c <= 'f') {
        currentDigit = (c - 'a') + 10;
      } else if ('x' == c || c == 'z') {
        InvariantChecks.checkTrue(base.getBase() != 10);

        currentDigit = 0;
        currentXMask |= (c == 'x' ? (base.getBase() - 1) : 0);
        currentZMask |= (c == 'z' ? (base.getBase() - 1) : 0);
      }

      InvariantChecks.checkTrue(currentDigit < base.getBase());
      InvariantChecks.checkTrue(xmaskList != null || (currentXMask | currentZMask) == 0);

      integer = integer.add(BigInteger.valueOf(currentDigit));

      if (xmaskList != null && (currentWidth % Integer.SIZE) == 0) {
        xmaskList.addAll(0, toByteList(currentXMask, Integer.SIZE));
        zmaskList.addAll(0, toByteList(currentZMask, Integer.SIZE));
      }
    }

    if (xmaskList != null) {
      xmaskList.addAll(0, toByteList(currentXMask, currentWidth % Integer.SIZE));
      zmaskList.addAll(0, toByteList(currentZMask, currentWidth % Integer.SIZE));
    }

    // TODO: Take into account base.isSigned().

    final int bitSize = base.getBitSize() > 0 ? base.getBitSize() : Integer.SIZE;
    final BitVector value = BitVector.valueOf(integer, bitSize);

    if (xmaskList == null) {
      return new VerilogLiteral(value);
    }

    final BitVector xmask = BitVector.valueOf(toByteArray(xmaskList), bitSize);
    final BitVector zmask = BitVector.valueOf(toByteArray(zmaskList), bitSize);

    return new VerilogLiteral(value, xmask, zmask);
  }

  /**
   * Parses the string and extracts the size/base information.
   * 
   * <p>
   * The size and base format is follows:
   * 
   * SIZE_BASE : ((SIZE)? BASE)?
   * SIZE      : (DIGIT)+
   * BASE      : '\'' ('s' | 'S')? ('d' | 'h' | 'o' | 'b')
   * DIGIT     : ('0'..'9')'
   * </p>
   * 
   * @param text the string to be parsed.
   * @param base the size and the base of the literal (output).
   * @return the string suffix with no size/base information.
   */
  private static String parseSizeBase(final String text, final SizeBase base /* OUTPUT */) {
    final int singleQuoteIdx = text.indexOf('\'');

    int nextCharIdx;
    if (singleQuoteIdx != -1) {
      nextCharIdx = singleQuoteIdx + 1;
      InvariantChecks.checkTrue(nextCharIdx < text.length());

      if (text.charAt(nextCharIdx) == 's') {
        base.setSigned(true);
        nextCharIdx++;
      }

      InvariantChecks.checkTrue(nextCharIdx < text.length());
      switch (text.charAt(nextCharIdx)) {
        case 'b':
          base.setBase(2);
          break;
        case 'o':
          base.setBase(8);
          break;
        case 'd':
          base.setBase(10);
          break;
        case 'h':
          base.setBase(16);
          break;
      }

      if (singleQuoteIdx > 0) {
        base.setBitSize(Integer.parseInt(text.substring(0, singleQuoteIdx)));
      }

      nextCharIdx++;
    } else {
      nextCharIdx = 0;
    }

    return text.substring(nextCharIdx);
  }

  /**
   * Preprocesses the given string.
   * 
   * <ul>
   * <li>transforms the string into the lower case;</li>
   * <li>removes all occurrences of the character '_';</li>
   * <li>replaces all occurrences of the character '?' to 'z'.</li>
   * </ul>
   * 
   * @param text the string to be preprocessed.
   * @return the preprocessing result.
   */
  private static String preProcess(final String text) {
    return text.toLowerCase().replace(" ", "").replace("_", "").replace("?", "z");
  }

  /**
   * Transforms the given integer into the list of bytes.
   * 
   * @param value the integer value.
   * @param width the bit width.
   * @return the list of bytes.
   */
  private static List<Byte> toByteList(int value, int width) {
    final List<Byte> list = new ArrayList<>();

    while (width > 0) {
      list.add(0, (byte) (value & ((1 << Byte.SIZE) - 1)));

      value >>>= Byte.SIZE;
      width -= Byte.SIZE;
    }

    return list;
  }

  /**
   * Transforms the given list of bytes into the array of bytes.
   * 
   * @param list the list to be transformed.
   * @return the array of bytes.
   */
  private static byte[] toByteArray(final List<Byte> list) {
    if (list == null) {
      return null;
    }

    final byte[] array = new byte[list.size()];

    for (int i = 0; i < list.size(); i++) {
      array[i] = list.get(i);
    }

    return array;
  }
}
