/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sample;

import ru.ispras.fortress.expression.printer.ExprPrinter;
import ru.ispras.verilog.parser.model.VerilogActivity;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogAttribute;
import ru.ispras.verilog.parser.model.VerilogBlockGenerate;
import ru.ispras.verilog.parser.model.VerilogBlockStatement;
import ru.ispras.verilog.parser.model.VerilogCaseGenerate;
import ru.ispras.verilog.parser.model.VerilogCaseGenerateItem;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatementItem;
import ru.ispras.verilog.parser.model.VerilogCode;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogDefineParameter;
import ru.ispras.verilog.parser.model.VerilogDelayedStatement;
import ru.ispras.verilog.parser.model.VerilogDisableStatement;
import ru.ispras.verilog.parser.model.VerilogGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerateBranch;
import ru.ispras.verilog.parser.model.VerilogIfStatement;
import ru.ispras.verilog.parser.model.VerilogIfStatementBranch;
import ru.ispras.verilog.parser.model.VerilogInstantiation;
import ru.ispras.verilog.parser.model.VerilogLoopGenerate;
import ru.ispras.verilog.parser.model.VerilogLoopStatement;
import ru.ispras.verilog.parser.model.VerilogModule;
import ru.ispras.verilog.parser.model.VerilogNullStatement;
import ru.ispras.verilog.parser.model.VerilogPathDeclaration;
import ru.ispras.verilog.parser.model.VerilogPort;
import ru.ispras.verilog.parser.model.VerilogPortConnection;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.VerilogPulseStyle;
import ru.ispras.verilog.parser.model.VerilogShowCancelled;
import ru.ispras.verilog.parser.model.VerilogSpecify;
import ru.ispras.verilog.parser.model.VerilogTable;
import ru.ispras.verilog.parser.model.VerilogTableEntry;
import ru.ispras.verilog.parser.model.VerilogTaskStatement;
import ru.ispras.verilog.parser.model.VerilogTriggerStatement;
import ru.ispras.verilog.parser.model.VerilogWaitStatement;
import ru.ispras.verilog.parser.model.basis.VerilogDelay;
import ru.ispras.verilog.parser.model.basis.VerilogEdge;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogEventControl;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathDescription;
import ru.ispras.verilog.parser.model.basis.VerilogRange;
import ru.ispras.verilog.parser.model.basis.VerilogReference;
import ru.ispras.verilog.parser.sva.SvaAssertionStatement;
import ru.ispras.verilog.parser.sva.SvaInstantiation;
import ru.ispras.verilog.parser.sva.SvaPort;
import ru.ispras.verilog.parser.sva.SvaPropertyDeclaration;
import ru.ispras.verilog.parser.sva.SvaSequenceDeclaration;
import ru.ispras.verilog.parser.sva.basis.SvaProperty;
import ru.ispras.verilog.parser.walker.VerilogNodeVisitor;

import java.io.PrintStream;
import java.util.List;

/**
 * {@link VerilogNodePrinter} illustrates development of a Verilog node visitor.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogNodePrinter extends VerilogNodeVisitor {
  /** The flag indicating whether to print debug comments. */
  private final boolean debugComments;

  /** The output stream. */
  private final PrintStream out = System.out;

  /** The current indentation. */
  private int indent = 0;

  /**
   * Creates a node printer.
   *
   * @param debugComments the flag indicating whether to print debug comments.
   */
  public VerilogNodePrinter(final boolean debugComments) {
    this.debugComments = debugComments;
  }

  //================================================================================================
  // Activity
  //================================================================================================

  @Override
  public Result onActivityBegin(final VerilogActivity node) {
    indent();

    switch (node.getType()) {
      case INITIAL:
        text("initial");
        break;
      case ALWAYS:
        text("always");
        break;
    }

    endl();
    begin();

    return Result.OK;
  }

  @Override
  public Result onActivityEnd(final VerilogActivity node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // Assign
  //================================================================================================

  @Override
  public Result onAssignBegin(final VerilogAssign node) {
    indent();

    text("assign");

    final VerilogDelay delay = node.getDelay();

    if (delay.isSpecified()) {
      text(" ");
      text(delay);
    }

    text(" ");
    text(node.getAssignment());
    text(";");

    endl();

    return Result.OK;
  }

  @Override
  public Result onAssignEnd(final VerilogAssign node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // AssignStatement
  //================================================================================================

  @Override
  public Result onAssignStatementBegin(final VerilogAssignStatement node) {
    indent();

    String type = "=";

    switch (node.getType()) {
      case BLOCKING: {
        break;
      }
      case NON_BLOCKING: {
        type = "<=";
        break;
      }
      case ASSIGN: {
        text("assign ");
        break;
      }
      case DEASSIGN: {
        text("deassign ");
        break;
      }
      case FORCE: {
        text("force ");
        break;
      }
      case RELEASE: {
        text("release ");
        break;
      }
    }

    final VerilogEventControl control = node.getEventControl();

    if (control != null) {
      text(control);
      text(" ");
    }

    text(type, node.getAssignment());
    text(";");
    endl();

    return Result.OK;
  }

  @Override
  public Result onAssignStatementEnd(final VerilogAssignStatement node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // Assignment
  //================================================================================================

  @Override
  public Result onAssignmentBegin(final VerilogAssignment node) {
    indent();

    text(node);
    text(";");
    endl();

    return Result.OK;
  }

  @Override
  public Result onAssignmentEnd(final VerilogAssignment node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // Attribute
  //================================================================================================

  @Override
  public Result onAttributeBegin(final VerilogAttribute node) {
    // Nothing.
    return Result.OK;
  }

  @Override
  public Result onAttributeEnd(final VerilogAttribute node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // BlockGenerate
  //================================================================================================

  @Override
  public Result onBlockGenerateBegin(final VerilogBlockGenerate node) {
    indent();
    text("begin");

    final String name = node.getName();

    if (name != null) {
      text(": ");
      text(name);
    }

    endl();
    begin();

    return Result.OK;
  }

  @Override
  public Result onBlockGenerateEnd(final VerilogBlockGenerate node) {
    end();

    indent();
    text("end");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // BlockStatement
  //================================================================================================

  @Override
  public Result onBlockStatementBegin(final VerilogBlockStatement node) {
    indent();

    switch (node.getType()) {
      case BEGIN:
        text("begin");
        break;
      case FORK:
        text("fork");
        break;
    }

    final String name = node.getName();

    if (name != null) {
      text(": ");
      text(name);
    }

    endl();
    begin();

    return Result.OK;
  }

  @Override
  public Result onBlockStatementEnd(final VerilogBlockStatement node) {
    end();

    indent();
    text("end");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // CaseGenerate
  //================================================================================================

  @Override
  public Result onCaseGenerateBegin(final VerilogCaseGenerate node) {
    indent();
    text("case(");
    text(node.getExpression());
    text(")");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onCaseGenerateEnd(final VerilogCaseGenerate node) {
    end();

    indent();
    text("end");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // CaseGenerateItem
  //================================================================================================

  @Override
  public Result onCaseGenerateItemBegin(final VerilogCaseGenerateItem node) {
    indent();

    boolean comma = false;
    for (final VerilogExpression expression : node.getExpressions()) {
      if (comma) {
        text(", ");
      }

      comma = true;
      text(expression);
    }

    if (!comma) {
      text("default");
    }

    text(":");
    endl();

    begin();

    // Print the case condition (comment).
    if (debugComments) {
      final VerilogExpression condition = node.getExpression();

      indent();
      text("/* ASSERT: ");
      text(condition);
      text(" */");
      endl();
    }

    return Result.OK;
  }

  @Override
  public Result onCaseGenerateItemEnd(final VerilogCaseGenerateItem node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // CaseStatement
  //================================================================================================

  @Override
  public Result onCaseStatementBegin(final VerilogCaseStatement node) {
    indent();

    switch (node.getType()) {
      case CASE:
        text("case");
        break;
      case CASEX:
        text("casex");
        break;
      case CASEZ:
        text("casez");
        break;
    }

    text("(");
    text(node.getExpression());
    text(")");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onCaseStatementEnd(final VerilogCaseStatement node) {
    end();

    indent();
    text("end");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // CaseStatementItem
  //================================================================================================

  @Override
  public Result onCaseStatementItemBegin(final VerilogCaseStatementItem node) {
    indent();

    boolean comma = false;
    for (final VerilogExpression expression : node.getExpressions()) {
      if (comma) {
        text(", ");
      }

      comma = true;
      text(expression);
    }

    if (!comma) {
      text("default");
    }

    text(":");
    endl();

    begin();

    // Print the case condition (comment).
    if (debugComments) {
      final VerilogExpression condition = node.getExpression();

      indent();
      text("/* ASSERT: ");
      text(condition);
      text(" */");
      endl();
    }

    return Result.OK;
  }

  @Override
  public Result onCaseStatementItemEnd(final VerilogCaseStatementItem node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // Code
  //================================================================================================

  @Override
  public Result onCodeBegin(final VerilogCode node) {
    return Result.OK;
  }

  @Override
  public Result onCodeEnd(final VerilogCode node) {
    return Result.OK;
  }

  //================================================================================================
  // Declaration
  //================================================================================================

  @Override
  public Result onDeclarationBegin(final VerilogDeclaration node) {
    indent();

    boolean printed = false;

    if (node.getType() != VerilogDeclaration.Type.VARIABLE) {
      printed = true;
    }

    switch (node.getType()) {
      case INPUT: {
        text("input");
        break;
      }
      case OUTPUT: {
        text("output");
        break;
      }
      case INOUT: {
        text("inout");
        break;
      }
      case EVENT: {
        text("event");
        break;
      }
      case PARAMETER: {
        text("parameter");
        break;
      }
      case SPECPARAM: {
        text("specparam");
        break;
      }
      case LOCALPARAM: {
        text("localparam");
        break;
      }
      case GENVAR: {
        text("genvar");
        break;
      }
      case VARIABLE: {
        break;
      }
    }

    final VerilogElementType type = node.getElementType();

    if (type != null && type.isSpecified()) {
      if (printed) {
        text(" ");
      }
      text(type);

      printed = true;
    }

    if (printed) {
      text(" ");
    }

    text(node.getName());
    text(node.getDimensions());

    if (node.getExpression() != null) {
      text(" = ");
      text(node.getExpression());
    }

    text(";");
    endl();

    return Result.OK;
  }

  @Override
  public Result onDeclarationEnd(final VerilogDeclaration node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // DefineParameter
  //================================================================================================

  @Override
  public Result onDefineParameterBegin(final VerilogDefineParameter node) {
    indent();

    text("defparam");
    text(" ");
    text(node.getPath());
    text(" = ");
    text(node.getExpression());
    text(";");
    endl();

    return Result.OK;
  }

  @Override
  public Result onDefineParameterEnd(final VerilogDefineParameter node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // DelayedStatement
  //================================================================================================

  @Override
  public Result onDelayedStatementBegin(final VerilogDelayedStatement node) {
    indent();

    text(node.getEventControl());
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onDelayedStatementEnd(final VerilogDelayedStatement node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // DisableStatement
  //================================================================================================

  @Override
  public Result onDisableStatementBegin(final VerilogDisableStatement node) {
    indent();

    text("disable ");
    text(node.getPath());
    endl();

    return Result.OK;
  }

  @Override
  public Result onDisableStatementEnd(final VerilogDisableStatement node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // Generate
  //================================================================================================

  @Override
  public Result onGenerateBegin(final VerilogGenerate node) {
    indent();
    text("generate");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onGenerateEnd(final VerilogGenerate node) {
    end();

    indent();
    text("endgenerate");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // IfGenerate
  //================================================================================================

  @Override
  public Result onIfGenerateBegin(final VerilogIfGenerate node) {
    indent();
    text("if");
    text("(");
    text(node.getExpression());
    text(")");
    endl();

    return Result.OK;
  }

  @Override
  public Result onIfGenerateEnd(final VerilogIfGenerate node) {
    indent();
    text("end");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // IfGenerateBranch
  //================================================================================================

  @Override
  public Result onIfGenerateBranchBegin(final VerilogIfGenerateBranch node) {
    if (node.isElse()) {
      indent();
      text("else");
      endl();
    }

    begin();

    // Print the branch condition (comment).
    if (debugComments) {
      final VerilogExpression condition = node.getExpression();

      indent();
      text("/* ASSERT: ");
      text(condition);
      text(" */");
      endl();
    }

    return Result.OK;
  }

  @Override
  public Result onIfGenerateBranchEnd(final VerilogIfGenerateBranch node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // IfStatement
  //================================================================================================

  @Override
  public Result onIfStatementBegin(final VerilogIfStatement node) {
    indent();
    text("if");
    text("(");
    text(node.getExpression());
    text(")");
    endl();

    return Result.OK;
  }

  @Override
  public Result onIfStatementEnd(final VerilogIfStatement node) {
    indent();
    text("end");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // IfStatementBranch
  //================================================================================================

  @Override
  public Result onIfStatementBranchBegin(final VerilogIfStatementBranch node) {
    if (node.isElse()) {
      indent();
      text("else");
      endl();
    }

    begin();

    // Print the branch condition (comment).
    if (debugComments) {
      final VerilogExpression condition = node.getExpression();

      indent();
      text("/* ASSERT: ");
      text(condition);
      text(" */");
      endl();
    }

    return Result.OK;
  }

  @Override
  public Result onIfStatementBranchEnd(final VerilogIfStatementBranch node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // Instantiation
  //================================================================================================

  @Override
  public Result onInstantiationBegin(final VerilogInstantiation node) {
    final VerilogModule declaration = node.getDeclaration();

    if (debugComments) {
      indent();
      text("/* ");
      text(String.format("DECL: %s", declaration != null ? declaration.getName() : "null"));
      text(" */");
      endl();
    }

    indent();
    text(node.getModuleName());

    text(" ");
    text(node.getName());
    endl();

    indent();
    text("(");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onInstantiationEnd(final VerilogInstantiation node) {
    end();

    indent();
    text(");");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // LoopGenerate
  //================================================================================================

  @Override
  public Result onLoopGenerateBegin(final VerilogLoopGenerate node) {
    indent();
    text("for(");
    text(node.getInitialization());
    text("; ");
    text(node.getExpression());
    text("; ");
    text(node.getIteration());
    text(")");

    begin();

    return Result.OK;
  }

  @Override
  public Result onLoopGenerateEnd(final VerilogLoopGenerate node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // LoopStatement
  //================================================================================================

  @Override
  public Result onLoopStatementBegin(final VerilogLoopStatement node) {
    indent();

    switch (node.getType()) {
      case FOREVER: {
        text("forever");
        break;
      }
      case REPEAT: {
        text("repeat(");
        text(node.getExpression());
        text(")");
        break;
      }
      case WHILE: {
        text("while(");
        text(node.getExpression());
        text(")");
        break;
      }
      case FOR: {
        text("for(");
        text(node.getInitialization());
        text("; ");
        text(node.getExpression());
        text("; ");
        text(node.getIteration());
        text(")");
        break;
      }
    }

    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onLoopStatementEnd(final VerilogLoopStatement node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // Module
  //================================================================================================

  @Override
  public Result onModuleBegin(final VerilogModule node) {
    // Gate- and switch-level primitives are not printed.
    if (node.isGatePrimitive()) {
      return Result.ABORT;
    }

    indent();

    switch (node.getType()) {
      case MODULE:
        text("module");
        break;
      case MACROMODULE:
        text("macromodule");
        break;
      case USER_PRIMITIVE:
        text("primitive");
      default:
        break;
    }

    text(" " + node.getName());
    text("(");

    boolean comma = false;
    for (final VerilogPort port : node.getPorts()) {
      if (comma) {
        text(", ");
      }

      comma = true;

      if (port.getName() != null) {
        text("." + port.getName());
      }

      final List<VerilogReference> references = port.getReferences();

      if (references.size() > 1) {
        text("(");
      }
      textReferences(references);
      if (references.size() > 1) {
        text(")");
      }
    }
    text(");");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onModuleEnd(final VerilogModule node) {
    end();
    indent();

    switch (node.getType()) {
      case MODULE:
      case MACROMODULE:
        text("endmodule");
        break;
      case USER_PRIMITIVE:
        text("endprimitive");
        break;
      default:
        break;
    }

    endl();
    endl();

    return Result.OK;
  }

  //================================================================================================
  // NullStatement
  //================================================================================================

  @Override
  public Result onNullStatementBegin(final VerilogNullStatement node) {
    indent();
    text(";");
    endl();

    return Result.OK;
  }

  @Override
  public Result onNullStatementEnd(final VerilogNullStatement node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // PathDeclaration
  //================================================================================================

  @Override
  public Result onPathDeclarationBegin(final VerilogPathDeclaration node) {
    indent();

    switch (node.getType()) {
      case IF: {
        text("if(");
        text(node.getExpression());
        text(")");
        break;
      }
      case IF_NONE: {
        text("ifnone");
        break;
      }
      case SIMPLE: {
        final VerilogPathDescription description = node.getDescription();

        text("(");
        {
          text(description.getEdge());

          if (description.getEdge() != VerilogEdge.UNDEFINED) {
            text(" ");
          }

          textReferences(description.getInputs());

          text(" ");
          switch (description.getType()) {
            case PARALLEL: {
              text("=>");
              break;
            }
            case FULL: {
              text("*>");
              break;
            }
          }
          text(" ");

          textReferences(description.getOutputs());

          final VerilogExpression expression = description.getExpression();

          if (expression != null) {
            text(": ");
            text(expression);
          }

        }
        text(")");
        text(" = ");
        text(node.getDelay());

        break;
      }
    }

    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onPathDeclarationEnd(final VerilogPathDeclaration node) {
    end();

    return Result.OK;
  }

  //================================================================================================
  // Port
  //================================================================================================

  @Override
  public Result onPortBegin(final VerilogPort node) {
    // Nothing.
    return Result.OK;
  }

  @Override
  public Result onPortEnd(final VerilogPort node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // PortConnection
  //================================================================================================

  @Override
  public Result onPortConnectionBegin(final VerilogPortConnection node) {
    indent();

    if (node.getName() != null) {
      text(".");
      text(node.getName());
      text("(");
    }

    if (node.getExpression() != null) {
      text(node.getExpression());
    }

    if (node.getName() != null) {
      text(")");
    }

    endl();

    return Result.OK;
  }

  @Override
  public Result onPortConnectionEnd(final VerilogPortConnection node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // Procedure
  //================================================================================================

  @Override
  public Result onProcedureBegin(final VerilogProcedure node) {
    // System tasks and functions are not printed.
    if (node.isSystem()) {
      return Result.ABORT;
    }

    indent();

    final VerilogElementType type = node.getElementType();

    if (node.isFunction()) {
      text("function");
    } else {
      text("task");
    }

    if (node.isAutomatic()) {
      text(" ");
      text("automatic");
    }

    if (type != null && type.isSpecified()) {
      text(" ");
      text(type);
    }

    text(" ");
    text(node.getName());
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onProcedureEnd(final VerilogProcedure node) {
    end();
    indent();

    if (node.isFunction()) {
      text("endfunction");
    } else {
      text("endtask");
    }

    endl();

    return Result.OK;
  }

  //================================================================================================
  // PulseStyle
  //================================================================================================

  @Override
  public Result onPulseStyleBegin(final VerilogPulseStyle node) {
    switch (node.getType()) {
      case ON_EVENT: {
        text("pulsestyle_onevent");
        break;
      }
      case ON_DETECT: {
        text("pulsestyle_ondetect");
        break;
      }
    }

    text(" ");
    textReferences(node.getReferences());
    text(";");

    return Result.OK;
  }

  @Override
  public Result onPulseStyleEnd(final VerilogPulseStyle node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // ShowCancelled
  //================================================================================================

  @Override
  public Result onShowCancelledBegin(final VerilogShowCancelled node) {
    switch (node.getType()) {
      case SHOW: {
        text("showcancelled");
        break;
      }
      case NO_SHOW: {
        text("noshowcancelled");
        break;
      }
    }

    text(" ");
    textReferences(node.getReferences());
    text(";");

    return Result.OK;
  }

  @Override
  public Result onShowCancelledEnd(final VerilogShowCancelled node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // Specify
  //================================================================================================

  @Override
  public Result onSpecifyBegin(final VerilogSpecify node) {
    indent();
    text("specify");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onSpecifyEnd(final VerilogSpecify node) {
    end();

    indent();
    text("endspecify");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // Table
  //================================================================================================

  @Override
  public Result onTableBegin(final VerilogTable node) {
    indent();
    text("table");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onTableEnd(final VerilogTable node) {
    end();

    indent();
    text("endtable");
    endl();

    return Result.OK;
  }

  //================================================================================================
  // TableEntry
  //================================================================================================

  @Override
  public Result onTableEntryBegin(final VerilogTableEntry node) {
    indent();
    text(" ", node.getInputs1());
    text(" ");
    text(" ", node.getEdge());
    text(" ");
    text(" ", node.getInputs2());
    text(" : ");
    text(" ", node.getOutputs());
    text(";");
    endl();

    return Result.OK;
  }

  @Override
  public Result onTableEntryEnd(final VerilogTableEntry node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // TaskStatement
  //================================================================================================

  @Override
  public Result onTaskStatementBegin(final VerilogTaskStatement node) {
    final VerilogProcedure declaration = node.getDeclaration();

    if (debugComments) {
      indent();
      text("/* ");
      text(String.format("DECL: %s", declaration != null ? declaration.getName() : "null"));
      text(" */");
      endl();
    }

    indent();
    text(node.getPath());
    text("(");

    boolean comma = false;
    for (final VerilogExpression expression : node.getArguments()) {
      if (comma) {
        text(", ");
      }

      comma = true;
      text(expression);
    }

    text(");");
    endl();

    return Result.OK;
  }

  @Override
  public Result onTaskStatementEnd(final VerilogTaskStatement node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // TriggerStatement
  //================================================================================================

  @Override
  public Result onTriggerStatementBegin(final VerilogTriggerStatement node) {
    indent();
    text("-> ");
    text(node.getPath());

    for (final VerilogExpression expression : node.getArguments()) {
      text("(");
      text(expression);
      text(")");
    }

    endl();

    return Result.OK;
  }

  @Override
  public Result onTriggerStatementEnd(final VerilogTriggerStatement node) {
    // Nothing.
    return Result.OK;
  }

  //================================================================================================
  // WaitStatement
  //================================================================================================

  @Override
  public Result onWaitStatementBegin(final VerilogWaitStatement node) {
    indent();

    text("wait(");
    text(node.getExpression());
    text(")");
    endl();

    begin();

    return Result.OK;
  }

  @Override
  public Result onWaitStatementEnd(final VerilogWaitStatement node) {
    end();

    return Result.OK;
  }


  //================================================================================================
  // SVA nodes
  //================================================================================================

  @Override
  public Result onSequenceDeclarationBegin(final SvaSequenceDeclaration node) {
    indent();

    text(String.format("sequence %s", node.getName()));

    if (!node.getPorts().isEmpty()) {
      printPorts(node.getPorts());
    }
    text(";");
    endl();
    return Result.OK;
  }

  @Override
  public Result onSequenceDeclarationEnd(final SvaSequenceDeclaration node) {
    indent();

    text(String.format("%s", node.getExpression()));
    endl();
    indent();
    text("endsequence");
    endl();
    return Result.OK;
  }

  @Override
  public Result onPropertyDeclarationBegin(final SvaPropertyDeclaration node) {
    indent();

    text(String.format("property %s", node.getName()));

    if (!node.getPorts().isEmpty()) {
      printPorts(node.getPorts());
    }
    text(";");
    endl();
    return Result.OK;
  }

  @Override
  public Result onPropertyDeclarationEnd(final SvaPropertyDeclaration node) {
    indent();

    text(String.format("%s", node.getProperty()));
    endl();
    indent();
    text("endproperty");
    endl();
    return Result.OK;
  }

  @Override
  public Result onAssertionStatementBegin(final SvaAssertionStatement node) {
    indent();

    final String label = node.getName();

    if (label != null && !label.equals("")) {
      text(String.format("%s: ", label));
    }

    if (node.isAssert()) {
      text("assert ");
    }

    if (node.isAssume()) {
      text("assume ");
    }

    final VerilogExpression expression = node.getExpression();

    final SvaProperty property = node.getProperty();

    if (expression != null) {
      text(String.format("(%s)", expression));
    }

    if (property != null) {
      text(String.format("property (%s)", property.toString()));
    }

    endl();
    return Result.OK;
  }

  @Override
  public Result onAssertionStatementEnd(SvaAssertionStatement node) {
    return Result.OK;
  }

  @Override
  public Result onSvaInstanceBegin(final SvaInstantiation node) {
    return Result.OK;
  }

  @Override
  public Result onSvaInstanceEnd(final SvaInstantiation node) {
    return Result.OK;
  }

  //================================================================================================
  // Internals
  //================================================================================================

  /**
   * Increases the indentation.
   */
  private void begin() {
    indent++;
  }

  /**
   * Decreases the indentation.
   */
  private void end() {
    indent--;
  }

  /**
   * Prints the indentation.
   */
  private void indent() {
    for (int i = 0; i < indent; i++) {
      out.print("  ");
    }
  }

  /**
   * Prints the new line character.
   */
  private void endl() {
    out.println();
  }

  /**
   * Prints the list of references.
   *
   * @param node the list of references to be printed.
   */
  private void textReferences(final List<VerilogReference> node) {
    boolean comma = false;
    for (final VerilogReference reference : node) {
      if (comma) {
        text(", ");
      }

      comma = true;
      text(reference);
    }
  }

  /**
   * Prints the text.
   *
   * @param node the text to be printed.
   */
  private void text(final String node) {
    out.print(node);
  }

  /**
   * Prints the list of strings.
   *
   * @param delim the delimiter.
   * @param nodes the list.
   */
  private void text(final String delim, final List<String> nodes) {
    boolean comma = false;
    for (final String string : nodes) {
      if (comma) {
        text(delim);
      }

      comma = true;
      text(string);
    }
  }

  /**
   * Prints the expression.
   *
   * @param node the expression to be printed.
   */
  private void text(final VerilogExpression node) {
    text(ExprPrinter.VERILOG.toString(node.getNode()));

    if (debugComments) {
      text(String.format(" /* %s */", node.getNode()));
    }
  }

  /**
   * Prints the extended expression.
   *
   * @param node the extended expression to be printed.
   */
  private void text(final VerilogMinTypMax node) {
    if (node.isUsual()) {
      text(node.getMinExpression());
    } else {
      text(node.getMinExpression());
      text(":");
      text(node.getTypExpression());
      text(":");
      text(node.getMaxExpression());
    }
  }

  // Prints the range.
  private void text(final VerilogRange node) {
    final VerilogExpression lhs = node.getLeftExpression();
    final VerilogExpression rhs = node.getRightExpression();

    text("[");
    text(lhs);

    if (rhs != null) {
      switch (node.getType()) {
        case RANGE:
          text(":");
          break;
        case PLUS:
          text("+:");
          break;
        case MINUS:
          text("-:");
          break;
      }

      text(rhs);
    }

    text("]");
  }

  /**
   * Prints the list of ranges.
   *
   * @param node the list of ranges to be printed.
   */
  private void text(final List<VerilogRange> node) {
    for (final VerilogRange range : node) {
      text(range);
    }
  }

  /**
   * Prints the path
   *
   * @param node the path to be printed.
   */
  private void text(final VerilogPath node) {
    text(node.toString());
  }

  /**
   * Prints the reference.
   *
   * @param node the reference to be printed.
   */
  private void text(final VerilogReference node) {
    text(ExprPrinter.VERILOG.toString(node.getNode()));

    if (debugComments) {
      // Bits selection is not printed (it is included into expression as BVEXTRACT).
      final VerilogDeclaration declaration = node.getDeclaration();

      text(" /* DECL: ");
      text(declaration != null ? declaration.getName() : "null");
      text(" */");
    }
  }

  /**
   * Prints the assignment.
   *
   * @param type the assignment type.
   * @param node the assignment.
   */
  private void text(final String type, final VerilogAssignment node) {
    final List<VerilogReference> references = node.getReferences();

    if (references.size() > 1) {
      text("{");
    }

    textReferences(references);

    if (references.size() > 1) {
      text("}");
    }

    final VerilogMinTypMax expression = node.getRhsExpression();

    if (expression != null) {
      text(" " + type + " ");
      text(expression);
    }
  }

  /**
   * Prints the assignment.
   *
   * @param node the assignment to be printed.
   */
  private void text(final VerilogAssignment node) {
    text("=", node);
  }

  /**
   * Prints the event control.
   *
   * @param node the event control to be printed.
   */
  private void text(final VerilogEventControl node) {
    if (node.isDelays()) {
      text(node.getDelay());
    }

    if (node.isRepeat()) {
      text("repeat (");
      text(node.getRepeat());
      text(") ");
    }

    if (node.isEvents() || node.isRepeat()) {
      text("@(");

      boolean comma = false;
      for (final VerilogEvent event : node.getEvents()) {
        if (comma) {
          text(", ");
        }

        comma = true;
        text(event.getEdge());

        if (event.getEdge() != VerilogEdge.UNDEFINED) {
          text(" ");
        }

        text(event.getExpression());
      }

      text(")");
    }
  }

  /**
   * Prints the element type.
   *
   * @param node the element type to be printed.
   */
  private void text(final VerilogElementType node) {
    boolean printed = false;

    if (node.isVectored()) {
      text("vectored");
      printed = true;
    }

    if (node.isScalared()) {
      if (printed) {
        text(" ");
      }
      text("scalared");
      printed = true;
    }

    if (node.isSigned()) {
      if (printed) {
        text(" ");
      }
      text("signed");
      printed = true;
    }

    if (!node.isUndefined()) {
      if (printed) {
        text(" ");
      }
      printed = true;
    }

    switch (node.getType()) {
      case UNDEFINED: {
        break;
      }
      case INTEGER: {
        text("integer");
        break;
      }
      case REAL: {
        text("real");
        break;
      }
      case TIME: {
        text("time");
        break;
      }
      case REALTIME: {
        text("realtime");
        break;
      }
      case SUPPLY0: {
        text("supply0");
        break;
      }
      case SUPPLY1: {
        text("supply1");
        break;
      }
      case TRI: {
        text("tri");
        break;
      }
      case TRIAND: {
        text("triand");
        break;
      }
      case TRIOR: {
        text("trior");
        break;
      }
      case TRI0: {
        text("tri0");
        break;
      }
      case TRI1: {
        text("tri1");
        break;
      }
      case UWIRE: {
        text("uwire");
        break;
      }
      case WIRE: {
        text("wire");
        break;
      }
      case WAND: {
        text("wand");
        break;
      }
      case WOR: {
        text("wor");
        break;
      }
      case TRIREG: {
        text("trireg");
        break;
      }
      case REG: {
        text("reg");
        break;
      }
    }

    final VerilogRange range = node.getRange();

    if (range != null) {
      if (printed) {
        text(" ");
      }
      text(range);
      printed = true;
    }
  }

  /**
   * Prints the edge.
   *
   * @param node the edge to be printed.
   */
  private void text(final VerilogEdge node) {
    switch (node) {
      case POSITIVE: {
        text("posedge");
        break;
      }
      case NEGATIVE: {
        text("negedge");
        break;
      }
      default: {
        break;
      }
    }
  }

  /**
   * Prints the delay.
   *
   * @param node the delay to be printed.
   */
  private void text(final VerilogDelay node) {
    if (node.isSpecified()) {
      final List<VerilogMinTypMax> delays = node.getDelays();

      text("#");

      if (delays.size() > 1) {
        text("(");
      }
      {
        boolean comma = false;
        for (final VerilogMinTypMax expression : delays) {
          if (comma) {
            text(", ");
          }

          comma = true;
          text(expression);
        }
      }
      if (delays.size() > 1) {
        text(")");
      }
    }
  }

  private void printPorts(final List<SvaPort> ports) {
    boolean comma = false;
    text("(");
    for (SvaPort port : ports) {
      if (comma) {
        text(", ");
      }

      if (port.isLocal()) {
        text("local ");
      }

      if (port.getDirection() != null) {
        switch (port.getDirection()) {
          case INPUT:
            text("input ");
            break;
          case OUTPUT:
            text("output ");
            break;
          case INOUT:
            text("inout ");
            break;
        }
      }

      if (port.getType() != null) {
        text(String.format("%s ", port.getType().toString()));
      }

      text(String.format("%s", port.getName()));

      for (VerilogRange dimension : port.getDimensions()) {
        text(dimension.toString());
      }

      if (port.getArgument() != null) {
        text(String.format(" = %s", port.getArgument().toString()));
      }

      comma = true;
    }
    text(")");
  }
}
