/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva;

import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogStatement;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.sva.basis.SvaProperty;

import java.util.EnumSet;

/**
 * {@link SvaAssertionStatement} represents the SystemVerilog Assertion statement.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaAssertionStatement extends VerilogStatement {

  public static final Tag TAG = Tag.ASSERTION_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.SVA_INSTANCE);


  /**
   * Statement types.
   */
  public enum StatementType {
    CONCURRENT,
    IMMEDIATE
  }


  /**
   * Assertion kinds.
   */
  public enum AssertionKind {
    ASSERT,
    ASSUME
  }

  /** Assertion statement type. */
  private StatementType type;

  /** Assertion kind. */
  private AssertionKind kind;

  /** Immediate assertion expression. */
  private VerilogExpression expression;

  /** Concurrent assertion property. */
  private SvaProperty property;

  /** Assertion is procedural. */
  private boolean procedural = false;

  /**
   * Creates a new assertion statement.
   *
   * @param parent - the specified parent node.
   */
  public SvaAssertionStatement(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Copy constructor.
   *
   * @param other - the specified statement to copy.
   * @param parent - the specified parent node.
   */
  private SvaAssertionStatement(final SvaAssertionStatement other, final VerilogNode parent) {
    super(other, parent);
    this.type = other.type;
    this.kind = other.kind;
    this.expression = other.expression != null ? other.expression.clone() : null;
    this.property = other.property != null ? other.property.clone() : null;
    this.procedural = other.procedural;
  }

  @Override
  public VerilogStatement clone() {
    return new SvaAssertionStatement(this, this.getParentNode());
  }

  /**
   * Sets the assertion kind.
   *
   * @param kind - the specified assertion kind.
   */
  public void setKind(final AssertionKind kind) {
    this.kind = kind;
  }

  /**
   * Sets the statement type.
   *
   * @param type - the specified statement type.
   */
  public void setType(final StatementType type) {
    this.type = type;
  }

  /**
   * Sets the immediate assertion expression.
   *
   * @param expression - the specified {@link VerilogExpression}.
   */
  public void setExpression(final VerilogExpression expression) {
    this.expression = expression;
  }

  /**
   * Returns the immediate assertion expression.
   *
   * @return the assertion expression.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the concurrent assertion property.
   *
   * @param property - the specified {@link SvaProperty}.
   */
  public void setProperty(final SvaProperty property) {
    this.property = property;
  }

  /**
   * Returns the concurrent assertion property.
   *
   * @return the assertion property.
   */
  public SvaProperty getProperty() {
    return property;
  }

  /**
   * Shows whether the assertion is concurrent.
   *
   * @return {@code true} if the assertion is concurrent, {@code false} otherwise.
   */
  public boolean isConcurrent() {
    return type == StatementType.CONCURRENT;
  }

  /**
   * Shows whether the assertion is immediate.
   *
   * @return {@code true} if the assertion is immediate, {@code false} otherwise.
   */
  public boolean isImmediate() {
    return type == StatementType.IMMEDIATE;
  }

  /**
   * Shows whether the statement is 'assert'.
   *
   * @return {@code true} if the statement is 'assert', {@code false} otherwise.
   */
  public boolean isAssert() {
    return kind == AssertionKind.ASSERT;
  }

  /**
   * Shows whether the statement is 'assume'.
   *
   * @return {@code true} if the statement is 'assume', {@code false} otherwise.
   */
  public boolean isAssume() {
    return kind == AssertionKind.ASSUME;
  }

  /**
   * Shows whether the statement is located in the procedural code.
   *
   * @return {@code true} if the statement is procedural, {@code false} otherwise.
   */
  public boolean isProcedural() {
    return procedural;
  }

  /**
   * Makes the statement procedural.
   */
  public void setProcedural() {
    this.procedural = true;
  }
}
