/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.basis.VerilogRange;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link SvaDeclaration} represents the declaration of a property or a sequence.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public abstract class SvaDeclaration extends VerilogNode {

  /** Ports. */
  protected List<SvaPort> ports;

  /** Variable declarations. */
  protected List<VerilogDeclaration> variables;

  protected SvaDeclaration(final Tag tag, final EnumSet<Tag> childrenTags, final NodeKind kind,
      final VerilogNode parent) {
    super(tag, childrenTags, kind, parent);
    ports = new ArrayList<>();
    variables = new ArrayList<>();
  }

  protected SvaDeclaration(final SvaDeclaration other, final VerilogNode parent) {
    super(other, parent);
    ports = new ArrayList<>();
    variables = new ArrayList<>();
    for (SvaPort port : other.ports) {
      //addPort(port.clone());
      this.ports.add(port.clone());
    }
    for (VerilogDeclaration variable : other.variables) {
      //addDeclaration(variable.clone());
      this.variables.add(variable.clone());
    }
  }

  /**
   * Returns the property variables.
   *
   * @return the list of property variables.
   */
  public List<VerilogDeclaration> getVariables() {
    return variables;
  }

  /**
   * Adds a variable to the property variable list.
   *
   * @param variable - the specified variable declaration.
   */
  public void addDeclaration(final VerilogDeclaration variable) {
    add(variable);
    this.variables.add(variable);
  }

  /**
   * Returns the property ports.
   *
   * @return the list of ports.
   */
  public List<SvaPort> getPorts() {
    return ports;
  }

  /**
   * Adds a port to the property ports list.
   *
   * @param port - the specified port.
   */
  public void addPort(final SvaPort port) {
    InvariantChecks.checkNotNull(port);
    this.ports.add(port);

    final VerilogDeclaration declaration = new VerilogDeclaration(this);
    declaration.setName(port.getName());
    if (port.getType() != null) {
      declaration.setElementType(port.getType());
    }
    for (VerilogRange range : port.getDimensions()) {
      declaration.addDimension(range);
    }
    add(declaration);
  }

}
