/*
 * Copyright 2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.sva.basis.SvaPortConnection;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link SvaInstantiation} represents the instance of a sequence or a property.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class SvaInstantiation extends VerilogNode {

  public static final Tag TAG = Tag.SVA_INSTANCE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** Path to an object. */
  private VerilogPath path;

  /** Instance node. */
  private Node node;

  /** List of port connections. */
  private List<SvaPortConnection> connections;

  /** Sequence/property declaration. */
  private SvaDeclaration declaration;

  public SvaInstantiation(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
    connections = new ArrayList<>();
  }

  public SvaInstantiation(final SvaInstantiation other, final VerilogNode parent) {
    super(other, parent);
    connections = new ArrayList<>();
    for (final SvaPortConnection connection : other.connections) {
      addConnection(connection.clone());
    }
    this.declaration = other.declaration;
    this.node = other.node != null ? other.node.deepestCopy() : null;
    this.path = other.path != null ? other.path.clone() : null;
  }

  @Override
  public VerilogNode clone() {
    return new SvaInstantiation(this, getParentNode());
  }

  /**
   * Returns the reference path.
   *
   * @return the path.
   */
  public VerilogPath getPath() {
    return path;
  }

  /**
   * Sets the reference path.
   *
   * @param path the path to be set.
   */
  public void setPath(final VerilogPath path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  /**
   * Returns the associated instance node.
   *
   * @return the node.
   */
  public Node getNode() {
    return node;
  }

  /**
   * Sets the instance node.
   *
   * @param node - the specified node.
   */
  public void setNode(final Node node) {
    InvariantChecks.checkNotNull(node);
    this.node = node;
  }

  /**
   * Adds the port connection to the reference.
   *
   * @param connection - the specified port connection.
   */
  public void addConnection(final SvaPortConnection connection) {
    InvariantChecks.checkNotNull(connection);
    connections.add(connection);
  }

  /**
   * Returns the list of port connections.
   *
   * @return the list of port connections.
   */
  public List<SvaPortConnection> getConnections() {
    return connections;
  }

  /**
   * Returns the sequence/property declaration.
   *
   * @return the declaration node.
   */
  public SvaDeclaration getDeclaration() {
    return declaration;
  }

  /**
   * Sets the sequence/property declaration.
   *
   * @param declaration - the specified declaration node.
   */
  public void setDeclaration(final SvaDeclaration declaration) {
    this.declaration = declaration;
  }

  /**
   * Returns the connection arguments as a list of {@link Node}.
   *
   * @return the connection arguments.
   */
  public List<Node> getConnectionNodes() {
    final ArrayList<Node> connectionNodes = new ArrayList<>();

    for (SvaPortConnection connection : connections) {
      if (connection.isExpression()) {
        connectionNodes.add(((VerilogExpression) connection.getArgument()).getNode());
      }
      if (connection.isEvent()) {
        connectionNodes.add(((VerilogEvent) connection.getArgument()).getExpression().getNode());
      }
    }
    return connectionNodes;
  }
}
