/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.VerilogTranslator;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogRange;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link SvaPort} represents the SVA sequence/property port.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaPort extends VerilogNode {

  public static final Tag TAG = Tag.SVA_PORT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);


  /**
   * Port direction types.
   */
  public enum Direction {
    INPUT,
    OUTPUT,
    INOUT
  }

  /** Port type. */
  private VerilogElementType type;

  /** Port direction type. */
  private Direction direction;

  /** Local flag. */
  private boolean isLocal = false;

  /** Port dimensions. */
  private List<VerilogRange> dimensions;

  /** Default port argument. */
  private Object argument;

  /**
   * Creates a new port.
   *
   * @param parent - the specified parent node.
   */
  public SvaPort(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
    dimensions = new ArrayList<>();
  }

  /**
   * Creates a new port using the other.
   *
   * @param other - the specified port to copy.
   * @param parent - the specified parent node.
   */
  public SvaPort(final SvaPort other, final VerilogNode parent) {
    super(other, parent);
    this.type = other.type != null ? other.type.clone() : null;
    this.direction = other.direction;
    this.isLocal = other.isLocal;
    this.argument = other.argIsEvent()
        ? ((VerilogEvent) other.argument).clone()
        : other.argIsExpression() ? ((VerilogExpression) other.argument).clone() : null;
    dimensions = new ArrayList<>();
    for (VerilogRange range : other.dimensions) {
      this.addDimension(range.clone());
    }
  }

  @Override
  public SvaPort clone() {
    return new SvaPort(this, this.getParentNode());
  }

  /**
   * Returns the port type.
   *
   * @return the port type.
   */
  public VerilogElementType getType() {
    return type;
  }

  /**
   * Sets the port type.
   *
   * @param type - the specified port type.
   */
  public void setType(final VerilogElementType type) {
    this.type = type;
  }

  /**
   * Returns the port direction.
   *
   * @return the port direction.
   */
  public Direction getDirection() {
    return direction;
  }

  /**
   * Sets the port direction.
   *
   * @param direction - the specified port direction.
   */
  public void setDirection(final Direction direction) {
    this.direction = direction;
  }

  /**
   * Returns the port dimensions.
   *
   * @return the list of port dimensions.
   */
  public List<VerilogRange> getDimensions() {
    return dimensions;
  }

  /**
   * Adds a dimension to the port.
   *
   * @param dimension - the specified dimension.
   */
  public void addDimension(final VerilogRange dimension) {
    InvariantChecks.checkNotNull(dimension);
    this.dimensions.add(dimension);
  }

  /**
   * Returns the default port argument.
   *
   * @return the default port argument.
   */
  public Object getArgument() {
    return argument;
  }

  /**
   * Sets the default port argument.
   *
   * @param argument - the specified port argument.
   */
  public void setArgument(final Object argument) {
    InvariantChecks.checkNotNull(argument);
    if (direction == Direction.INOUT) {
      VerilogLogger.error("Default argument is illegal for inout port: %s", getName());
      VerilogTranslator.exit();
    }

    if (type == null) {
      VerilogLogger.error("Port argument type must be specified explicitly: %s", getName());
      VerilogTranslator.exit();
    }

    this.argument = argument;
  }

  /**
   * Shows whether the default argument is an event.
   *
   * @return {@code true} if the default argument is an event, {@code false} otherwise.
   */
  public boolean argIsEvent() {
    return argument instanceof VerilogEvent;
  }

  /**
   * Shows whether the default port argument is an expression.
   *
   * @return {@code true} if the default port argument is an expression, {@code false} otherwise.
   */
  public boolean argIsExpression() {
    return argument instanceof VerilogExpression;
  }

  /**
   * Shows whether the port is local.
   *
   * @return {@code true} if the port is local, {@code false} otherwise.
   */
  public boolean isLocal() {
    return isLocal;
  }

  /**
   * Makes the port local.
   */
  public void setLocal() {
    isLocal = true;
  }
}
