/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

/**
 * {@link SvaPortConnection} represents the property/sequence port connection.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaPortConnection {

  /** Port name. */
  private String name;

  /** Connected argument. */
  private Object argument;

  /**
   * Creates a new port connection.
   */
  public SvaPortConnection() {
  }

  /**
   * Copy constructor.
   *
   * @param other - the {@link SvaPortConnection} to copy.
   */
  private SvaPortConnection(final SvaPortConnection other) {
    InvariantChecks.checkNotNull(other);
    this.name = other.name;
    this.argument = other.isExpression()
        ? ((VerilogExpression) other.argument).clone()
        : other.isEvent() ? ((VerilogEvent) other.argument).clone() : null;
  }

  /**
   * Returns the port name.
   *
   * @return the port name.
   */
  public String getName() {
    return name;
  }

  /**
   * Sets the port name.
   *
   * @param name - the specified name.
   */
  public void setName(final String name) {
    this.name = name;
  }

  /**
   * Returns the argument connected to the port.
   *
   * @return the port argument.
   */
  public Object getArgument() {
    return argument;
  }

  /**
   * Sets the connected argument.
   *
   * @param argument - the specified argument.
   */
  public void setArgument(final Object argument) {
    InvariantChecks.checkNotNull(argument);
    this.argument = argument;
  }

  /**
   * Checks whether the connected argument is an expression.
   *
   * @return {@code true} if the connected argument is an expression, {@code false} otherwise.
   */
  public boolean isExpression() {
    return argument instanceof VerilogExpression;
  }

  /**
   * Checks whether the connected argument is an event.
   *
   * @return {@code true} if the connected argument is an event, {@code false} otherwise.
   */
  public boolean isEvent() {
    return argument instanceof VerilogEvent;
  }

  @Override
  public SvaPortConnection clone() {
    return new SvaPortConnection(this);
  }

  @Override
  public String toString() {
    if (name != null) {
      return String.format(".%s(%s)", name, argument);
    } else {
      return argument.toString();
    }
  }
}
