/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link SvaProperty} represents the SVA property specification.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaProperty {

  /** Property triggering clocking event. */
  private List<VerilogEvent> event;

  /** Property disabling expression. */
  private VerilogExpression disableExpression;

  /** Property expression. */
  private VerilogExpression property;

  /**
   * Creates a new property.
   */
  public SvaProperty() {
  }

  /**
   * Copy constructor.
   *
   * @param other - the specified property to copy.
   */
  private SvaProperty(final SvaProperty other) {
    InvariantChecks.checkNotNull(other);
    this.disableExpression =
        other.disableExpression != null ? other.disableExpression.clone() : null;
    this.property = other.property != null ? other.property.clone() : null;
    event = new ArrayList<>();
    if (other.event != null) {
      for (VerilogEvent otherEvent : other.event) {
        event.add(otherEvent.clone());
      }
    }
  }

  /**
   * Returns the clocking events.
   *
   * @return the list of {@link VerilogEvent}.
   */
  public List<VerilogEvent> getEvent() {
    return event;
  }

  /**
   * Sets the clocking event.
   *
   * @param event - the specified list of {@link VerilogEvent}.
   */
  public void setEvent(final List<VerilogEvent> event) {
    InvariantChecks.checkNotNull(event);
    this.event = event;
  }

  /**
   * Returns the property disabling expression.
   *
   * @return the disabling expression.
   */
  public VerilogExpression getDisableExpression() {
    return disableExpression;
  }

  /**
   * Sets the disabling expression.
   *
   * @param disableExpression - the specified expression.
   */
  public void setDisableExpression(final VerilogExpression disableExpression) {
    InvariantChecks.checkNotNull(disableExpression);
    this.disableExpression = disableExpression;
  }

  /**
   * Returns the property expression.
   *
   * @return the property expression.
   */
  public VerilogExpression getExpression() {
    return property;
  }

  /**
   * Sets the property expression.
   *
   * @param property - the specified property expression.
   */
  public void setExpression(final VerilogExpression property) {
    InvariantChecks.checkNotNull(property);
    this.property = property;
  }

  @Override
  public SvaProperty clone() {
    return new SvaProperty(this);
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    if (event != null && !event.isEmpty()) {
      builder.append(String.format("%s ", event.iterator().next().toString()));
    }
    if (disableExpression != null) {
      builder.append(String.format("%s ", disableExpression.toString()));
    }
    if (property != null) {
      builder.append(String.format("%s ", property.toString()));
    }

    return builder.toString();
  }

}
