/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.basis;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogRange;

/**
 * {@link SvaRange} represents the ranges used in SVA expressions.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class SvaRange {

  /** Verilog-style range. */
  private VerilogRange range;

  /** Infinity flag. */
  private boolean unbounded;

  /** Node representation of a range. */
  private Node node;

  /**
   * Creates a new range.
   */
  public SvaRange() {
    unbounded = false;
  }

  /**
   * Copy constructor.
   *
   * @param other - the specified range to copy.
   */
  protected SvaRange(final SvaRange other) {
    InvariantChecks.checkNotNull(other);
    this.unbounded = other.unbounded;
    this.range = other.range != null ? other.range.clone() : null;
    this.node = other.node;
  }

  /**
   * Sets the {@link VerilogRange} of this range.
   *
   * @param range - the specified range.
   */
  public void setRange(final VerilogRange range) {
    InvariantChecks.checkNotNull(range);
    this.range = range;
  }

  /**
   * Returns the left expression of the range.
   *
   * @return the left expression.
   */
  public VerilogExpression getMin() {
    return range != null ? range.getLeftExpression() : null;
  }

  /**
   * Sets the left expression of the range.
   *
   * @param min - the specified expression.
   */
  public void setMin(final VerilogExpression min) {
    if (range == null) {
      range = new VerilogRange();
    }
    range.setLeftExpression(min);
  }

  /**
   * Returns the right expression of the range.
   *
   * @return the right expression.
   */
  public VerilogExpression getMax() {
    return isUnbounded() | range == null ? null : range.getRightExpression();
  }

  /**
   * Sets the right expression of the range.
   *
   * @param max - the specified expression.
   */
  public void setMax(final VerilogExpression max) {
    if (range == null) {
      range = new VerilogRange();
    }
    range.setRightExpression(max);
  }

  /**
   * Shows whether this range is unbounded.
   *
   * @return {@code true} if the range is unbounded, {@code false} otherwise.
   */
  public boolean isUnbounded() {
    return unbounded;
  }

  /**
   * Makes this range unbounded.
   */
  public void setUnbounded() {
    this.unbounded = true;
  }

  @Override
  public SvaRange clone() {
    return new SvaRange(this);
  }

  /**
   * Returns the {@link Node} representation of the range.
   *
   * @return the corresponding node.
   */
  public Node getNode() {
    return node;
  }

  /**
   * Sets the {@link Node} representation of the range.
   *
   * @param node - the specified node.
   */
  public void setNode(final Node node) {
    InvariantChecks.checkNotNull(node);
    this.node = node;
  }
}
