/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.expression;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.LinkedList;
import java.util.List;

/**
 * {@link SvaCaseExpression} represents the SVA case expression.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaCaseExpression extends VerilogExpression {

  /** Case condition. */
  private VerilogExpression caseExpression;

  /** Case alternatives. */
  private List<SvaCaseExpressionItem> items;

  /**
   * Creates a new case expression.
   */
  public SvaCaseExpression() {
    items = new LinkedList<>();
  }

  /**
   * Copy constructor.
   *
   * @param other - the expression to copy.
   */
  private SvaCaseExpression(final SvaCaseExpression other) {
    super(other);
    this.caseExpression = other.caseExpression != null ? other.caseExpression.clone() : null;
    items = new LinkedList<>();
    for (SvaCaseExpressionItem item : other.items) {
      addItem(item.clone());
    }
  }

  /**
   * Returns the case condition.
   *
   * @return the case condition.
   */
  public VerilogExpression getCaseExpression() {
    return caseExpression;
  }

  /**
   * Sets the case condition.
   *
   * @param caseExpression - the specified expression.
   */
  public void setCaseExpression(final VerilogExpression caseExpression) {
    InvariantChecks.checkNotNull(caseExpression);
    this.caseExpression = caseExpression;
  }

  /**
   * Returns the case alternatives.
   *
   * @return the list of alternatives.
   */
  public List<SvaCaseExpressionItem> getItems() {
    return items;
  }

  /**
   * Adds an alternative to the case.
   *
   * @param item - the specified case item.
   */
  public void addItem(final SvaCaseExpressionItem item) {
    InvariantChecks.checkNotNull(item);
    items.add(item);
  }

  @Override
  public SvaCaseExpression clone() {
    return new SvaCaseExpression(this);
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();

    builder.append(String.format("case (%s)%n", caseExpression.toString()));
    for (SvaCaseExpressionItem item : items) {
      builder.append(String.format("  %s%n", item.toString()));
    }
    builder.append("endcase\n");

    return builder.toString();
  }
}
