/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.expression;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.HashSet;
import java.util.Set;

/**
 * {@link SvaCaseExpressionItem} represents the SVA case expression alternative.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaCaseExpressionItem {

  /** Set of values. */
  private Set<VerilogExpression> values;

  /** Alternative's property expression. */
  private VerilogExpression property;

  /** Default flag. */
  private boolean isDefault;

  /**
   * Creates a new case alternative.
   */
  public SvaCaseExpressionItem() {
    values = new HashSet<>();
  }

  /**
   * Copy constructor.
   *
   * @param other - the item to copy.
   */
  private SvaCaseExpressionItem(final SvaCaseExpressionItem other) {
    InvariantChecks.checkNotNull(other);
    this.isDefault = other.isDefault;
    this.property = other.property != null ? other.property.clone() : null;
    values = new HashSet<>();
    for (VerilogExpression value : other.values) {
      addValue(value.clone());
    }
  }

  /**
   * Returns the alternative's values.
   *
   * @return the set of values.
   */
  public Set<VerilogExpression> getValues() {
    return values;
  }

  /**
   * Adds a value to the set of values.
   *
   * @param value - the specified value.
   */
  public void addValue(final VerilogExpression value) {
    InvariantChecks.checkNotNull(value);
    this.values.add(value);
  }

  /**
   * Returns the alternative's property.
   *
   * @return the property expression.
   */
  public VerilogExpression getProperty() {
    return property;
  }

  /**
   * Sets the property expression.
   *
   * @param property - the specified property expression.
   */
  public void setProperty(final VerilogExpression property) {
    this.property = property;
  }

  /**
   * Shows whether the alternative is default.
   *
   * @return {@code true} if the alternative is default, {@code false} otherwise.
   */
  public boolean isDefault() {
    return isDefault;
  }

  /**
   * Makes the alternative default.
   */
  public void setDefault() {
    isDefault = true;
  }

  /**
   * Clones this case item.
   *
   * @return the copy of this case item.
   */
  @Override
  public SvaCaseExpressionItem clone() {
    return new SvaCaseExpressionItem(this);
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    if (isDefault()) {
      builder.append("default: ");
    } else {
      boolean comma = false;
      for (VerilogExpression value : values) {
        if (comma) {
          builder.append(", ");
        }
        builder.append(value.toString());
        comma = true;
      }
      builder.append(": ");
    }
    builder.append(String.format("%s;", property.toString()));
    return builder.toString();
  }
}
