/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.expression;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.fortress.util.Pair;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.VerilogTranslator;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.sva.basis.SvaRange;
import ru.ispras.verilog.parser.sva.basis.SvaTemporalOperations;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * {@link SvaDelayedExpression} represents the SVA sequence of expressions separated by delays.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaDelayedExpression extends VerilogExpression {

  /** List of delay-expression pairs. */
  private LinkedList<Pair<SvaRange, VerilogExpression>> elements;

  /**
   * Creates a new expression.
   */
  public SvaDelayedExpression() {
    elements = new LinkedList<>();
  }

  /**
   * Copy constructor.
   *
   * @param other - the expression to copy.
   */
  private SvaDelayedExpression(final SvaDelayedExpression other) {
    super(other);
    elements = new LinkedList<>();
    for (Pair<SvaRange, VerilogExpression> element : other.getElements()) {
      this.addElement(element.first == null ? null : element.first.clone(), element.second.clone());
    }
  }

  /**
   * Adds a delay-expression pair to the sequence.
   *
   * @param delay - the specified delay range.
   * @param sequence - the specified expression.
   */
  public void addElement(final SvaRange delay, final VerilogExpression sequence) {
    InvariantChecks.checkNotNull(sequence);
    elements.add(new Pair<>(delay, sequence));
  }

  /**
   * Returns the elements of the sequence expression.
   *
   * @return the list of delay-expression pairs.
   */
  public List<Pair<SvaRange, VerilogExpression>> getElements() {
    return elements;
  }

  /**
   * Computes the node representation of the sequence.
   */
  public void computeNode() {
    final Iterator<Pair<SvaRange, VerilogExpression>> iterator = elements.descendingIterator();

    if (iterator.hasNext()) {

      Pair<SvaRange, VerilogExpression> prevElem = iterator.next();
      Node node = prevElem.second.getNode();
      while (iterator.hasNext()) {
        final Pair<SvaRange, VerilogExpression> currentElem = iterator.next();
        node = new NodeOperation(SvaTemporalOperations.DELAY, prevElem.first.getNode(),
              currentElem.second.getNode(), node);
        prevElem = currentElem;
      }

      // If a sequence starts with ##
      if (prevElem.first != null) {
        node = new NodeOperation(SvaTemporalOperations.DELAY, prevElem.first.getNode(), node);
      }

      setNode(node);
    } else {
      VerilogLogger.error("Temporal sequence has no values!");
      VerilogTranslator.exit();
    }
  }

  @Override
  public SvaDelayedExpression clone() {
    return new SvaDelayedExpression(this);
  }

}
