/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.expression;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.sva.basis.SvaTemporalOperations;

import java.util.LinkedList;
import java.util.List;

/**
 * {@link SvaOperatorExpression} represents the SVA operator-based expression.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class SvaOperatorExpression extends VerilogExpression {

  /** Operator. */
  protected SvaTemporalOperations operator;

  /** Operands. */
  protected List<VerilogExpression> operands;

  /**
   * Creates a new operator-based expression.
   */
  public SvaOperatorExpression() {
    operands = new LinkedList<>();
  }

  /**
   * Creates a new operator-based expression using the other.
   *
   * @param other - the specified expression to copy.
   */
  protected SvaOperatorExpression(final SvaOperatorExpression other) {
    super(other);
    this.operator = other.operator;
    operands = new LinkedList<>();
    for (VerilogExpression operand : other.operands) {
      addOperand(operand.clone());
    }
  }

  /**
   * Returns the expression's operator.
   *
   * @return the operator.
   */
  public SvaTemporalOperations getOperator() {
    return operator;
  }

  /**
   * Sets the expression's operator.
   *
   * @param operator - the specified operator.
   */
  public void setOperator(final SvaTemporalOperations operator) {
    this.operator = operator;
  }

  /**
   * Adds an operand to the expression.
   *
   * @param operand - the specified operand.
   */
  public void addOperand(final VerilogExpression operand) {
    InvariantChecks.checkNotNull(operand);
    operands.add(operand);
  }

  /**
   * Returns the operand by index.
   *
   * @param index - the specified operand's index.
   * @return the operand.
   */
  public VerilogExpression getOperand(final int index) {
    return operands.get(index);
  }

  /**
   * Returns the operands of the expression.
   *
   * @return the list of operands.
   */
  public List<VerilogExpression> getOperands() {
    return operands;
  }

  @Override
  public SvaOperatorExpression clone() {
    return new SvaOperatorExpression(this);
  }

}
