/*
 * Copyright 2019-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.expression;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.sva.basis.SvaRepetition;
import ru.ispras.verilog.parser.sva.basis.SvaSequenceMatchAction;

import java.util.LinkedList;
import java.util.List;

/**
 * {@link SvaRepeatedExpression} represents the SVA repeated expression.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaRepeatedExpression extends VerilogExpression {

  /** Expression. */
  private VerilogExpression expression;

  /** Expression repetition. */
  private SvaRepetition repetition;

  /** Match actions.*/
  private List<SvaSequenceMatchAction> actions;

  /**
   * Creates a new repeated expression.
   */
  public SvaRepeatedExpression() {
    actions = new LinkedList<>();
  }

  /**
   * Copy constructor.
   *
   * @param other - the expression to copy.
   */
  private SvaRepeatedExpression(final SvaRepeatedExpression other) {
    super(other);
    this.expression = other.expression != null ? other.expression.clone() : null;
    this.repetition = other.repetition != null ? other.repetition.clone() : null;
    actions = new LinkedList<>();
    for (SvaSequenceMatchAction action : other.actions) {
      addMatchAction(action.clone());
    }
  }

  /**
   * Returns the expression that is repeated.
   *
   * @return the expression.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the expression that is repeated.
   *
   * @param expression - the specified expression.
   */
  public void setExpression(final VerilogExpression expression) {
    this.expression = expression;
  }

  /**
   * Returns the repetition.
   *
   * @return the repetition.
   */
  public SvaRepetition getRepetition() {
    return repetition;
  }

  /**
   * Sets the repetition.
   *
   * @param repetition - the specified repetition.
   */
  public void setRepetition(final SvaRepetition repetition) {
    InvariantChecks.checkNotNull(repetition);
    this.repetition = repetition;
  }

  /**
   * Returns the match actions.
   *
   * @return the list of {@link SvaSequenceMatchAction}.
   */
  public List<SvaSequenceMatchAction> getMatchActions() {
    return actions;
  }

  /**
   * Adds a match action to the expression.
   *
   * @param action - the specified match action.
   */
  public void addMatchAction(final SvaSequenceMatchAction action) {
    InvariantChecks.checkNotNull(action);
    this.actions.add(action);
  }

  @Override
  public void setNode(final Node node) {
    if (repetition != null) {
      super.setNode(node);
    } else {
      InvariantChecks.checkNotNull(node);
      final Object oldUserData = node.getUserData();
      super.setNode(node);
      // Set the old user data for node that is passed from another rule.
      node.setUserData(oldUserData);
    }
  }

  @Override
  public SvaRepeatedExpression clone() {
    return new SvaRepeatedExpression(this);
  }

}
