/*
 * Copyright 2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.sva.util;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.sva.SvaDeclaration;
import ru.ispras.verilog.parser.sva.SvaInstantiation;
import ru.ispras.verilog.parser.sva.basis.SvaPortConnection;
import ru.ispras.verilog.parser.sva.basis.SvaTemporalOperations;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Utils for SVA syntax tree building.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SvaUtils {

  private static Map<String, Integer> svaInstances = new HashMap<>();

  /*
   * Searches for a sequence/property declaration.
   */
  public static SvaDeclaration findSvaDeclaration(final String name, final VerilogNode parent) {
    final SvaDeclaration seqDecl =
        (SvaDeclaration) parent.findRecursively(name, VerilogNode.Tag.SEQUENCE_DECLARATION);
    final SvaDeclaration propDecl =
        (SvaDeclaration) parent.findRecursively(name, VerilogNode.Tag.PROPERTY_DECLARATION);
    return seqDecl != null ? seqDecl : propDecl;
  }

  /*
   * Creates the sequence/property instance's node representation.
   */
  public static Node createSvaInstNode(final String path, final VerilogNode declaration,
      final SvaInstantiation instance, final Collection<Node> nodes) {

    final Collection<Node> args;
    if (nodes == null) {
      args = new ArrayList<>();
      args.add(NodeValue.newUnknown("<" + path + ">"));
      args.addAll(instance.getConnectionNodes());
    } else {
      args = nodes;
    }
    SvaTemporalOperations op = null;
    if (declaration.isSequenceDeclaration()) {
      op = SvaTemporalOperations.SEQ_INSTANCE;
    }
    if (declaration.isPropertyDeclaration()) {
      op = SvaTemporalOperations.PROP_INSTANCE;
    }
    return new NodeOperation(op, args.toArray(new Node[0]));
  }

  /*
   * Creates the sequence/property AST node.
   */
  public static SvaInstantiation createSvaInstance(final VerilogNode parent, final String path,
      final SvaDeclaration declaration, final Collection<SvaPortConnection> connections,
      final Collection<Node> nodes) {
    final SvaInstantiation instance = new SvaInstantiation(parent);

    instance.setDeclaration(declaration);
    instance.setPath(new VerilogPath(new VerilogPathItem(path)));
    if (connections != null) {
      for (final SvaPortConnection connection : connections) {
        instance.addConnection(connection);
      }
    }

    final Node node = createSvaInstNode(path, declaration, instance, nodes);
    instance.setNode(node);

    // Avoid name similarity.
    int instNum = 0;
    String name = path;
    while (svaInstances.containsKey(name)) {
      instNum = svaInstances.get(name) + 1;
      name = path + "_" + instNum;
    }
    svaInstances.put(path, instNum);
    instance.setName(name);
    parent.add(instance);
    return instance;
  }

  /*
   * Creates the sequence/property Verilog expression.
   */
  public static VerilogExpression createExpressionAndInstance(final VerilogNode parent,
      final VerilogPath path, final SvaDeclaration declaration,
      final Collection<SvaPortConnection> connections, final Collection<Node> nodes) {

    final SvaInstantiation instance =
        createSvaInstance(parent, path.toString(), declaration, connections, nodes);
    return new VerilogExpression(instance.getNode());
  }

  public static VerilogExpression createExpressionAndInstance(final VerilogNode parent,
      final VerilogPath path, final SvaDeclaration declaration) {
    return createExpressionAndInstance(parent, path, declaration, null, null);
  }

  public static VerilogExpression createExpressionAndInstance(final VerilogNode parent,
      final VerilogPath path, final SvaDeclaration declaration,
      final Collection<SvaPortConnection> connections) {
    return createExpressionAndInstance(parent, path, declaration, connections, null);
  }

  /*
   * Adds connections to the reference.
   */
  public static Collection<SvaPortConnection> getConnections(
      final List<VerilogExpression> expressions) {
    final Collection<SvaPortConnection> connections = new ArrayList<>();
    for (final VerilogExpression expression : expressions) {
      final SvaPortConnection connection = new SvaPortConnection();
      connection.setArgument(expression);
      connections.add(connection);
    }
    return connections;
  }

}
