/*
 * Copyright 2018-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.transformer;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeProvider;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.elaborator.VerilogBinding;
import ru.ispras.verilog.parser.elaborator.VerilogContext;
import ru.ispras.verilog.parser.elaborator.VerilogDescriptor;
import ru.ispras.verilog.parser.elaborator.VerilogParameter;
import ru.ispras.verilog.parser.interpreter.VerilogCalculator;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.util.IndexOutOfBoundStructure;

import java.util.Map;

/**
 * {@link VerilogTransformerVariableSubstitute} implements an engine that substitutes variables with
 * expressions taking into account bit ranges.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTransformerVariableSubstitute extends VerilogTransformer {

  /** All elaborated variables. */
  private final Map<String, VerilogDescriptor> variables;
  /** Contains bindings to be applied. */
  private final VerilogContext context;
  /** Contains parameters to be applied. */
  private final Map<String, VerilogParameter> parameters;

  public VerilogTransformerVariableSubstitute(
      final Map<String, VerilogDescriptor> variables,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(context);
    InvariantChecks.checkNotNull(parameters);

    this.variables = variables;
    this.context = context;
    this.parameters = parameters;
  }

  public static Node transform(
      final Node node,
      final Map<String, VerilogDescriptor> variables,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(context);
    InvariantChecks.checkNotNull(parameters);

    final VerilogTransformer transformer =
        new VerilogTransformerVariableSubstitute(variables, context, parameters);
    return transformer.transform(node);
  }

  public static void transform(
      final VerilogNode node,
      final Map<String, VerilogDescriptor> variables,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(context);
    InvariantChecks.checkNotNull(parameters);

    final VerilogTransformer transformer =
        new VerilogTransformerVariableSubstitute(variables, context, parameters);
    transformer.run(node);
  }

  @Override
  public Node transform(final Node node) {
    final Map<String, VerilogBinding> bindings = context.getBindings();

    // Before substituting the variables, ranges should be updated.
    final Node node1 = shiftRanges(node);

    return Transformer.substitute(node1, (NodeProvider) variable -> {
      final Node newNode;

      final String localName = variable.getName();
      final VerilogBinding binding = bindings.get(localName);

      if (binding != null) {
        // Apply the binding.
        switch (binding.getType()) {
          case CURRENT_CONTEXT:
            // Current bindings should be applied.
            newNode = transform(binding.getRhs(), variables, context, parameters);
            break;
          case PARENT_CONTEXT:
            // Parent bindings should be applied.
            newNode = transform(binding.getRhs(), variables, context.getParentContext(),
                parameters);
            break;
          default:
            // Expression is ready to use.
            newNode = shiftRanges(binding.getRhs());
            break;
        }
      } else {
        final String fullName = context.getFullName(localName);
        final VerilogDescriptor descriptor = variables.get(fullName);

        if (descriptor != null) {
          // Access is done via a qualified name not presented in the bindings.
          newNode = descriptor.getVariable().getGlobalNode();
        } else {
          // Variable is not a real one (a generate variable or a module reference).
          newNode = new NodeVariable(variable);
        }
      }

      // Replace parameters and generate variables with the corresponding values.
      return Transformer.substitute(newNode, (NodeProvider) variable1 -> {
        final String fullName = variable1.getName();
        final VerilogParameter parameter = parameters.get(fullName);
        if (parameter != null) {
          // Substitute the parameter with its value.
          return new NodeValue(parameter.getValue());
        }

        final Data value = context.getGenvarValue(fullName);
        if (value != null) {
          // Substitute the generate variable with its value.
          return new NodeValue(value);
        }

        final VerilogDescriptor descriptor = variables.get(fullName);
        if (descriptor != null) {
          // Replace the variable with the elaborated one.
          return descriptor.getVariable().getGlobalNode();
        }

        // Variable is a module reference.
        return new NodeVariable(variable1);
      });
    });
  }

  private Node shiftRanges(final Node node) {
    return Transformer.transform(node, StandardOperation.BVEXTRACT, new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVEXTRACT);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation operation = (NodeOperation) node;
        InvariantChecks.checkTrue(operation.getOperandCount() == 3);

        final Node operand0 = operation.getOperand(0);
        final Node operand1 = operation.getOperand(1);
        final Node operand2 = operation.getOperand(2);

        if (operand2.getKind() != Node.Kind.VARIABLE) {
          return node;
        }

        final NodeVariable operand = (NodeVariable) operand2;

        // If there is a binding, then the variable has not been substituted yet.
        final VerilogBinding binding = context.getBindings().get(operand.getName());
        final NodeVariable variable = binding != null ? binding.getLhs().getGlobalNode() : operand;

        final VerilogDescriptor descriptor = variables.get(variable.getName());
        InvariantChecks.checkNotNull(descriptor,
            String.format("Descriptor for '%s' has not been found: %s", variable, variables));

        int firstCandidate;
        int secondCandidate;

        try {
          firstCandidate = VerilogCalculator.evaluateInteger(operand1);
          secondCandidate = VerilogCalculator.evaluateInteger(operand0);
        } catch (Exception ex) {
          if (descriptor.getOffset() == 0) {
            return node;
          }

          final Node offset = NodeValue.newBitVector(descriptor.getOffset(), Integer.SIZE);
          final Node upper = VerilogCalculator.reduce(Nodes.bvsub(operand0, offset));
          final Node lower = VerilogCalculator.reduce(Nodes.bvsub(operand1, offset));

          return Nodes.bvextract(upper, lower, operand2);
        }

        final int lower = Math.min(firstCandidate, secondCandidate);
        final int upper = Math.max(firstCandidate, secondCandidate);

        final int lowerBound = descriptor.getOffset();
        final int upperBound = descriptor.getDataType().getSize() + lowerBound - 1;

        if (lowerBound == 0 && lower >= lowerBound && upper <= upperBound) {
          return node;
        }

        final Node upperNode = outOfBoundChecker(upper, operand2, upperBound, lowerBound);
        final Node lowerNode = outOfBoundChecker(lower, operand2, upperBound, lowerBound);

        return Nodes.bvextract(upperNode, lowerNode, operand2);
      }
    });
  }

  private Node outOfBoundChecker(final int index, final Node vector, final int upperBound,
      final int lowerBound) {
    if (index > upperBound || index < lowerBound) {
      final IndexOutOfBoundStructure indexOutOfBoundStructure =
          new IndexOutOfBoundStructure(index, vector, upperBound, lowerBound);
      VerilogLogger.warn(indexOutOfBoundStructure.toString());

      return index > upperBound
          ? NodeValue.newBitVector(upperBound - lowerBound, Integer.SIZE)
          : NodeValue.newBitVector(0, Integer.SIZE);
    }

    return NodeValue.newBitVector(index - lowerBound, Integer.SIZE);
  }
}
