/*
 * Copyright 2012-2020 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.util;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.Token;
import org.antlr.runtime.TokenSource;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.Stack;

/**
 * Composite token source for hierarchically organized sub-sources.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class TokenSourceStack implements TokenSource {
  public static final int NEW_SOURCE_TOKEN = Integer.MAX_VALUE;

  public static class TokenSourceEntry {
    public final TokenSource source;

    /** The latest token of the parent stream. */
    public Token lastParentToken = null;

    public TokenSourceEntry(final TokenSource source) {
      InvariantChecks.checkNotNull(source);
      this.source = source;
    }
  }

  private static final Token EOF_TOKEN = new CommonToken(Token.EOF);
  private Stack<TokenSourceEntry> sources = new Stack<>();
  private final TokenSourceCallback callback;

  public TokenSourceStack() {
    this.callback = null;
  }

  public TokenSourceStack(final TokenSourceCallback callback) {
    InvariantChecks.checkNotNull(callback);
    this.callback = callback;
  }

  public void push(final TokenSource source) {
    sources.push(new TokenSourceEntry(source));
  }

  public void pop() {
    final TokenSourceEntry entry = sources.pop();

    // If the callback is specified, it is launched.
    if (callback != null) {
      callback.action(entry.source);
    }
  }

  public TokenSource getSource() {
    return sources.peek().source;
  }

  public boolean hasSources() {
    return !sources.empty();
  }

  private static boolean isEof(final Token token) {
    return token == null || token.getType() == Token.EOF;
  }

  private Token getNextToken() {
    // If there are no sources, returns EOF.
    if (!hasSources()) {
      return EOF_TOKEN;
    }

    Token token = EOF_TOKEN;
    boolean subsourceCreated = true;

    while (subsourceCreated) {
      // Request the active source for the next token.
      token = getSource().nextToken();

      // New sub-sources were created during the recent nextToken() call and pushed into the stack
      // (e.g., for macro expansion or file inclusion). Note that if there are several includes in
      // a row, the corresponding number of sub-sources are created.
      subsourceCreated = (token.getType() == NEW_SOURCE_TOKEN);
    }

    return token;
  }

  @Override
  public Token nextToken() {
    Token token = EOF_TOKEN;

    while (!sources.isEmpty()) {
      token = getNextToken();

      // Skip EOFs of sub-sources (sub-sources are invisible for a user).
      if (!isEof(token)) {
        return token;
      }

      // Remove an exhausted sub-source from the stack.
      pop();
    }

    return token;
  }

  @Override
  public String getSourceName() {
    if (!hasSources()) {
      return "";
    }

    final TokenSource source = getSource();
    return source.getSourceName();
  }
}
