/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.verilog.parser.model.util;

import org.junit.Assert;
import org.junit.Test;

import ru.ispras.verilog.parser.model.basis.VerilogLiteral;

import java.math.BigInteger;

/**
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class NumberTestCase {
  public NumberTestCase() {}

  @Test
  public void runUnderscoreTest() {
    final String[] numbers = {
      "100'h0123456789abcdef",
      "100'h01234567_89ABCDEF",
      "100'h0123_4567_89ab_cdef",
      "100'h01_23_45_67_89_ab_cd_ef",
      "100'h0_1_2_3_4_5_6_7_8_9_A_B_C_D_E_F",
      "100'h0__1__2__3__4__5__6__7__8__9__a__b__c__d__e__f",
      "100'h01234567________89abcdef________"
    };

    runTest(numbers, new BigInteger("81985529216486895"));
  }

  @Test
  public void runSizeBaseTest() {
    final String[] numbers = {
      "32'b10101010101010101010101010101010",
      "'b10101010101010101010101010101010",
      "32'o25252525252",
      "'o25252525252",
      "32'haaaaaaaa",
      "'haaaaaaaa",
      "32'd2863311530",
      "'d2863311530",
      "2863311530"
    };

    runTest(numbers, new BigInteger("2863311530"));
  }


  @Test
  public void runSpecialSymbolsTest() {
    String[] numbers = {
      "100'h1230_5670_9ab0_def",
      "100'h123x_567X_9abX_def",
      "100'h123z_567Z_9abZ_def",
      "100'h123?_567?_9ab?_def",
      "100'h123x_567?_9abz_def"
    };

    runTest(numbers, new BigInteger("81915158324776431"));
  }

  @Test
  public void runZeroTest() {
    final String[] numbers = {
      "0'b0",
      "'b0",
      "0'b",
      "'b",
      "0",
      "",
      "1000000'b00000000000000000000000"
    };

    runTest(numbers, BigInteger.ZERO);
  }

  @Test
  public void runXzOnlyTest() {
    final String[] numbers = {
      "0'bX",
      "'bX",
      "0'b",
      "'b",
      "2'hX",
      "4'hX",
      "0'bZ",
      "'bZ",
      "0'b",
      "'b",
      "2'hZ",
      "4'hZ"
    };

    runTest(numbers, BigInteger.ZERO);
  }

  @Test
  public void runTwoTest() {
    final String[] numbers = {
      "2",
      "002",
      "2'h2",
      "3'h2",
      "4'h2",
      "5'h2"
    };

    runTest(numbers, BigInteger.valueOf(2));
  }

  @Test
  public void runRealTest() {
    final String[] numbers = {
      "10.5",
      "10.5e0",
      "10.5e+0",
      "10.5e-0",
      "1.05e1",
      "1.05e+1",
      "105.0e-1"
    };

    runTest(numbers, 10.5);
  }


  /**
   * Parses the strings of the {@code equalNumbers} array and checks whether the results are
   * equal to {@code goldenValue}.
   * 
   * @param equalNumbers the array of strings representing numbers.
   * @param goldenValue the golden value.
   */
  private void runTest(final String[] equalNumbers, final BigInteger goldenValue) {
    for (int i = 0; i < equalNumbers.length; i++) {
      final String number = equalNumbers[i];
      final VerilogLiteral result = Parser.parseNumber(number);

      Assert.assertTrue(result.isBitVector() || result.isInteger());

      if (!goldenValue.equals(result.getBitVector().bigIntegerValue(false))) {
        Assert.fail("Number " + number + " is not equal to " + goldenValue);
      }
    }
  }

  /**
   * Parses the strings of the {@code equalNumbers} array and checks whether the results are
   * equal to {@code goldenValue}.
   * 
   * @param equalNumbers the array of strings representing numbers.
   * @param goldenValue the golden value.
   */
  private void runTest(final String[] equalNumbers, final double goldenValue) {
    for (int i = 0; i < equalNumbers.length; i++) {
      final String number = equalNumbers[i];
      final VerilogLiteral result = Parser.parseNumber(number);

      Assert.assertTrue(result.isReal());

      if (result.getReal() != goldenValue) {
        Assert.fail("Number " + number + " is not equal to " + goldenValue);
      }
    }
  }
}

