
`include "dispatcher.v"

module dispatcher_checker (
  input clk,
  input reset,

  input [255:0] descriptor_writedata,
  input [31:0] descriptor_byteenable,
  input descriptor_write,
  input descriptor_waitrequest,
  
  input [255:0] snk_descriptor_data,
  input snk_descriptor_valid,
  input snk_descriptor_ready,

  input [31:0] csr_writedata,
  input [3:0] csr_byteenable,
  input csr_write,
  input [31:0] csr_readdata,
  input csr_read,
  input [2:0] csr_address,
  input csr_irq,

  // Used by a host with a master (like Nios II)
  input [31:0] mm_response_readdata,
  input mm_response_read,
  input mm_response_address,
  input [3:0] mm_response_byteenable,
  input mm_response_waitrequest,

  // Used by a pre-fetching master
  input [255:0] src_response_data,      // making wide in case we need to jam more signals in here, unnecessary bits will be grounded/optimized away
  input src_response_valid,
  input src_response_ready,

  input [255:0] src_write_master_data,  // don't know how many bits the master will use, unnecessary bits will be grounded/optimized away 
  input src_write_master_valid,
  input src_write_master_ready,

  input [255:0] snk_write_master_data,       // might need to jam more bits in......
  input snk_write_master_valid,
  input snk_write_master_ready,

  input [255:0] src_read_master_data,  // don't know how many bits the master will use, unnecessary bits will be grounded/optimized away 
  input src_read_master_valid,
  input src_read_master_ready,

  input [255:0] snk_read_master_data,       // might need to jam more bits in......
  input snk_read_master_valid,
  input snk_read_master_ready
);

wire data_push;
wire data_pop;

assign data_push = descriptor_write && ~descriptor_waitrequest && descriptor_byteenable[31] && descriptor_writedata[255];
assign data_pop = src_read_master_valid && src_read_master_ready && !dispatcher.the_descriptor_buffers.read_park_enable;

reg [7:0]desc_in_cnt;
always @ (posedge clk or posedge reset)
if (reset) begin
   desc_in_cnt <= 0;
end
else begin
   desc_in_cnt <= desc_in_cnt + (data_push) - (data_pop);
end

//data path

reg symb_data_in;
reg symb_data_out;
reg [255:0]symb_data;
reg [7:0]data_ahead_cnt;

always @ (posedge clk or posedge reset)
if (reset) begin
   symb_data_in <= 0;
   symb_data_out <= 0;
end
else if (symb_data_in && data_pop && (data_ahead_cnt==1)) begin
   symb_data_out <= 1;
end
else if (data_push && &descriptor_byteenable && (descriptor_writedata==symb_data)) begin
   symb_data_in <= 1;
end

always @ (posedge clk or posedge reset)
if (reset) begin
   data_ahead_cnt <= 0;
end
else begin
   data_ahead_cnt <= data_ahead_cnt + (data_push && !symb_data_in) - (data_pop && !symb_data_out);
end

//verify at fifo depth 4, then assume for larger depth
reg [7:0]symb_data_ptr;
always @ (posedge clk)
if (data_push && (descriptor_writedata==symb_data) && !symb_data_in) symb_data_ptr <= dispatcher.the_descriptor_buffers.the_read_command_FIFO.write_address;

dispatcher dispatcher(
    .clk (clk),
    .reset (reset),
    .descriptor_writedata (descriptor_writedata),
    .descriptor_byteenable (descriptor_byteenable),
    .descriptor_write (descriptor_write),
    .descriptor_waitrequest (descriptor_waitrequest),
    .snk_descriptor_data (snk_descriptor_data),
    .snk_descriptor_valid (snk_descriptor_valid),
    .snk_descriptor_ready (snk_descriptor_ready),
    .csr_writedata (csr_writedata),
    .csr_byteenable(csr_byteenable),
    .csr_write (csr_write),
    .csr_readdata (csr_readdata),
    .csr_read (csr_read),
    .csr_address (csr_address),
    .csr_irq (csr_irq),
    .mm_response_readdata (mm_response_readdata),
    .mm_response_read (mm_response_read),
    .mm_response_address (mm_response_address),
    .mm_response_byteenable (mm_response_byteenable),
    .mm_response_waitrequest (mm_response_waitrequest),
    .src_response_data (src_response_data),
    .src_response_valid (src_response_valid),
    .src_response_ready (src_response_ready),
    .src_write_master_data (src_write_master_data),
    .src_write_master_valid (src_write_master_valid),
    .src_write_master_ready (src_write_master_ready),
    .snk_write_master_data (snk_write_master_data),
    .snk_write_master_valid (snk_write_master_valid),
    .snk_write_master_ready (snk_write_master_ready),
    .src_read_master_data (src_read_master_data),
    .src_read_master_valid (src_read_master_valid),
    .src_read_master_ready (src_read_master_ready),
    .snk_read_master_data (snk_read_master_data),
    .snk_read_master_valid (snk_read_master_valid),
    .snk_read_master_ready (snk_read_master_ready)
  );

endmodule
